/*-------------------------------------------------------------------------*/
/*                                                                         */
/*  COPYRIGHT[copyright mark] 1996, Hughes Aircraft Company, its vendors,  */
/*  and suppliers.  ALL RIGHTS RESERVED.                                   */
/*                                                                         */
/*-------------------------------------------------------------------------*/
/*******************************************************************************
BEGIN_FILE_PROLOG:

FILENAME:
  PGS_CSC_ECRtoGEO.c

DESCRIPTION:
  This file contains the function PGS_CSC_ECRtoGEO()

AUTHOR:
  Peter Noerdlinger  / Applied Research Corporation
  Anubha Singhal     / Applied Research Corporation
  Guru Tej S. Khalsa / Applied Research Corporation

HISTORY:
  01-Feb-1994       PDN   Designed       
  19-Mar-1994       PDN   Coded
  31-Aug-1994       AS    Included call to PGS_CSC_GetEarthFigure and changed
                          calling sequence
  15-Sep-1994       AS    Modified prolog to conform to latest ECS/PGS 
                          standards
  21-Sep-1994       AS    Made changes from code inspection
  22-Sep-1994       PDN   Designed new algorithm for determining the latitude
                          and altitude (more robust at large geocentric 
                          distances; converges faster; no trig functions in
                          iterative loop)
  23-Sep-1994       AS    Coded new algorithm
  31-May-1995      GTSK   Altered algorithm to remember last earth tag passed
                          in. 
  12-Jul-1995       PDN   Rewrote prolog; changed comments
  29-Oct-1996       PDN   Put in early check point is not too
                          near Earth center - convergence is poor there
                          Output approximate answer in that case
  02-Jan-1997       PDN   Replaced the above approximation with the Borkowski
                          method

END_FILE_PROLOG:
*******************************************************************************/
/*******************************************************************************
BEGIN_PROLOG:

TITLE:  
      This function converts from ECR to geodetic coordinates

NAME:   
      PGS_CSC_ECRtoGEO()                                    

SYNOPSIS:
C:
      #include <PGS_CSC.h>

      PGSt_SMF_status
      PGS_CSC_ECRtoGEO(
             PGSt_double       posECR[3],
             char              *earthEllipsTag,
             PGSt_double       *longitude,
             PGSt_double       *latitide,
             PGSt_double       *altitude);

FORTRAN:     
      include 'PGS_SMF.f'
      include 'PGS_CSC_4.f'
 
      integer function  pgs_csc_ecrtogeo(posecr,earthellipstag,longitude,
                                         latitude,height,
         
      double precision  posecr(3)
      character*20      earthellipstag
      double precision  longitude
      double precision  latitude
      double precision  altitude
              
DESCRIPTION:
      This function converts from ECR to geodetic coordinates.

INPUTS:
      Name            Description           Units   Min           Max
      ----            -----------           -----   ---           ---          
      posECR[3]       Geocentric position   meters  N/A           N/A

      earthEllipsTag  earth model           N/A     N/A           N/A          
                      used                                           
						 
OUTPUTS:
      Name          Description           Units       Min       Max
      ----          -----------           -----       ---       ---     
      latitude      geodetic latitude     radians     -pi/2     pi/2

      longitude     longitude             radians     -pi       pi
 
      altitude      altitude               meters    -.1*Earth   sky's the limit
 
RETURNS:
      PGS_S_SUCCESS		    Successful return
      PGSCSC_W_TOO_MANY_ITERS       Normal Iteration Count exceeded - could
                                    indicate inconsistent units for Spacecraft
                                    and Earth data, or corrupted Earth Axis
                                    values
      PGSCSC_W_INVALID_ALTITUDE     Spacecraft underground - probably 
                                    indicates bad input data
      PGSCSC_W_SPHERE_BODY          using a spherical earth model
      PGSCSC_W_LARGE_FLATTENING     issued if flattening factor is greater
                                    then 0.01  
      PGSCSC_W_DEFAULT_EARTH_MODEL  Uses default earth model
      PGSCSC_E_BAD_EARTH_MODEL      The equatorial or polar radius is negative
                                    or zero OR the radii define a prolate Earth
      PGS_E_TOOLKIT                 Something unexpected happened, execution
                                    of function ended prematurely

EXAMPLES:
C:
      PGSt_SMF_status      returnStatus
      PGSt_double          longitude
      PGSt_double          latitude
      PGSt_double          altitude
      char                 earthEllipsTag[5],
      PGSt_double          posECR[3] = {1000.5,64343.56,34343.92} 
      char                 err[PGS_SMF_MAX_MNEMONIC_SIZE];  
      char                 msg[PGS_SMF_MAX_MSG_SIZE];      
      
      strcpy(earthEllipsTag,"WGS84"); 
       
      returnStatus = PGS_CSC_ECRtoGEO(posECR[3],earthEllipsTag,longitude,
                                      latitude,altitude);
      if(returnStatus != PGS_S_SUCCESS)
      {
	  PGS_SMF_GetMsg(&returnStatus,err,msg);
	  printf("\nERROR: %s",msg);
      }

FORTRAN:
      implicit none
      integer              pgs_csc_geotoecr
      integer              returnstatus
      double precision     longitude
      double precision     latitude
      double precision     altitude
      character*20         earthellipstag,
      double precision     posecr(3)
      character*33 	   err
      character*241 	   msg

      data posecr/1000.5,64343.56,34343.92/
      earthellipstag = 'WGS84'

      returnstatus = pgs_csc_ecrtogeo(posecr,earthellipstag,longitude,
     >                                latitude,altitude)

      if(returnstatus .ne. pgs_s_success) then
	  returnstatus = pgs_smf_getmsg(returnstatus,err,msg)
	  write(*,*) err, msg
      endif

NOTES:  
      The Earth axes will accessed from the earthfigure.dat file.
      The input must always be in meters and should never be a unit vector.

REQUIREMENTS: 
      PGSTK-1050, 930

DETAILS:
      See "Theoretical Basis of the SDP Toolkit Geolocation Package for the
      ECS Project", Document 445-TP-002-002, May 1995, by P. Noerdlinger, 
      for more information on the algorithm.

      In brief, the method is:
      To find the latitude and altitude, denote the point (X,Y,Z) as B and
      normalize all lengths by dividing by equatRad_A, the Earth's 
      equatorial radius.  Let a line be drawn through B and the center of 
      the Earth, intersecting the Earth ellipsoid at point U . The normal 
      to the ellipsoid is then dropped from U towards the Earth's equatorial 
      plane, intersecting it at point Q, and a distance lengthOQ computed 
      from Q to earth center O. Then a line BQ is drawn through B and Q, 
      (the algebra requires use of the value of lengthOQ) and a new normal is
      dropped from the intersection (again denoted U) of BQ with the 
      ellipsoid. The algorithm converges in a few iterations such that the 
      normal becomes coincident with the line BQ, and U becomes the footpoint
      of that normal on the ellipsoid.   The altitude is the (signed) distance
      UB and the latitude is easily computed from the coordinates of U.
      As usual, the longitude is found from atan(Y,X).


GLOBALS:
      NONE

FILES:
      earthfigure.dat
 
FUNCTIONS_CALLED:
      PGS_SMF_SetStaticMsg      set error/status message
      PGS_SMF_SetDynamicMsg     set error/status message
      PGS_SMF_SetUnknownMsg     set error/status message for unknown message
      PGS_CSC_GetEarthFigure    get equatorial and polar radius 

END_PROLOG:   
*******************************************************************************/

#include <string.h>
#include <PGS_math.h>            /*  Standard header file for math libraries */
#include <PGS_CSC.h>         /*  General header file for Geolocation Tools */

/* name of this function */

#define FUNCTION_NAME "PGS_CSC_ECRtoGEO()"

PGSt_SMF_status  
PGS_CSC_ECRtoGEO(                 /* converts from ECR to geodetic
				     coordinates */
    PGSt_double  posECR[3],       /* Geocentric position vector in ECR */
    char         *earthEllipsTag, /* tag for selecting Earth Ellipsoid model */
    PGSt_double  *longitude,      /* latitude in radians */
    PGSt_double  *latitude,       /* longitude in radians */
    PGSt_double  *altitude)       /* altitude in meters */
{
    int          ncount = 0;      /* number of iterations */
    
    PGSt_double  rho;             /* Component of ECR vector in equatorial 
                                     plane */
    PGSt_double  normRho;         /* Component of ECR vector in equatorial 
                                     plane normalized */ 
    PGSt_double  normZ;           /* Z component of ECR vector normalized */
    PGSt_double  lengthOQ;        /* distance of Q from O */
    PGSt_double  signn;           /* used to check if the point is in or 
                                     outside the ellipsoid */
    PGSt_double  lengthOQStore;   /* stored value of lengthOQ */
    PGSt_double  normRhoFoot;     /* normalized radial cylindrical coordinate of
                                     point U at foot of normal from B to 
                                     ellipsoid */
    PGSt_double  normZFoot;       /* location of point at foot of normal */
    PGSt_double  srct;            /* used for obtaining geocentric coordinates
                                     of U reference: denominator of equations on
				     p. 29 of "Methods of Orbit Determination"
				     by P.Escobal (Wiley, 1965) */
    
    register PGSt_double sinlat;  /* Sine of the latitude */
    register PGSt_double coslat;  /* Cosine of the latitude */

    PGSt_SMF_status returnStatus; /* return status */

    /* returnStatus of call to PGS_CSC_GetEarthFigure() */

    static PGSt_SMF_status modelStatus=PGS_S_SUCCESS;
    
    /* earth tag from any previous call to this function, initialized to some
       nonsense value (hopefully) */

    static char oldEarthTag[50]="THIS IS THE DEFAULT (WGS84) EARTH MODEL";

    /* Earth radii (initialized to default (WGS84) values) */

    static PGSt_double equatRad_A=6378137.0;        /* equatorial radius */
    static PGSt_double polarRad_C=6356752.314245;    /* polar radius */

    /* ellipsoid eccentricity (initialized to default (WGS84) value) */
    
    static PGSt_double ellipsEccSq=6.69438002301199714E-3;

    /* flattening factor - which is related to eccentricity of the ellipsoid of
       revolution (initialized to default (WGS84) value) */

    static PGSt_double flatFac=3.35281068123811075E-3;

    /* Initialization */

    returnStatus = modelStatus;

    /* if the earthEllipsTag passed in is not the same as the last value used,
       get equatorial and polar radius and calculate flattening and ellipsoid
       eccentricity */

    if (strcmp(earthEllipsTag,oldEarthTag) != 0) /* BEGIN setup */
    {
	
	returnStatus = PGS_CSC_GetEarthFigure(earthEllipsTag,&equatRad_A,
					      &polarRad_C);
	switch (returnStatus)
	{
	  case PGS_S_SUCCESS:
	  case PGSCSC_W_DEFAULT_EARTH_MODEL:
	    break;
	  default:
	    PGS_SMF_SetUnknownMsg(returnStatus,FUNCTION_NAME);
	    return PGS_E_TOOLKIT;
	}
	
	/* issue an error if the equatorial radius is negative */
	
	if (equatRad_A <= 0.0)
	{
	    returnStatus = PGSCSC_E_BAD_EARTH_MODEL;
	    PGS_SMF_SetDynamicMsg(returnStatus,
				  "the equatorial radius is negative or zero",
				  FUNCTION_NAME);
	    return returnStatus;
	}
	
	/* issue an error if the polar radius is negative */
	
	if (polarRad_C <= 0.0)
	{
	    returnStatus = PGSCSC_E_BAD_EARTH_MODEL;
	    PGS_SMF_SetDynamicMsg(returnStatus,"the polar radius is negative "
				  "or zero",FUNCTION_NAME);
	    return returnStatus;
	}
	
	/* determine the flattening factor */
	
	flatFac = (equatRad_A - polarRad_C) / equatRad_A;
	
	/* check for errors in the value of the flattening factor */
	
	if (flatFac < 0.0)
	{
	    returnStatus = PGSCSC_E_BAD_EARTH_MODEL;
	    PGS_SMF_SetDynamicMsg(returnStatus,"the specified earth model "
				  "defines a prolate earth",FUNCTION_NAME);
	    return returnStatus;
	} 
	else if (flatFac == 0.0)
	{
	    returnStatus = PGSCSC_W_SPHERE_BODY;
	    PGS_SMF_SetDynamicMsg(returnStatus,
				  "a spherical earth model is being used",
				  FUNCTION_NAME);
	}
	else if (flatFac > 0.01)
	{
	    returnStatus = PGSCSC_W_LARGE_FLATTENING;
	    PGS_SMF_SetStaticMsg(returnStatus,FUNCTION_NAME);
	}
	
	/*  determine the square of eccentricity of the ellipsoid */
	
	ellipsEccSq = (2.0 * flatFac) - (flatFac * flatFac);
	
	/* set the oldEarthTag to the current value */
	
	oldEarthTag[49] = '\0';
	strncpy(oldEarthTag,earthEllipsTag,49);

	/* set modelStatus to current value of returnStatus */

	modelStatus = returnStatus;

    } /* END setup */
    else if (returnStatus != PGS_S_SUCCESS)
	PGS_SMF_SetStaticMsg(returnStatus,FUNCTION_NAME);
    
    /* Compute radius projected onto X-Y plane */

    rho = sqrt(posECR[0]*posECR[0] + posECR[1]*posECR[1]);  
                                                               
    /*  Compute Longitude */

    if ( rho > 0.0 )
	*longitude = atan2(posECR[1], posECR[0]);     /* Arctangent of Y/X */
    else
	*longitude = 0.0;

    /* normalize the coordinates */

    normRho  = rho / equatRad_A;
    normZ    = posECR[2]/ equatRad_A;
        
    /* if point is within 0.5 Earth radii of the center convergence
       may be poor.  Use Borkowski method (actually quite good all
       over - but perhaps slower */ 

    if ( normRho*normRho + normZ*normZ < 0.80 )
    { 
       returnStatus = PGS_CSC_BorkowskiGeo(equatRad_A,polarRad_C,rho,
                       posECR[2],latitude,altitude);
                     
       returnStatus = PGSCSC_W_INVALID_ALTITUDE;
       PGS_SMF_SetStaticMsg(returnStatus,FUNCTION_NAME);
       return returnStatus;
 
    } 
    
    /* test if the point is inside or outside the ellipsoid */

    signn = (normRho*normRho + normZ*normZ / 
             (1.0 - ellipsEccSq)) > 1.0 ? 1.0 : -1.0;

    /* initialize the distance of Q from O and the stored value */

    lengthOQ  = 0.0;
    lengthOQStore = 1.0;
 
    /* iterate to about 11 significant digits accuracy (0.05 mm) */

    while (((fabs(lengthOQStore  - lengthOQ)) > 1.0e-11) && (ncount < 10))
    {
        /* reinitialize the lagged value lengthOQ */

	lengthOQStore = lengthOQ;

        /* increment ncount */
	++ncount;

        /* calculate the new length */

	lengthOQ = ellipsEccSq * (normRho - lengthOQStore) / 
                   sqrt ((normRho - lengthOQStore) * (normRho - lengthOQStore)
                   + normZ * normZ * (1.0 - ellipsEccSq));
	
    }

    /* issue a warning if the number of iterations ran too high */

    if (ncount == 10)
    {
	returnStatus = PGSCSC_W_TOO_MANY_ITERS;
	PGS_SMF_SetStaticMsg(returnStatus,FUNCTION_NAME);
    }
    
    /* calculate the latitude */

    *latitude = atan2(normZ, normRho - lengthOQ);
    coslat = cos(*latitude);
    sinlat = sin(*latitude);
    srct = 1.0/sqrt(1.0 - ellipsEccSq * sinlat * sinlat);

    /* determine the location of point U on geoid at foot of normal from B to
       geoid (normalized coordinates) */

    normRhoFoot = coslat * srct;
    normZFoot = (1.0 - ellipsEccSq) * sinlat * srct;

    /* calculate the true altitude = normalized altitude * equatorial radius */

    *altitude  = signn * sqrt((normRhoFoot-normRho) * (normRhoFoot-normRho) + 
                 (normZFoot-normZ) * (normZFoot-normZ)) * equatRad_A;

    /* check for an invalid altitude */
    
    if (*altitude < (-0.1 * equatRad_A))
        if (returnStatus != PGSCSC_W_TOO_MANY_ITERS)
	{
	    returnStatus = PGSCSC_W_INVALID_ALTITUDE;
	    PGS_SMF_SetStaticMsg(returnStatus,FUNCTION_NAME);
	}

    if (returnStatus == PGS_S_SUCCESS)  
	PGS_SMF_SetStaticMsg(returnStatus,FUNCTION_NAME);

    return returnStatus;
}
