/*-------------------------------------------------------------------------*/
/*                                                                         */
/*  COPYRIGHT[copyright mark] 1996, Hughes Aircraft Company, its vendors,  */
/*  and suppliers.  ALL RIGHTS RESERVED.                                   */
/*                                                                         */
/*-------------------------------------------------------------------------*/
/*******************************************************************************
BEGIN_FILE_PROLOG:

FILENAME:
   PGS_CSC_QuatToEuler.c

DESCRIPTION:
   This file contains the function PGS_CSC_QuatToEuler().
   This function gets Euler angles from a quaternion.
   
AUTHOR:
   Guru Tej S. Khalsa / Applied Research Corporation
   Anubha Singhal     / Applied Research Corporation

HISTORY:
   14-Oct-1994 AS/GTSK Initial version

END_FILE_PROLOG:
*******************************************************************************/
/*******************************************************************************
BEGIN_PROLOG

TITLE: 
   Obtains Euler angles from a quaternion

NAME:   
   PGS_CSC_QuatToEuler()

SYNOPSIS:
 C:
   #include <PGS_CSC.h>
   
   PGSt_SMF_status
   PGS_CSC_QuatToEuler( 
       PGSt_double  quaternion[4],            
       PGSt_integer eulerAngleOrder[3],
       PGSt_double  eulerAngle[3])   
	      
 FORTRAN:
      include 'PGS_CSC_4.f'
      include 'PGS_SMF.f'

      integer function pgs_csc_quattoeuler(quaternion,eulerangleorder,
     >                                     eulerangle)
      double precision quaternion(4)
      integer          eulerangleorder(3)
      double precision eulerangle(3)
    
DESCRIPTION:
   This function gets euler angles from a quaternion.

INPUTS:
   Name              Description            Units    Min        Max
   ----              -----------            -----    ---        --- 
   quaternion        quaternion defining    N/A      N/A        N/A
                     the desired rotation                
   eulerAngleOrder   order of rotations     N/A      N/A        N/A  
                                                   
OUTPUTS:
   Name              Description            Units    Min        Max
   ----              -----------            ----     ---        ---
   eulerAngle        Euler angles           radians  -Pi        Pi
                                                  
RETURNS:
   PGS_S_SUCCESS		 Success
   PGSCSC_E_EULER_REP_INVALID    Euler angle order is invalid
   PGSCSC_E_EULER_INDEX_ERROR    Error in value of euler index
   PGSCSC_E_BAD_QUATERNION       Input quaternion is invalid

NOTES:
   The indices 1, 2, 3 are used for the Euler angle order. They refer to the
   rotations about the body axes labeled 1, 2 and 3 which are assumed to be x,
   y and z.  The angles of rotation are phi, theta and psi in the order
   performed.  Thus, if the Euler angle order is, for example, 2, 1, 3, it means
   we rotate phi about axis 2, then by theta about new axis 1, then by psi about
   new axis 3.

   If the 3 Euler rotation axes are all different(e.g 3,1,2), the quaternion 
   is obtained from the following equations (ref: Schuster[1] Eq(191))

   xi(i,j) = 1.0 if {i,j} = 1,2 or 2,3 or 3,1
           = -1.0 if {i,j} = 2,1 or 3,2 or 1,3
	   = 0 otherwise

   [1] "A Survey of Attitude Representations", by Malcolm D. Schuster, The
       Journal of the Astronautical Sciences vol 41, #4, pp. 439-517 (AIAA,
       Oct 1993)

REQUIREMENTS:
   PGSTK - 1050 

DETAILS:
      See "Theoretical Basis of the SDP Toolkit Geolocation Package for the
      ECS Project", Document 445-TP-002-002, May 1995, by P. Noerdlinger, 
      for more information on the algorithm.

GLOBALS:
   NONE

FILES:
   NONE

FUNCTIONS CALLED:
   PGS_SMF_SetStaticMsg()
      
END_PROLOG
*******************************************************************************/

#include <PGS_math.h>
#include <PGS_CSC.h>

/* name of this function */

#define FUNCTION_NAME "PGS_CSC_QuatToEuler()"

PGSt_SMF_status
PGS_CSC_QuatToEuler(                       /* obtains the equivalent Euler
					      angles from a quaternion */  
    PGSt_double         quaternion[],      /* orbital */
    PGSt_integer        eulerAngleOrder[], /* order of the rotations */
    PGSt_double	        eulerAngle[])      /* Euler angles */ 
{  
    short               cnt;               /* loop counter */

    static PGSt_integer iIndex;            /* eulerAngleOrder[0] */
    static PGSt_integer jIndex;            /* eulerAngleOrder[1] */
    static PGSt_integer kIndex;            /* eulerAngleOrder[2] */
    static PGSt_integer oldEAOrder[3];     /* saved Euler angle order */ 
    static PGSt_integer psi;               /* right hand (1) or left hand (-1)
					      order */
    PGSt_double         alpha;             /* intermediate angle */
    PGSt_double         beta;              /* intermediate angle */
    PGSt_double         sumSquares;        /* sum of the squares of the
					      components of the quaternion */
    static PGSt_boolean indicesDiff;       /* whether a symmetric or asymmetric
					      Euler angle order */
    PGSt_SMF_status     returnStatus;      /* return status of function call */

    /* initialize return value to indicate success */

    returnStatus = PGS_S_SUCCESS;
    
    /* Make sure the input quaternion is indeed a quaternion.  The sum of the
       squares of the components of a quaternion should equal one (1). */

    sumSquares = quaternion[0]*quaternion[0] + quaternion[1]*quaternion[1] +
                 quaternion[2]*quaternion[2] + quaternion[3]*quaternion[3];
    if (sumSquares < .99999999 || sumSquares > 1.00000001)
    {
	returnStatus =  PGSCSC_E_BAD_QUATERNION;
	PGS_SMF_SetStaticMsg(returnStatus,FUNCTION_NAME);
	return returnStatus;
    }
    
    /* check if we have used this function  with the same euler angle order-- 
       if so, omit setup and proceed to obtain the Euler angles */ 
    
    if ((eulerAngleOrder[0] != oldEAOrder[0]) ||
	(eulerAngleOrder[1] != oldEAOrder[1]) ||
	(eulerAngleOrder[2] != oldEAOrder[2]) )
    {
	
	/* test that none of the values in the euler angle order is less than 
           0 or greater than 2 */

	for (cnt = 0 ;cnt < 3; cnt++)
	  switch (eulerAngleOrder[cnt])
	  {
	    case 1:
	    case 2:
	    case 3:
	      break;
	    default:
	      returnStatus = PGSCSC_E_EULER_INDEX_ERROR;
	      PGS_SMF_SetStaticMsg(returnStatus,FUNCTION_NAME);
	      return returnStatus;
	  }

	iIndex = eulerAngleOrder[0];
	jIndex = eulerAngleOrder[1];
	kIndex = eulerAngleOrder[2];

	/* test to see if all the values in the euler angle order are 
           different */

	if (((iIndex - jIndex) * (iIndex - kIndex) * (jIndex - kIndex)) != 0)
	  indicesDiff = PGS_TRUE;

        /* test to see if the last and first indices are the same and the 
	   middle one is different */

	else if ((iIndex == kIndex) && (jIndex != kIndex))
        {
	    kIndex = 6 - iIndex - jIndex;
	    indicesDiff = PGS_FALSE;
	}
     
        /* euler angle order is invalid - issue error and return */

	else 
	{
	    returnStatus = PGSCSC_E_EULER_REP_INVALID;
	    PGS_SMF_SetStaticMsg(returnStatus,FUNCTION_NAME);
	    return returnStatus; 
	}

	/* determine psi */
        psi = (jIndex  == (iIndex + 1) % 3) - (jIndex == (iIndex + 2) % 3);

	/* save the Euler angle order so that the setup can be omitted if the
	   order is the same the next time the function is called */

	for (cnt = 0; cnt < 3; cnt++)
	  oldEAOrder[cnt] = eulerAngleOrder[cnt];
    } 
    
    /* determine Euler angles for a non-repeating sequence of Euler angles */

    if (indicesDiff == PGS_TRUE)
    {
	eulerAngle[1] = asin(2.0*(psi*quaternion[iIndex]*quaternion[kIndex] +
				  quaternion[jIndex]*quaternion[0]));
	alpha = atan2((quaternion[iIndex] + quaternion[kIndex]),
		       (psi*quaternion[jIndex] + quaternion[0]));
	beta = atan2((quaternion[iIndex] - quaternion[kIndex]),
		    (-psi*quaternion[jIndex] + quaternion[0]));
	eulerAngle[0] = alpha + beta;
	eulerAngle[2] = alpha - beta;
    }

    /* determine Euler angles for a repeating sequence of Euler angles */

    else 
    {
	alpha = atan2(quaternion[iIndex],quaternion[0]);
	beta = atan2(psi*quaternion[kIndex],quaternion[jIndex]);
	eulerAngle[1] = 2.0*asin(sqrt(quaternion[jIndex]*quaternion[jIndex] +
			      quaternion[kIndex]*quaternion[kIndex]));
	eulerAngle[0] = alpha + beta;
	eulerAngle[2] = alpha - beta;
    } 
    
    PGS_SMF_SetStaticMsg(returnStatus,FUNCTION_NAME);
       
    return returnStatus;
}
