/*-------------------------------------------------------------------------*/
/*                                                                         */
/*  COPYRIGHT[copyright mark] 1996, Hughes Aircraft Company, its vendors,  */
/*  and suppliers.  ALL RIGHTS RESERVED.                                   */
/*                                                                         */
/*-------------------------------------------------------------------------*/
#include <PGS_math.h>
#include <PGS_CSC.h>

#define EPOCH_DAY            2448988.5 /* TAI Julian day of 0 hrs UTC 1-1-93 */
#define EPOCH_DAY_FRACTION   0.0003125 /* TAI Julian day fraction 0 hrs UTC */
#define SECONDSperDAY        86400.0   /* number of seconds in a day */ 

#define MAX_TIME_DIFF        1200.0    /* maximum time (in seconds) over which
					  to extrapolate nutation angles */
/* name of this function */

#define FUNCTION_NAME "PGS_CSC_quickWahr()"

PGSt_SMF_status 
PGS_CSC_quickWahr(           /* return nutation angles and rates and TDB at
				input TAI time (extrapolating over short 
				periods) */
    PGSt_double secTAI93,    /* TAI (toolkit internal time) */
    PGSt_double jedTDB[2],   /* TDB Julian Ephemeris Date */
    PGSt_double dvnut[4])    /* nutation angles and rates
				dvnut[0] - nutation in longitude (radians)
				dvnut[1] - nutation in obliquity (radians)
				dvnut[2] - nut. rate in longitude (radians/sec)
				dvnut[3] - nut. rate in obliquity (rad/sec) */

{
    
    PGSt_double jdTAI[2];            /* TAI as a Julian Day  */
    PGSt_double jedTDT[2];           /* TDT expressed in Julian days 
					(leap seconds added in, and 32.184) */
    PGSt_double timeDiff;            /* secTAI93 - oldsecTAI93, used to avoid
					repeated setup */
    
    static PGSt_double jedTDBold[2]; /* TDB expressed in Julian days (additional
					periodic correction) */
    static PGSt_double dvnutOld[4];  /* the two nutation angles and their rates,
					output from "PGS_CSC_wahr2" */

    static PGSt_double oldsecTAI93=-1.e50; /* previous value (to avoid repeated
					      setup) (initialized to junk) */

    timeDiff = secTAI93 - oldsecTAI93;
    
    if ( fabs(timeDiff) > MAX_TIME_DIFF ) /* setup */
    {
	oldsecTAI93 = secTAI93;    
        timeDiff = 0.0;
	
	/* Get TDB - this is needed to get precession and nutation */
	
	/* convert TAI input to Julian date format */
	
	PGS_TD_TAItoTAIjd(secTAI93,jdTAI);
	    	
	/* get TDT from TAI */
	
	PGS_TD_TAIjdtoTDTjed(jdTAI,jedTDT); 
	
	/* compute jedTDB */
	    
	PGS_TD_TDTjedtoTDBjed(jedTDT,jedTDBold);
	        
	/*  Obtain the nutation angles and rates  */
	
	PGS_CSC_wahr2(jedTDT,dvnutOld); 

    } /* end "if ( fabs(timeDiff .." */ 

    /* extrapolate */
    
    dvnut[0] = dvnutOld[0] + dvnutOld[2]*timeDiff;
    dvnut[1] = dvnutOld[1] + dvnutOld[3]*timeDiff;
    dvnut[2] = dvnutOld[2];
    dvnut[3] = dvnutOld[3];
    jedTDB[0] = jedTDBold[0];
    jedTDB[1] = jedTDBold[1] + timeDiff/SECONDSperDAY;

    return PGS_S_SUCCESS;
}




