/*-------------------------------------------------------------------------*/
/*                                                                         */
/*  COPYRIGHT[copyright mark] 1996, Hughes Aircraft Company, its vendors,  */
/*  and suppliers.  ALL RIGHTS RESERVED.                                   */
/*                                                                         */
/*-------------------------------------------------------------------------*/
/*******************************************************************************
BEGIN_FILE_PROLOG:

FILENAME:
   PGS_EPH_EphemAttit.c

DESCRIPTION:
   This file contains the function PGS_EPH_EphemAttit()
   This function gets ephemeris and/or attitude data for the specified
   spacecraft at the specified times.

AUTHOR:
   Guru Tej S. Khalsa / Applied Research Corporation
   Peter D. Noerdlinger / Applied Research Corporation

HISTORY:
   08-Apr-1994  GTSK  Initial version
   08-Sep-1994  GTSK  Updated variable types to latest PGS standards.  Added ypr
                      and yprRate to calling sequence.  Improved error handling
		      somewhat.
   01-Oct-1994  GTSK  Improved error handling.  Fixed code to handle case of
                      numValues = 0.  Return error on numValues < 0.  Continues
		      execution on PGSTD_E_NO_LEAP_SECS error now.
   17-Oct-1994  GTSK  Entirely rewritten to read s/c ephemeris files from disk.
                      Approximately 0% of code is same as previous version.
   01-Jan-1995  GTSK  Removed code to read s/c ephemeris files and moved to a
                      separate function: PGS_EPH_getEphemRecords.  Fixed bug in
		      determining TRMM (geodetic-based attitude) s/c to ECI
		      rotation quaternion.
   01-Feb-1995  GTSK  Cast all pointers being passed in to the memcpy() function
                      to type (char *).  This is to prevent issuance of spurious
                      compiler messages about incompatible arguments from the
                      SunOS 4.x platforms .  It should have no effect on other 
                      platforms since the function memcpy() takes void pointers 
                      and doesn't care what type of pointer is passed in.
   20-Mar-1995  GTSK  Replaced call to malloc() with call to PGS_MEM_Calloc().
                      Replaced string "PGS_EPH_EphemAttit()" with FUNCTION_NAME.
		      Renamed variable startUTC to asciiUTC.
   23-May-1995  GTSK  Added check on time interval between actual records when
                      interpolating.  No interpolation is done if the time 
                      between actual records is > MAX_TIME_INTERVAL.  The
		      quality flag(s) is(are) set to indicate no data for a 
		      requested time for which interpolation can't be done.
   21-Dec-1995  GTSK  Fixed previously implemented scheme to avoid reloading
                      ephemeris data into memory from disk files by checking if
		      the required data was loaded in a previous call to this
		      function.  The time range information was being properly
		      saved but the s/c tag value was NOT being saved (causing
		      this routine to always assume that each time it was called
		      a new value of s/c tag had been passed in).  This has been
		      fixed.
   05-May-1996  GTSK  Implemented major changes to accommodate new
                      ephemeris/attitude file formats.
   27-May-1997  PDN   Deleted references to predicted leap seconds
   12-Feb-1999  PDN   Fixed nomenclature consistent with AM1 and later spacecraft
                      note that the function is also OK for TRMM, but in that case
                      please observe that angularVelocity is really (yaw rate, 
                      pitch rate, and roll rate in that order).
   22-Feb-1999  PDN   Shortened the maximum interpolation interval from 121
                      seconds (useful in I&T) to 60 seconds, consistent with
                      DPREP capability for AM1 and probably later spacecraft.

END_FILE_PROLOG:
*******************************************************************************/

/*******************************************************************************
BEGIN_PROLOG:

TITLE:
   Get Ephemeris and Attitude

NAME:
   PGS_EPH_EphemAttit()

SYNOPSIS:
C:
   #include <PGS_EPH.h>

   PGSt_SMF_status
   PGS_EPH_EphemAttit(
        PGSt_tag        spacecraftTag,   
	PGSt_integer    numValues,        
	char            *asciiUTC,       
	PGSt_double     offsets[],       
	PGSt_boolean    orbFlag,          
	PGSt_boolean    attFlag,
        PGSt_integer    qualityFlags[][2],
	PGSt_double     positionECI[][3], 
	PGSt_double     velocityECI[][3], 
	PGSt_double     eulerAngles[][3],         
	PGSt_double     xyzRotRates[][3],    
	PGSt_double     attitQuat[][4])

FORTRAN:
      include 'PGS_SMF.f'
      include 'PGS_TD.f'
      include 'PGS_TD_3.f'
      include 'PGS_EPH_5.f'
      include 'PGS_MEM_7.f'

      integer function pgs_eph_ephemattit(spacecrafttag,numvalues,
     >                                    asciiutc,offsets,orbflag,
     >                                    attflag,qualityflags,
     > 				          positioneci,velocityeci,
     >                                    eulerangles,xyzrotrates,
     > 				          attitquat)
      integer           spacecrafttag
      integer           numvalues
      character*27      asciiutc
      double precision  offsets(*)
      integer           orbflag
      integer           attflag
      integer           qualityflags(2,*)
      double precision  positioneci(3,*)
      double precision  velocityeci(3,*)
      double precision  eulerangles(3,*)
      double precision  xyzrotrates(3,*)
      double precision  attitquat(4,*)

DESCRIPTION:
   This tool gets ephemeris and/or attitude data for the specified spacecraft
   at the specified times.

INPUTS:
   Name               Description               Units  Min        Max
   ----               -----------               -----  ---        ---
   spacecraftTag      spacecraft identifier     N/A

   numValues          num. of values requested  N/A

   asciiUTC           UTC time reference start  ASCII  1961-01-01 see NOTES
                      time in CCSDS ASCII time
	              code A format

   offsets            array of time offsets in  sec    ** depends on asciiUTC **
                      seconds relative to 
	              asciiUTC

   orbFlag            set to true to get        T/F
                      ephemeris data

   attFlag            set to true to get        T/F
                      attitude data

OUTPUTS:
   Name              Description                Units        Min         Max
   ----              -----------                -----        ---         ---
   qualityFlags      quality flags for                 ** see NOTES **
                     position and attitude
		     
   positionECI       ECI position               meters

   velocityECI       ECI velocity               meters/sec

   eulerAngles       s/c attitude as a set of   radians
                     Euler angles

   xyzRotRates       angular rates about body   radians/sec
                     x, y and z axes

   attitQuat         spacecraft to ECI          N/A
                     rotation quaternion
          
RETURNS:
   PGS_S_SUCCESS               successful return 
   PGSEPH_W_BAD_EPHEM_VALUE    one or more values could not be determined
   PGSMEM_E_NO_MEMORY          unable to allocate required dynamic memory space
   PGSEPH_E_BAD_EPHEM_FILE_HDR no s/c ephemeris and/or attitude files had
                               readable headers
   PGSEPH_E_NO_SC_EPHEM_FILE   no s/c ephemeris and/or attitude files could be
                               found for input times
   PGSEPH_E_NO_DATA_REQUESTED  both orb and att flags are set to false
   PGSTD_E_SC_TAG_UNKNOWN      unrecognized/unsupported spacecraft tag
   PGSEPH_E_BAD_ARRAY_SIZE     array size specified is less than 0
   PGSTD_E_TIME_FMT_ERROR      format error in asciiUTC
   PGSTD_E_TIME_VALUE_ERROR    value error in asciiUTC
   PGSTD_E_NO_LEAP_SECS        no leap seconds correction available for initial
                               time (asciiUTC)
   PGS_E_TOOLKIT               an unexpected error occured

EXAMPLES:
C:
        #define ARRAY_SIZE 10

	PGSt_double     offsets[ARRAY_SIZE];
	PGSt_double     positionECI[ARRAY_SIZE][3];
	PGSt_double     velocityECI[ARRAY_SIZE][3];
	PGSt_double     eulerAngles[ARRAY_SIZE][3];
	PGSt_double     xyzRotRates[ARRAY_SIZE][3];
	PGSt_double     attitQuat[ARRAY_SIZE][4];

	char            asciiUTC[28];

	PGSt_integer    qualityFlags[ARRAY_SIZE][2];

	int             i;
	
	PGSt_SMF_status returnStatus;


	** intialize asciiUTC and offsets array **

	strcpy(asciiUTC,"1998-02-03T19:23:45.123");
	for (i=0;i<ARRAY_SIZE;i++)
	  offsets[i] = (PGSt_double) i;

        returnStatus = PGS_EPH_EphemAttit(PGSd_EOS_AM,ARRAY_SIZE,asciiUTC,
	                                  offsets,PGS_TRUE,PGS_TRUE,
					  qualityFlags,positionECI,velocityECI,
					  eulerAngles,xyzRotRates,attitQuat);

	if (returnStatus != PGS_S_SUCCESS)
	{
	            :
	 ** do some error handling ***
		    :
	}
   
FORTRAN:
      implicit none

      include  'PGS_SMF.f'
      include  'PGS_TD.f'
      include  'PGS_TD_3.f'
      include  'PGS_EPH_5.f'
      include  'PGS_MEM_7.f'

      integer           pgs_eph_ephemattit

      integer           numvalues/10/
      integer           i
      integer           returnstatus
      integer           qualityflags(2,numvalues)
      
      character*27      asciiutc
      
      double precision  offsets(numvalues)
      double precision  positioneci(3,numvalues)
      double precision  velocityeci(3,numvalues)
      double precision  eulerangles(3,numvalues)
      double precision  xyzrotrates(3,numvalues)
      double precision  attitquat(4,numvalues)

!  intialize asciiutc and offsets array

      asciiutc = '1998-02-03T19:23:45.123'
      do 100 i = 1,numvalues
 100    offsets(i) = i-1

      returnstatus =  pgs_eph_ephemattit(pgsd_eos_am,numvalues,asciiutc,
     >                                   offsets,pgs_true,pgs_true,
     >                                   qualityflags,positioneci,
     >          			 velocityeci,eulerangles,
     >                                   xyzrotrates,attitquat)

      if (returnstatus .ne. pgs_s_success) then
                :
      *** do some error handling ***
                :
      endif

NOTES:
   QUALITY FLAGS:

     The quality flags are returned as integer quantities but should be
     interpreted as bit fields.  Only the first 32 bits of each quality flag is
     meaningfully defined, any additional bits should be ignored (currently
     integer quantities are 32 bits on most UNIX platforms, but this is not
     guaranteed to be the case--e.g. an integer is 64 bits on a Cray).

     Generally the quality flags are platform specific and are not defined by
     the Toolkit.  Two bits of these flags have, however, been reserved for SDP
     Toolkit usage.  Bit 12 will be set by the Toolkit if no data is available
     at a requested time, bit 14 will be set by the Toolkit if the data at the
     requested time has been interpolated (the least significant bit is "bit
     0").  Any other bits are platform specific and are the responsibility of
     the user to interpret.

   TIME ACRONYMS:
     
     TAI is:  International Atomic Time
     UTC is:  Coordinated Universal Time

   TIME BOUNDARIES:

     The minimum and maximum times that can successfully be processed by this
     function depend on the file leapsec.dat which relates leap second (TAI-UTC)
     values to UTC Julian dates.  The file leapsec.dat starts at Jan 1, 1961;
     therefore an error status message will be returned if this function is
     called with a time before that date.  The file, which is updated when a 
     new leap second event is announced, contains dates of leap second events.
     If an input date is outside of the range of dates in the file (or if the 
     file cannot be read) the function PGS_TD_LeapSec() which reads the file
     will return a calculated value of TAI-UTC based on a linear fit of the data
     in the table as of late 1997.  This value of TAI-UTC is relatively crude 
     estimate and may be off by as much as 1.0 or more seconds. The warning 
     return status from PGS_TD_LeapSec() is promoted to an error level by any 
     Toolkit function receiving it, which will terminate the present function.

   ASCII UTC:

     Toolkit functions that accept an ASCII time as input require the time to
     be UTC time in CCSDS ASCII Time Code A or CCSDS ASCII Time Code B format.
     Toolkit functions that return an ASCII time return the UTC time in CCSDS
     ASCII Time Code A format (see CCSDS 301.0-B-2 for details).
     (CCSDS => Consultative Committee for Space Data Systems)

      The general format is:
 
          YYYY-MM-DDThh:mm:ss.ddddddZ (Time Code A)
          YYYY-DDDThh:mm:ss.ddddddZ   (Time Code B)
 
          where:
              -,T,: are field delimiters
              Z is the (optional) time code terminator
              other fields are numbers as implied:
                Time Code A:
                   4 Year digits, 2 Month, 2 Day, 2 hour,
                   2 minute, 2 second, and up to 6 decimal
                   digits representing a fractional second
                Time Code B:
                   4 Year digits, 3 Day of year, 2 hour,
                   2 minute, 2 second, and up to 6 decimal
                   digits representing a fractional second

   TOOLKIT INTERNAL TIME (TAI):

     Toolkit internal time is the real number of continuous SI seconds since the
     epoch of UTC 12 AM 1-1-1993.  Toolkit internal time is also referred to in
     the toolkit as TAI (upon which it is based).

   TIME OFFSETS:

      This function accepts an ASCII UTC time, an array of time offsets and the
      number of offsets as input.  Each element in the offset array is an offset
      in seconds relative to the initial input ASCII UTC time.

      An error will be returned if the number of offsets specified is less than
      zero.  If the number of offsets specified is actually zero, the offsets
      array will be ignored.  In this case the input ASCII UTC time will be
      converted to Toolkit internal time (TAI) and this time will be used to
      process the data.  If the number of offsets specified is one (1) or
      greater, the input ASCII UTC time will converted to TAI and each element
      'i' of the input data will be processed at the time: (initial time) +
      (offset[i]).

      Examples:
         
         if numValues is 0 and asciiUTC is "1993-001T12:00:00" (TAI: 432000.0),
         then input[0] will be processed at time 432000.0 and return output[0]

	 if numValues is 1 and asciiUTC is "1993-001T12:00:00" (TAI: 432000.0),
         then input[0] will be processed at time 432000.0 + offsets[0] and
         return output[0]

         if numValues is N and asciiUTC is "1993-001T12:00:00" (TAI: 432000.0),
         then each input[i] will be processed at time 43000.0 + offsets[i] and
         the result will be output[i], where i is on the interval [0,N)

   ERROR HANDLING:

      This function processes data over an array of times (specified by an input
      ASCII UTC time and an array of time offsets relative to that time). 

      If processing at each input time is successful the return status of this
      function will be PGS_S_SUCCESS (status level of 'S').

      If processing at ALL input times was unsuccessful the status level of the
      return status of this function will be 'E'.

      If processing at some (but not all) input times was unsuccessful the
      status level (see SMF) of the return status of this function will be 'W'
      AND all high precision real number (C: PGSt_double, FORTRAN: DOUBLE
      PRECISION) output variables that correspond to the times for which
      processing was NOT successful will be set to the value:
      PGSd_GEO_ERROR_VALUE.  In this case users may (should) loop through the
      output testing any one of the aforementioned output variables against
      the value PGSd_GEO_ERROR_VALUE.  This indicates that there was an error in
      processing at the corresponding input time and no useful output data was
      produced for that time.  

      Note: a return status with a status of level of 'W' does not necessarily
      mean that some of the data could not be processed.  The 'W' level may
      indicate a general condition that the user may need to be aware of but
      that did not prohibit processing.  For example, if an Earth ellipsoid
      model is required, but the user supplied value is undefined, the WGS84
      model will be used, and processing will continue normally, except that the
      return status will be have a status level of 'W' to alert the user that
      the default earth model was used and not the one specified by the user.
      The reporting of such general warnings takes precedence over the generic
      warning (see RETURNS above) that processing was not successful at some of
      the requested times.  Therefore in the case of any return status of level
      'W', the returned value of a high precision real variable generally should
      be examined for errors at each time offset, as specified above.

      Special Note: for this tool, the associated quality flags will also
      indicate that no data is available for those points that could not be
      successfully processed (see QUALITY FLAGS above).

REQUIREMENTS:
   PGSTK - 0141, 0720, 0740

DETAILS:
   None

GLOBALS:
   None

FILES:
   The function requires the presence of the files:
      leapsec.dat
      earthfigure.dat
      utcpole.dat
   This function also requires ephemeris and attitude data files.

FUNCTIONS_CALLED:
   PGS_MEM_Calloc()              dynamically allocate (and initialize) memory
   PGS_TD_UTCtoTAI()             returns continuous seconds since 12AM 1/1/1993
   PGS_TD_sortArrayIndices()     sort offsets into ascending order
   PGS_EPH_getEphemRecords()     returns an array of s/c ephemeris records
   PGS_EPH_interpolatePosVel()   interpolates position and velocity
   PGS_EPH_interpolateAttitude() interpolates attitude Euler angles and rates
   PGS_CSC_getECItoORBquat()     returns the ECI to ORB quaternion
   PGS_CSC_EulerToQuat()         converts Euler angles to a quaternion
   PGS_CSC_quatMultiply()        multiplies two quaternions, returns result
   PGS_SMF_SetStaticMsg()        sets the message buffer
   PGS_SMF_SetDynamicMsg()       sets the message buffer
   PGS_SMF_SetUnknownMsg()       sets the message buffer

END_PROLOG:
*******************************************************************************/

#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <sys/types.h>
#include <PGS_CSC.h>
#include <PGS_EPH.h>
#include <PGS_MEM.h>

#define  BLOCK_SIZE            6000    /* size of ephemDataBlock array */

#define  MAX_EPH_TIME_INTERVAL 60.0   /* maximum time interval (in seconds)
					  over which ephemeris interpolation
					  will be done */

#define  MAX_ATT_TIME_INTERVAL 60.0   /* maximum time interval (in seconds)
					  over which attitude interpolation
					  will be done */
typedef struct
{
    PGSt_boolean    orbFlag;               /* get/don't get orbit data */
    PGSt_boolean    attFlag;               /* get/don't get attitude data */
    PGSt_integer    (*qFlags)[2];          /* quality flags */
    PGSt_double     (*positionECI)[3];     /* ECI position */
    PGSt_double     (*velocityECI)[3];     /* ECI velocity */
    PGSt_double     (*eulerAngle)[3];      /* s/c attitude Euler angles */
    PGSt_double     (*angularVelocity)[3]; /* ang. rates about s/c body axes */
    PGSt_double     (*attitQuat)[4];       /* s/c to ECI rotation quaternion */
} PGSt_scData;

/* prototype for function used to set bad s/c data to PGSd_GEO_ERROR_VALUE */

static void
PGS_EPH_setBadEphemRecord(PGSt_integer,PGSt_scData*,PGSt_SMF_status*);

/* name of this function */

#define  FUNCTION_NAME "PGS_EPH_EphemAttit()"

PGSt_SMF_status
PGS_EPH_EphemAttit(                      /* get ephemeris and attitude data */
    PGSt_tag        spacecraftTag,       /* spacecraft identifier */
    PGSt_integer    numValues,           /* number of values requested */
    char            *asciiUTC,           /* reference start time */
    PGSt_double     offsets[],           /* time offsets relative to asciiUTC */
    PGSt_boolean    orbFlag,             /* get/don't get orbit data */
    PGSt_boolean    attFlag,             /* get/don't get attitude data */
    void            *qualityFlagsPtr,    /* quality flags */
    PGSt_double     positionECI[][3],    /* ECI position */
    PGSt_double     velocityECI[][3],    /* ECI velocity */
    PGSt_double     eulerAngle[][3],     /* s/c attitude Euler angles */
    PGSt_double     angularVelocity[][3],/* angular rates about s/c body axes */
    PGSt_double     attitQuat[][4])      /* s/c to ECI rotation quaternion */
{
    static PGSt_ephemRecord ephemRecordArray[BLOCK_SIZE]; /* s/c ephemeris file 
							     records */
    static PGSt_attitRecord attitRecordArray[BLOCK_SIZE]; /* s/c ephemeris file 
							     records */
    static PGSt_integer     totalEphemRecords=0;          /* number of records
							     in scRecordArray */
    static PGSt_integer     totalAttitRecords=0;          /* number of records
							     in scRecordArray */
    static PGSt_double      beginTAI93eph=1.0e50;         /* first TAI time in
							     ephemRecordArray */
    static PGSt_double      endTAI93eph=-1.0e50;          /* last TAI time in
							     ephemRecordArray */
    static PGSt_double      beginTAI93att=1.0e50;         /* first TAI time in
							     attitRecordArray */
    static PGSt_double      endTAI93att=-1.0e50;          /* last TAI time in
							     attitRecordArray */
    static PGSt_tag         spacecraft=0;                 /* spacecraftTag value
							     from last call to
							     this function */

    PGSt_ephemRecord ephemRecord1;    /* s/c ephemeris file record */
    PGSt_ephemRecord ephemRecord2;    /* s/c ephemeris file record */

    PGSt_attitRecord attitRecord1;    /* s/c ephemeris file record */
    PGSt_attitRecord attitRecord2;    /* s/c ephemeris file record */

    PGSt_scData     scData;           /* structure of pointers to output
					 arrays */
    PGSt_boolean    gotData;          /* indicates outcome of data search */

    PGSt_integer    *sortedIndex=NULL;/* pointer to array holding sorted indices
					 of time offset array (offsets) */
    PGSt_integer    maxValues;        /* number of records requested */
    PGSt_integer    cnt;              /* loop counter */
    PGSt_integer    count;            /* loop counter */
    PGSt_integer    ephRecordCounter=0; /* loop counter */
    PGSt_integer    attRecordCounter=0; /* loop counter */
    PGSt_integer    (*qualityFlags)[2]; /* pointer to quality flag data */
    PGSt_integer    numBadPoints=0;   /* count of input times for which no
					 ephem/attit data could be determined */

    PGSt_double     secTAI93;         /* time of requested s/c data */
    PGSt_double     startTAI93;       /* TAI equivalent of asciiUTC */
    PGSt_double     tempPosVel[6];    /* s/c position and velocity at requested
					 time */
    PGSt_double     *tempPosition;    /* s/c position at requested time */
    PGSt_double     *tempVelocity;    /* s/c velocity at requested time */
    PGSt_double     quatORBtoECI[4];  /* orbital ref. frame to ECI quaternion */
    PGSt_double     quatEuler[4];     /* quaternion equivalent of s/c attitude
					 as defined by the s/c Euler angles */

    PGSt_scTagInfo  scTagInfo;        /* s/c tag info. structure */

    char mnemonic[PGS_SMF_MAX_MNEMONIC_SIZE]; /* error mnemonic */
    char msg[PGS_SMF_MAX_MSG_SIZE];           /* error message */
    char details[PGS_SMF_MAX_MSG_SIZE];       /* detailed error message */

    PGSt_SMF_status returnStatus1;    /* return value of PGS function calls */
    PGSt_SMF_status code;             /* returned from PGS_SMF_GetMsg() */
    PGSt_SMF_status returnStatus;     /* return value of this function */

    /* initialize return value to indicate success */

    returnStatus = PGS_S_SUCCESS;
    returnStatus1 = PGS_S_SUCCESS;

    /* initialize position and velocity pointers to point to the appropriate
       places is the array tmpPosVel */

    tempPosition = tempPosVel;
    tempVelocity = tempPosVel+3;

    /* The incoming pointer to the quality flags array must be cast to type:
       pointer to array of two integers.  This is what is expected.  A void type
       is being used for now to handle some backward compatibility issues,
       namely that this slot used to be for an array of character strings. */

    qualityFlags = (PGSt_integer (*)[2]) qualityFlagsPtr;

    /* check that at least SOME data have been requested */

    if (orbFlag != PGS_TRUE && attFlag != PGS_TRUE)
    {
	PGS_SMF_SetStaticMsg(PGSEPH_E_NO_DATA_REQUESTED,FUNCTION_NAME);
	return PGSEPH_E_NO_DATA_REQUESTED;
    }

    /* check spacecraft tag for validity */

    returnStatus = PGS_EPH_GetSpacecraftData(spacecraftTag, NULL,
					     PGSe_TAG_SEARCH, &scTagInfo);
    switch (returnStatus)
    {
      case PGS_S_SUCCESS:
	break;

      case PGSTD_E_SC_TAG_UNKNOWN:
      case PGS_E_TOOLKIT:
	return returnStatus;

      default:
	PGS_SMF_SetUnknownMsg(returnStatus, FUNCTION_NAME);
	return PGS_E_TOOLKIT;
    }

    /* check numValues to make sure it is not negative */

    if (numValues < 0)
    {
	PGS_SMF_SetStaticMsg(PGSEPH_E_BAD_ARRAY_SIZE,FUNCTION_NAME);
	return PGSEPH_E_BAD_ARRAY_SIZE;
    }

    /* The number of records requested is the number specified by numValues
       passed in the calling sequence unless this number is 0 in which case one
       record is still retrieved.  In the latter case this will be the record
       corresponding to the ASCII time input (asciiUTC). */

    maxValues = (numValues) ? numValues : 1;

    /* In order to expedite processing, it is desirable that the s/c data be
       searched for in monotonically increasing temporal order.  Therefore an
       array of PGSt_integers is allocated to hold the sorted indices of the
       input time offsets (offsets).  The array "offsets" is sorted into
       chronological order by storing the indices in "sortedIndex" in such a way
       that accessing the data by these indices accesses the data in
       chronological order.  The actual input array "offsets" is in no way
       affected. */

    returnStatus = PGS_MEM_Calloc((void **) &sortedIndex, 1,
				  sizeof(PGSt_integer)*maxValues);
    if (returnStatus != PGS_S_SUCCESS)
	return returnStatus;

    PGS_TD_sortArrayIndices(offsets,maxValues,sortedIndex);

    /* for simplified error handling the following structure is used to sort of
       keep a copy of the s/c data being returned by this function */

    scData.orbFlag = orbFlag;
    scData.attFlag = attFlag;
    scData.qFlags = qualityFlags;
    scData.positionECI = positionECI;
    scData.velocityECI = velocityECI;
    scData.eulerAngle = eulerAngle;
    scData.angularVelocity = angularVelocity;
    scData.attitQuat = attitQuat;

    /* convert ASCII UTC start time to real continuous seconds since
       12AM UTC 1/1/93, this is done so that the time offsets which are in
       seconds can be added to this value to get a unique numeric time for each 
       point */

    returnStatus = PGS_TD_UTCtoTAI(asciiUTC, &startTAI93);
    switch (returnStatus)
    {
      case PGS_S_SUCCESS:
      case PGSTD_E_NO_LEAP_SECS:
	PGS_SMF_GetMsg(&code,mnemonic,details);
	if (code != returnStatus)
	  PGS_SMF_GetMsgByCode(returnStatus,details);
	break;
      case PGSTD_E_TIME_FMT_ERROR:
      case PGSTD_E_TIME_VALUE_ERROR:
      case PGS_E_TOOLKIT:
	PGS_MEM_Free(sortedIndex);
	return returnStatus;
      default:
	PGS_SMF_SetUnknownMsg(returnStatus,FUNCTION_NAME);
	PGS_MEM_Free(sortedIndex);
	return PGS_E_TOOLKIT;
    }

    /* the first time requested is the reference time plus the first offset,
       unless numValues is 0 in which case the first time is just the reference
       time itself--here the time is initialized to this reference time */

    if (numValues)
      secTAI93 = startTAI93 + offsets[sortedIndex[0]];
    else
      secTAI93 = startTAI93;

    /* Step through the requested times until a s/c record for one of the
       requested time is retrieved.  Note that "a s/c record" is really two
       records if a requested time is not coincident with a time recorded in the
       s/c ephemeris file.  This is of course necessary in order to interpolate
       the data to the requested time.  If no s/c record can be found for any of
       the requested times exit with an error message. */

    cnt = 0;
    gotData = PGS_FALSE;

    if (secTAI93 < beginTAI93eph ||
	secTAI93 > endTAI93eph ||
	spacecraftTag != spacecraft)
    {
	do
	{
	    if (numValues)
	    {
		secTAI93 = startTAI93 + offsets[sortedIndex[cnt]];
	    }
	  
	    /* Fill the array ephemRecordArray with s/c ephemeris records. */

	    returnStatus1 = PGS_EPH_getEphemRecords(&scTagInfo,secTAI93,
						    BLOCK_SIZE,ephemRecordArray,
						    &totalEphemRecords);
	    switch (returnStatus1)
	    {
	      case PGS_S_SUCCESS:
	      case PGSEPH_M_SHORT_ARRAY:
		gotData = PGS_TRUE;
		beginTAI93eph = ephemRecordArray[0].secTAI93;
		endTAI93eph = ephemRecordArray[totalEphemRecords-1].secTAI93;
		break;
	      case PGSEPH_E_BAD_EPHEM_FILE_HDR:
	      case PGSEPH_E_NO_SC_EPHEM_FILE:
	      case PGS_E_TOOLKIT:
		PGS_SMF_GetMsg(&code,mnemonic,msg);
		if (code != returnStatus1)
		  PGS_SMF_GetMsgByCode(returnStatus1,msg);
		if (returnStatus == PGS_S_SUCCESS)
		  strcpy(details,msg);
		numBadPoints++;
		PGS_EPH_setBadEphemRecord(sortedIndex[cnt],&scData,
					  &returnStatus);
		break;
	      default:
		PGS_SMF_SetUnknownMsg(returnStatus1,FUNCTION_NAME);
		returnStatus1 = PGS_E_TOOLKIT;
		numBadPoints++;
		PGS_EPH_setBadEphemRecord(sortedIndex[cnt],&scData,
					  &returnStatus);
		break;
	    }
	}
	while (++cnt < maxValues && gotData == PGS_FALSE);
    }
    else /* END: if (secTAI93 < startTAI93eph ... */
    {
	gotData = PGS_TRUE;
	cnt++;
    }

    /* if gotData is still false here, all times were checked and no s/c
       ephemeris file was successfully opened, so exit with error message */

    if (gotData == PGS_FALSE)
    {
	PGS_MEM_Free(sortedIndex);
	PGS_SMF_SetDynamicMsg(returnStatus1,msg,FUNCTION_NAME);
	return returnStatus1;
    }
	    
    if (attFlag == PGS_TRUE)
    {
	count = cnt;
	cnt = 0;
	gotData = PGS_FALSE;
	
	if (secTAI93 < beginTAI93att ||
	    secTAI93 > endTAI93att ||
	    spacecraftTag != spacecraft)
	{
	    do
	    {
		if (numValues)
		{
		    secTAI93 = startTAI93 + offsets[sortedIndex[cnt]];
		}
	  
		/* Fill the array attitRecordArray with s/c attitude records. */

		returnStatus1 = PGS_EPH_getAttitRecords(&scTagInfo,
							secTAI93,
							BLOCK_SIZE,
							attitRecordArray,
							&totalAttitRecords);
		switch (returnStatus1)
		{
		  case PGS_S_SUCCESS:
		  case PGSEPH_M_SHORT_ARRAY:
		    gotData = PGS_TRUE;
		    beginTAI93att = attitRecordArray[0].secTAI93;
		    endTAI93att =attitRecordArray[totalAttitRecords-1].secTAI93;
		    break;
		  case PGSEPH_E_BAD_EPHEM_FILE_HDR:
		  case PGSEPH_E_NO_SC_EPHEM_FILE:
		  case PGS_E_TOOLKIT:
		    gotData = PGS_FALSE;
		    PGS_SMF_GetMsg(&code,mnemonic,msg);
		    if (code != returnStatus1)
		      PGS_SMF_GetMsgByCode(returnStatus1,msg);
		    if (returnStatus == PGS_S_SUCCESS)
		      strcpy(details,msg);
		    numBadPoints++;
		    PGS_EPH_setBadEphemRecord(sortedIndex[cnt],&scData,
					      &returnStatus);
		    break;
		  default:
		    gotData = PGS_FALSE;
		    PGS_SMF_SetUnknownMsg(returnStatus1,FUNCTION_NAME);
		    returnStatus1 = PGS_E_TOOLKIT;
		    numBadPoints++;
		    PGS_EPH_setBadEphemRecord(sortedIndex[cnt],&scData,
					      &returnStatus);
		    break;
		}
	    }
	    while (++cnt < maxValues && gotData == PGS_FALSE);
	}
	else /* END: if (secTAI93 < startTAI93att ... */
	{
	    gotData = PGS_TRUE;
	    cnt++;
	}
	
	/* If "count" is greater than "cnt" then assign the value of "count" to
	   "cnt".  "cnt" is used to determine which input times have
	   corresponding s/c data.  This should therefore be set to the greater
	   of "count" (first time index for which valid ephemeris data was
	   found) and "cnt" (first time index for which valid attitude data was
	   found). */

	cnt = (count > cnt) ? count : cnt;
    }

    /* if gotData is still false here, all times were checked and no s/c
       (necessary) attitude file was successfully opened, so exit with error
       message */

    if (gotData == PGS_FALSE)
    {
	PGS_MEM_Free(sortedIndex);
	PGS_SMF_SetDynamicMsg(returnStatus1,msg,FUNCTION_NAME);
	return returnStatus1;
    }
    else
    {
	/* data is available for the current spacecraft, save the value of the
	   spacecraft tag so the data isn't reloaded unnecessarily in subsequent
	   calls to this function */

	spacecraft = spacecraftTag;
    }
    
    gotData = PGS_FALSE;
    for (ephRecordCounter=0;
	 ephRecordCounter<totalEphemRecords;
	 ephRecordCounter++)
    {
	if (secTAI93 < ephemRecordArray[ephRecordCounter].secTAI93)
	{
	    memcpy((char *) &ephemRecord2,
		   (char *)&ephemRecordArray[ephRecordCounter],
		   sizeof(PGSt_ephemRecord));

	    memcpy((char *) &ephemRecord1,
		   (char *)&ephemRecordArray[ephRecordCounter-1],
		   sizeof(PGSt_ephemRecord));

	    gotData = PGS_TRUE;
	    break;
	}
    }

    if (gotData == PGS_FALSE)
    {
	memcpy((char *) &ephemRecord2,
	       (char *) &ephemRecordArray[ephRecordCounter-1],
	       sizeof(PGSt_ephemRecord));

	memcpy((char *) &ephemRecord1,
	       (char *) &ephemRecordArray[ephRecordCounter-1],
	       sizeof(PGSt_ephemRecord));

	gotData = PGS_TRUE;
    }	    

    if (attFlag == PGS_TRUE)
    {
	gotData = PGS_FALSE;
	for (attRecordCounter=0;
	     attRecordCounter<totalAttitRecords;
	     attRecordCounter++)
	{
	    if (secTAI93 < attitRecordArray[attRecordCounter].secTAI93)
	    {
		memcpy((char *) &attitRecord2,
		       (char *) &attitRecordArray[attRecordCounter],
		       sizeof(PGSt_attitRecord));
		memcpy((char *) &attitRecord1,
		       (char *) &attitRecordArray[attRecordCounter-1],
		       sizeof(PGSt_attitRecord));
		gotData = PGS_TRUE;
		break;
	}
	}
	
	if (gotData == PGS_FALSE)
	{
	    memcpy((char *) &attitRecord2,
		   (char *)&attitRecordArray[attRecordCounter-1],
		   sizeof(PGSt_attitRecord));

	    memcpy((char *) &attitRecord1,
		   (char *)&attitRecordArray[attRecordCounter-1],
		   sizeof(PGSt_attitRecord));

	    gotData = PGS_TRUE;
	}
    }
    

    /* Step through the requested times retrieving s/c records.  Note that
       "a s/c record" is really two records if a requested time is not
       coincident with a time recorded in the s/c ephemeris file.  This is of
       course necessary in order to interpolate the data to the requested time.
       If no s/c record can be found for a requested time fill the record at
       that time with an "error value". */

    for (cnt--;cnt<maxValues;cnt++)
    {
	/* The current time is the reference time plus the current offset,
	   unless numValues is 0 in which case the current time is just the
	   reference time itself--to which it has been initialize above. */

	if (numValues)
	{
	    secTAI93 = startTAI93 + offsets[sortedIndex[cnt]];
	}

	/* If the current time falls within bounds of the most recently obtained
	   record go to the interpolation code... */

	if (!(secTAI93 >= ephemRecord1.secTAI93 && 
	      secTAI93 < ephemRecord2.secTAI93))
	{
	    /* ...otherwise get the next record. */

	    /* Search ephemRecordArray for the next record (starting at the last
	       place (time) a record was found in the array). */

	    gotData = PGS_FALSE;
	    for (;ephRecordCounter<totalEphemRecords;ephRecordCounter++)
	      if (secTAI93 < ephemRecordArray[ephRecordCounter].secTAI93)
	      {
		  memcpy((char*)&ephemRecord2,
			 (char*)&ephemRecordArray[ephRecordCounter],
			 sizeof(PGSt_ephemRecord));

		  memcpy((char*)&ephemRecord1,
			 (char*)&ephemRecordArray[ephRecordCounter-1],
			 sizeof(PGSt_ephemRecord));

		  gotData = PGS_TRUE;
		  break;
	      }	    

	    /* If no record was found in ephemRecordArray in the above search,
	       check the last record in the array (the above will miss the last
	       record) to see if the requested time exactly corresponds to the
	       last time in the array. */

	    if (gotData == PGS_FALSE)
	      if (secTAI93 == ephemRecordArray[totalEphemRecords-1].secTAI93)
	      {
		  memcpy((char*)&ephemRecord2,
			 (char*)&ephemRecordArray[ephRecordCounter-1],
			 sizeof(PGSt_ephemRecord));
		  memcpy((char*)&ephemRecord1,
			 (char*)&ephemRecordArray[ephRecordCounter-1],
			 sizeof(PGSt_ephemRecord));
		  gotData = PGS_TRUE;
	      }

	    /* If the appropriate record STILL has not been found in
	       ephemRecordArray, call PGS_EPH_getEphemRecords to read the next
	       block of spacecraft ephemeris data from disk. */

	    if (gotData == PGS_FALSE)
	    {
		returnStatus1 = PGS_EPH_getEphemRecords(&scTagInfo, secTAI93,
							BLOCK_SIZE,
							ephemRecordArray,
							&totalEphemRecords);
		switch (returnStatus1)
		{
		  case PGS_S_SUCCESS:
		  case PGSEPH_M_SHORT_ARRAY:
		    beginTAI93eph = ephemRecordArray[0].secTAI93;
		    endTAI93eph =ephemRecordArray[totalEphemRecords-1].secTAI93;
		    for (ephRecordCounter=1;
			 ephRecordCounter<totalEphemRecords;
			 ephRecordCounter++)
		    {
			if (secTAI93 < 
			    ephemRecordArray[ephRecordCounter].secTAI93)
			{
			    memcpy((char*)&ephemRecord2,
				   (char*)&ephemRecordArray[ephRecordCounter],
				   sizeof(PGSt_ephemRecord));

			    memcpy((char*)&ephemRecord1,
				   (char*)&ephemRecordArray[ephRecordCounter-1],
				   sizeof(PGSt_ephemRecord));

			    gotData = PGS_TRUE;
			    break;
			}
		    }
		    
		    if (gotData == PGS_TRUE)
			break;
		    if (secTAI93 == 
			ephemRecordArray[totalEphemRecords-1].secTAI93)
		    {
			memcpy((char *) &ephemRecord2,
			       (char *) &ephemRecordArray[ephRecordCounter-1],
			       sizeof(PGSt_ephemRecord));

			memcpy((char *) &ephemRecord1,
			       (char *) &ephemRecordArray[ephRecordCounter-1],
			       sizeof(PGSt_ephemRecord));

			gotData = PGS_TRUE;
			break;
		    }
		    returnStatus1 = PGS_E_TOOLKIT;
		  case PGSEPH_E_BAD_EPHEM_FILE_HDR:
		  case PGSEPH_E_NO_SC_EPHEM_FILE:
		  case PGS_E_TOOLKIT:
		    PGS_SMF_GetMsg(&code,mnemonic,msg);
		    if (code != returnStatus1)
		      PGS_SMF_GetMsgByCode(returnStatus1,msg);
		    if (returnStatus == PGS_S_SUCCESS)
		      strcpy(details,msg);
		    numBadPoints++;
		    PGS_EPH_setBadEphemRecord(sortedIndex[cnt],&scData,
					      &returnStatus);
		    continue;
		  default:
		    PGS_SMF_SetUnknownMsg(returnStatus1,
					  FUNCTION_NAME);
		    returnStatus1 = PGS_E_TOOLKIT;
		    numBadPoints++;
		    PGS_EPH_setBadEphemRecord(sortedIndex[cnt],&scData,
					      &returnStatus);
		    continue;
		}
	    }
	}

	/* If the current time falls within bounds of the most recently obtained
	   record go to the interpolation code... */

	if ( attFlag == PGS_TRUE &&
	     (! (secTAI93 >= attitRecord1.secTAI93 && 
		secTAI93 < attitRecord2.secTAI93)))
	{
	    /* ...otherwise get the next record. */

	    /* Search attitRecordArray for the next record (starting at the last
	       place (time) a record was found in the array). */

	    gotData = PGS_FALSE;
	    for (;attRecordCounter<totalAttitRecords;attRecordCounter++)
	      if (secTAI93 < attitRecordArray[attRecordCounter].secTAI93)
	      {
		  memcpy((char*)&attitRecord2,
			 (char*)&attitRecordArray[attRecordCounter],
			 sizeof(PGSt_attitRecord));

		  memcpy((char*)&attitRecord1,
			 (char*)&attitRecordArray[attRecordCounter-1],
			 sizeof(PGSt_attitRecord));

		  gotData = PGS_TRUE;
		  break;
	      }	    

	    /* If no record was found in attitRecordArray in the above search,
	       check the last record in the array (the above will miss the last
	       record) to see if the requested time exactly corresponds to the
	       last time in the array. */

	    if (gotData == PGS_FALSE)
	      if (secTAI93 == attitRecordArray[totalAttitRecords-1].secTAI93)
	      {
		  memcpy((char*)&attitRecord2,
			 (char*)&attitRecordArray[attRecordCounter-1],
			 sizeof(PGSt_attitRecord));
		  memcpy((char*)&attitRecord1,
			 (char*)&attitRecordArray[attRecordCounter-1],
			 sizeof(PGSt_attitRecord));
		  gotData = PGS_TRUE;
	      }

	    /* If the appropriate record STILL has not been found in
	       attitRecordArray, call PGS_EPH_getEphemRecords to read the next
	       block of spacecraft ephemeris data from disk. */

	    if (gotData == PGS_FALSE)
	    {
		returnStatus1 = PGS_EPH_getAttitRecords(&scTagInfo,secTAI93,
							BLOCK_SIZE,
							attitRecordArray,
							&totalAttitRecords);
		switch (returnStatus1)
		{
		  case PGS_S_SUCCESS:
		  case PGSEPH_M_SHORT_ARRAY:
		    beginTAI93att = attitRecordArray[0].secTAI93;
		    endTAI93att =attitRecordArray[totalAttitRecords-1].secTAI93;
		    for (attRecordCounter=1;
			 attRecordCounter<totalAttitRecords;
			 attRecordCounter++)
		    {
			if (secTAI93 <
			    attitRecordArray[attRecordCounter].secTAI93)
			{
			    memcpy((char*)&attitRecord2,
				   (char*)&attitRecordArray[attRecordCounter],
				   sizeof(PGSt_attitRecord));
			    memcpy((char*)&attitRecord1,
				   (char*)&attitRecordArray[attRecordCounter-1],
				   sizeof(PGSt_attitRecord));
			    gotData = PGS_TRUE;
			    break;
			}
		    }
		    
		    if (gotData == PGS_TRUE)
			break;
		    if (secTAI93 == 
			attitRecordArray[totalAttitRecords-1].secTAI93)
		    {
			memcpy((char *) &attitRecord2,
			       (char *) &attitRecordArray[attRecordCounter-1],
			       sizeof(PGSt_attitRecord));

			memcpy((char *) &attitRecord1,
			       (char *) &attitRecordArray[attRecordCounter-1],
			       sizeof(PGSt_attitRecord));

			gotData = PGS_TRUE;
			break;
		    }
		    returnStatus1 = PGS_E_TOOLKIT;
		  case PGSEPH_E_BAD_EPHEM_FILE_HDR:
		  case PGSEPH_E_NO_SC_EPHEM_FILE:
		  case PGS_E_TOOLKIT:
		    PGS_SMF_GetMsg(&code,mnemonic,msg);
		    if (code != returnStatus1)
		      PGS_SMF_GetMsgByCode(returnStatus1,msg);
		    if (returnStatus == PGS_S_SUCCESS)
		      strcpy(details,msg);
		    numBadPoints++;
		    PGS_EPH_setBadEphemRecord(sortedIndex[cnt],&scData,
					      &returnStatus);
		    continue;
		  default:
		    PGS_SMF_SetUnknownMsg(returnStatus1,
					  FUNCTION_NAME);
		    returnStatus1 = PGS_E_TOOLKIT;
		    numBadPoints++;
		    PGS_EPH_setBadEphemRecord(sortedIndex[cnt],&scData,
					      &returnStatus);
		    continue;
		}
	    }
	}

	/* Get the current position and velocity--interpolating if necessary.
	   The pos. and vel. must be determined regardless of whether or not
	   the orbFlag has been set to PGS_TRUE since they are needed to
	   determine the attitQuat.  Therefore at this point the values are
	   being assigned to temporary variables. */

	if (secTAI93 == ephemRecord1.secTAI93)
	{
	    for (count=0;count<3;count++)
	    {
		tempPosition[count] = ephemRecord1.position[count];
		tempVelocity[count] = ephemRecord1.velocity[count];
		qualityFlags[sortedIndex[cnt]][0] = 
		  (PGSt_integer) ephemRecord1.qualityFlag;
	    }
	}
	else
	{
	    /* if the time interval between bounding ephemeris records is too
	       large, don't interpolate, consider this condition an error
	       (i.e. no data available at the requested time) */

	    if ((ephemRecord2.secTAI93 - ephemRecord1.secTAI93) >
		MAX_EPH_TIME_INTERVAL)
	    {
		if (returnStatus == PGS_S_SUCCESS)
		{
		    sprintf(details,"%s%s%.2f%s",
			    "interpolation error for some requested values, ",
			    "time between actual ephemeris records is too "
			    "large (> ", MAX_EPH_TIME_INTERVAL, " seconds).");
		}
		numBadPoints++;
		PGS_EPH_setBadEphemRecord(sortedIndex[cnt],&scData,
					  &returnStatus);
		continue;
	    }
	    
	    returnStatus1 = PGS_EPH_interpolatePosVel(ephemRecord1.secTAI93,
						      ephemRecord1.position,
						      ephemRecord1.velocity,
						      ephemRecord2.secTAI93,
						      ephemRecord2.position,
						      ephemRecord2.velocity,
						      secTAI93,
						      tempPosition,
						      tempVelocity);
	    if (returnStatus1 != PGS_S_SUCCESS)
	    {
		if (returnStatus == PGS_S_SUCCESS)
		{
		    PGS_SMF_SetUnknownMsg(returnStatus1,FUNCTION_NAME);
		    PGS_SMF_GetMsg(&code,mnemonic,details);
		}
		numBadPoints++;
		PGS_EPH_setBadEphemRecord(sortedIndex[cnt],&scData,
					  &returnStatus);
		continue;
	    }
	    qualityFlags[sortedIndex[cnt]][0]= (PGSt_integer)
	                                       (ephemRecord1.qualityFlag |
						ephemRecord2.qualityFlag |
						PGSd_INTERPOLATED_POINT);
	}

	/* If attitude data has been requested get the current Euler angles and
	   their respective rates--interpolating if necessary.  Then determine
	   the s/c to ECI quaternion. */

	if (attFlag == PGS_TRUE)
	{
	    if (secTAI93 == attitRecord1.secTAI93)
	    {
		for (count=0;count<3;count++)
		{
		    eulerAngle[sortedIndex[cnt]][count] = 
			attitRecord1.eulerAngle[count];
		    angularVelocity[sortedIndex[cnt]][count] = 
			attitRecord1.angularVelocity[count];
		}
		qualityFlags[sortedIndex[cnt]][1] =
		  (PGSt_integer) attitRecord1.qualityFlag;
	    }
	    else  /* interpolate Euler angles and angular rates */
	    {
		/* if the time interval between bounding attitude records is
		   too large, don't interpolate, consider this condition an
		   error (i.e. no data available at the requested time) */
		
		if ((attitRecord2.secTAI93 - attitRecord1.secTAI93) >
		    MAX_ATT_TIME_INTERVAL)
		{
		    if (returnStatus == PGS_S_SUCCESS)
		    {
			sprintf(details,"%s%.2f%s",
				"interpolation error for some requested values,"
				" time between actual attitude records is too "
				"large (> ", MAX_ATT_TIME_INTERVAL,
				" seconds).");
		    }
		    numBadPoints++;
		    PGS_EPH_setBadEphemRecord(sortedIndex[cnt],&scData,
					      &returnStatus);
		    continue;
		}
		
		returnStatus1 = 
		  PGS_EPH_interpolateAttitude(attitRecord1.secTAI93,
					      attitRecord1.eulerAngle,
					      attitRecord1.angularVelocity,
					      attitRecord2.secTAI93,
					      attitRecord2.eulerAngle,
					      attitRecord2.angularVelocity,
					      scTagInfo.eulerAngleOrder,
					      secTAI93,
					      eulerAngle[sortedIndex[cnt]],
					      angularVelocity[sortedIndex[cnt]]
					      );
		
		if (returnStatus1 != PGS_S_SUCCESS)
		{
		    if (returnStatus == PGS_S_SUCCESS)
		    {
			PGS_SMF_SetUnknownMsg(returnStatus1,
					      FUNCTION_NAME);
			PGS_SMF_GetMsg(&code,mnemonic,details);
		    }
		    numBadPoints++;
		    PGS_EPH_setBadEphemRecord(sortedIndex[cnt],&scData,
					      &returnStatus);
		    continue;
		}

		qualityFlags[sortedIndex[cnt]][1] = (PGSt_integer)
                                                    (attitRecord1.qualityFlag |
						     attitRecord2.qualityFlag |
						     PGSd_INTERPOLATED_POINT);
	    }

	    /* Get the quaternion that gives the equivalent rotation as the s/c
	       Euler angles.  The Euler angle order depends on the s/c.  The
	       quaternion quatEuler can be used to rotate a vector from the s/c
	       reference frame to the orbital reference frame. */

	    returnStatus1 = PGS_CSC_EulerToQuat(eulerAngle[sortedIndex[cnt]],
						scTagInfo.eulerAngleOrder,
						quatEuler);
	    if (returnStatus1 != PGS_S_SUCCESS)
	    {
		if (returnStatus == PGS_S_SUCCESS)
		{
		    PGS_SMF_SetUnknownMsg(returnStatus1,
					  FUNCTION_NAME);
		    PGS_SMF_GetMsg(&code,mnemonic,details);
		}
		numBadPoints++;
		PGS_EPH_setBadEphemRecord(sortedIndex[cnt],&scData,
					  &returnStatus);
		continue;
	    }

	    /* get the orbital to ECI reference frame quaternion */
	    
	    returnStatus1 = PGS_CSC_getORBtoECIquat(tempPosition,
						    tempVelocity,
						    quatORBtoECI);
	    if (returnStatus1 != PGS_S_SUCCESS)
	    {
		if (returnStatus == PGS_S_SUCCESS)
		{
		    PGS_SMF_SetUnknownMsg(returnStatus1, FUNCTION_NAME);
		    PGS_SMF_GetMsg(&code,mnemonic,details);
		}
		numBadPoints++;
		PGS_EPH_setBadEphemRecord(sortedIndex[cnt],&scData,
					  &returnStatus);
		continue;
	    }

	    returnStatus1 = PGS_CSC_quatMultiply(quatORBtoECI, quatEuler,
						 attitQuat[sortedIndex[cnt]]);

	    /* Check the return status from the last operation any of the cases
	       above. */

	    if (returnStatus1 != PGS_S_SUCCESS)
	    {
		if (returnStatus == PGS_S_SUCCESS)
		{
		    PGS_SMF_SetUnknownMsg(returnStatus1, FUNCTION_NAME);
		    PGS_SMF_GetMsg(&code, mnemonic, details);
		}
		numBadPoints++;
		PGS_EPH_setBadEphemRecord(sortedIndex[cnt], &scData,
					  &returnStatus);
		continue;
	    }
	} /* END: if (attitudeFlag == PGS_TRUE) */

	/* finally if the orbit flag is set to PGS_TRUE write the position and
	   velocity data into their respective arrays */

	if (orbFlag == PGS_TRUE)
	{
	    memcpy((char *) positionECI[sortedIndex[cnt]],(char *) tempPosition,
		   sizeof(PGSt_double)*3);
	    memcpy((char *) velocityECI[sortedIndex[cnt]],(char *) tempVelocity,
		   sizeof(PGSt_double)*3);
	}

    }
    PGS_MEM_Free(sortedIndex);
    
    if (numBadPoints >= maxValues)
    {
	returnStatus = PGSEPH_E_NO_SC_EPHEM_FILE;
    }

    PGS_SMF_SetDynamicMsg(returnStatus,details,FUNCTION_NAME);
    
    return returnStatus;
}

/* the following function sets the values of the elements of bad records
   (i.e. records that could not be determined) to PGSd_GEO_ERROR_VALUE */

static void
PGS_EPH_setBadEphemRecord(
    PGSt_integer    recordNum,
    PGSt_scData     *scData,
    PGSt_SMF_status *returnStatus)
{
    short           cntr;             /* loop counter */

    if (scData->orbFlag == PGS_TRUE)
    {	
	scData->qFlags[recordNum][0] = PGSd_NO_DATA | 0x1;
	for (cntr=0;cntr<3;cntr++)
	{
	    scData->positionECI[recordNum][cntr] = PGSd_GEO_ERROR_VALUE;
	    scData->velocityECI[recordNum][cntr] = PGSd_GEO_ERROR_VALUE;
	}
    }
    
    if (scData->attFlag == PGS_TRUE)
    {
	scData->qFlags[recordNum][1] = PGSd_NO_DATA | 0x1;	
	for (cntr=0;cntr<3;cntr++)
	{
	    scData->eulerAngle[recordNum][cntr] = PGSd_GEO_ERROR_VALUE;
	    scData->angularVelocity[recordNum][cntr] = PGSd_GEO_ERROR_VALUE;
	    scData->attitQuat[recordNum][cntr] = PGSd_GEO_ERROR_VALUE;
	}
	scData->attitQuat[recordNum][3] = PGSd_GEO_ERROR_VALUE;
    }
    if (*returnStatus == PGS_S_SUCCESS)
      *returnStatus = PGSEPH_W_BAD_EPHEM_VALUE;
    return;
}
