/*-------------------------------------------------------------------------*/
/*                                                                         */
/*  COPYRIGHT[copyright mark] 1996, Hughes Aircraft Company, its vendors,  */
/*  and suppliers.  ALL RIGHTS RESERVED.                                   */
/*                                                                         */
/*-------------------------------------------------------------------------*/
/*******************************************************************************
BEGIN_FILE_PROLOG:

FILENAME:
  PGS_EPH_GetEphMet.c

DESCRIPTION:
  This file contains the function PGS_EPH_GetEphMet()
   This function returns the metadata associated with toolkit spacecraft
   ephemeris files.

AUTHOR:
  Guru Tej S. Khalsa / Applied Research Corporation
  Peter D. Noerdlinger / Applied Research Corporation
  Abe Taaheri          / Space Applications Corporation

HISTORY:
  01-Jul-1996  GTSK  Initial version
  27-May-1997  PDN   removed references to predicted leap seconds
  28-Sep-1998  AT    Added PGS_EPH_GetEphMetAux function to wrap around
                     PGS_EPH_GetEphMet. The new function (which has one more
                     argument than PGS_EPH_GetEphMet) is used by fortran
                     binder to resolve out of bound write when the function is
                     called from a fortran code (NCR ECSed17947 )

END_FILE_PROLOG:
*******************************************************************************/

/*******************************************************************************
BEGIN_PROLOG:

TITLE:
   Get Spacecraft Ephemeris Metadata

NAME:
   PGS_EPH_GetEphMet()

SYNOPSIS:
C:
   #include <PGS_EPH.h>

    PGSt_SMF_status
    PGS_EPH_GetEphMet(
        PGSt_tag      spacecraftTag,
	PGSt_integer  numValues,
	char          asciiUTC[28],
        PGSt_double   offsets[],
        PGSt_integer* numOrbits,
        PGSt_integer  orbitNumber[],
        char          orbitAscendTime[][28],
        char          orbitDescendTime[][28],
        PGSt_double   orbitDownLongitude[])

FORTRAN:
      include 'PGS_EPH_5.f'
      include 'PGS_TD_3.f'
      include 'PGS_SMF.f'

      integer function pgs_eph_getephmet(spacecrafttag,numvalues,
     >                                   asciiutc,offsets,numorbits,
     >                                   orbitnumber,orbitascendtime,
     >                                   orbitdescendtime,
     >                                   orbitdownlongitude)

      integer          spacecrafttag
      integer          numvalues
      character*27     asciiutc
      double precision offsets(*)
      integer          numorbits
      integer          orbitnumber(*)
      character*27     orbitascendtime(*)
      character*27     orbitdescendtime(*)
      double precision orbitdownlongitude(*)

DESCRIPTION:
   This function returns the metadata associated with toolkit spacecraft
   ephemeris files.

INPUTS:
   Name              Description               Units   Min         Max
   ----              -----------               -----   ---         ---
   spacecraftTag     spacecraft identifier     N/A

   numValues         num. of values requested  N/A

   asciiUTC          UTC time reference start  ASCII   1961-01-01 see NOTES
                     time in CCSDS ASCII time
	             code A format

   offsets           array of time offsets in  sec     ** depends on asciiUTC **
                     seconds relative to
	             asciiUTC

OUTPUTS:
   Name               Description              Units   Min         Max
   ----               -----------              -----   ---         ---
   numOrbits          number of orbits         N/A
                      spanned by data set

   orbitNumber        array of orbit numbers   N/A
                      spanned by data set

   orbitAscendTime    array of times of        ASCII
                      spacecraft northward
                      equator crossings

   orbitDescendTime   array of times of        ASCII
                      spacecraft southward
                      equator crossings

   orbitDownLongitude array of longitudes      radians
                      or spacecraft
		      southward equator
		      crossings

RETURNS:
   PGS_S_SUCCESS               successful return
   PGSTD_E_SC_TAG_UNKNOWN      unknown/unsupported spacecraft tag
   PGSEPH_E_NO_SC_EPHEM_FILE   no s/c ephem files could be found for input times
   PGSEPH_E_BAD_ARRAY_SIZE     array size specified is less than 0
   PGSTD_E_TIME_FMT_ERROR      format error in asciiUTC
   PGSTD_E_TIME_VALUE_ERROR    value error in asciiUTC
   PGS_E_TOOLKIT               an unexpected error occurred

EXAMPLES:
C:
    #include <PGS_EPH.h>

    #define ORBIT_ARRAY_SIZE 5   /# maximum number of orbits expected #/
    #define EPHEM_ARRAY_SIZE 100 /# number of ephemeris data points #/

    PGSt_double     offsets[EPHEM_ARRAY_SIZE];
    PGSt_double     orbitDownLongitude[ORBIT_ARRAY_SIZE];

    PGSt_integer    numOrbits;

    char            asciiUTC[28];
    char            orbitAscendTime[ORBIT_ARRAY_SIZE][28];
    char            orbitDescendTime[ORBIT_ARRAY_SIZE][28];

    /# initialize asciiUTC and offsets array with the times for actual
       ephemeris records that will be processed (i.e. by some other tool) #/

    strcpy(asciiUTC,"1998-02-03T19:23:45.123");
    for (i=0;i<EPHEM_ARRAY_SIZE;i++)
    {
        offsets[i] = (PGSt_double) i*60.0;
    }

    /# get the ephemeris metadata associated with these times #/

    returnStatus = PGS_EPH_GetEphMet(PGSd_EOS_AM,EPHEM_ARRAY_SIZE,asciiUTC,
                                     offsets,&numOrbits,orbitAscendTime,
				     orbitDescendTime,orbitDownLongitude);

    if (returnStatus != PGS_S_SUCCESS)
    {
                :
     ** do some error handling ***
                :
    }

    /# numOrbits will now contain the number of orbits spanned by the data set
       (as defined by asciiUTC and EPHEM_ARRAY_SIZE offsets). orbitAscendTime
       will contain numOrbits ASCII UTC times representing the time of northward
       equator crossing of the spacecraft for each respective orbit.
       orbitDescendTime will similarly contain the southward equator crossing
       times and orbitDownLongitude will contain the southward equator crossing
       longitudes #/

FORTRAN:
      implicit none

      include 'PGS_EPH_5.f'
      include 'PGS_TD.f'
      include 'PGS_TD_3.f'
      include 'PGS_SMF.f'

      integer orbit_array_size/5/   ! maximum number of orbits expected
      integer ephem_array_size/100/ ! number of ephemeris data points

      double precision offsets(ephem_array_size)
      double precision orbitdownlongitude(orbit_array_size)

      integer          numorbits

      character*27     asciiutc
      character*27     orbitascendtime(orbit_array_size)
      character*27     orbitdescendtime(orbit_array_size)

!    initialize asciiutc and offsets array with the times for actual
!    ephemeris records that will be processed (i.e. by some other tool)

      asciiutc = '1998-02-03t19:23:45.123'
      do 100 i=1,ephem_array_size
          offsets(i) = i*60.D0
 100  continue

!    get the ephemeris metadata associated with these times

      returnStatus = pgs_eph_getephmet(pgsd_eos_am,ephem_array_size,
     >                                 asciiutc,offsets,numorbits,
     >			               orbitascendtime,orbitdescendtime,
     >  			       orbitdownlongitude)

      if (returnStatus .ne. pgs_s_success) then
                  :
       ** do some error handling ***
                  :
      endif

!    numOrbits will now contain the number of orbits spanned by the data set
!    (as defined by asciiUTC and EPHEM_ARRAY_SIZE offsets). orbitAscendTime
!    will contain numOrbits ASCII UTC times representing the time of northward
!    equator crossing of the spacecraft for each respective orbit.
!    orbitDescendTime will similarly contain the southward equator crossing
!    times and orbitDownLongitude will contain the southward equator crossing
!    longitudes

NOTES:
   This function will determine the time span of the data set from the input UTC
   reference time and the offsets array.  It will then determine the metadata
   for all orbits spanned by the data set.

   TIME ACRONYMS:

     TAI is:  International Atomic Time
     UTC is:  Coordinated Universal Time

   TIME BOUNDARIES:

     The minimum and maximum times that can successfully be processed by this
     function depend on the file leapsec.dat which relates leap second (TAI-UTC)
     values to UTC Julian dates.  The file leapsec.dat starts at Jan 1, 1961;
     therefore an error status message will be returned if this function is
     called with a time before that date.  The file, which is updated when a new
     leap second event is announced, contains actual (definitive) and predicted
     (long term; very approximate) dates of leap second events.  The latter can
     be used only in simulations.   If an input date is outside of the range of
     dates in the file (or if the file cannot be read) the function will use
     a calculated value of TAI-UTC based on a linear fit of the data known to be
     in the table.  This value of TAI-UTC is relatively crude estimate and may
     be off by as much as 1.0 or more seconds.  Thus, when the function is used
     for dates in the future of the date and time of invocation, the user ought
     to carefully check the return status.  The status level of the return
     status of this function will be 'W' (at least) if the TAI-UTC value used
     is a predicted value.  The status level will be 'E' if the TAI-UTC value
     is calculated (although processing will continue in this case, using the
     calculated value of TAI-UTC).

   ASCII UTC:

     Toolkit functions that accept an ASCII time as input require the time to
     be UTC time in CCSDS ASCII Time Code A or CCSDS ASCII Time Code B format.
     Toolkit functions that return an ASCII time return the UTC time in CCSDS
     ASCII Time Code A format (see CCSDS 301.0-B-2 for details).
     (CCSDS => Consultative Committee for Space Data Systems)

      The general format is:

          YYYY-MM-DDThh:mm:ss.ddddddZ (Time Code A)
          YYYY-DDDThh:mm:ss.ddddddZ   (Time Code B)

          where:
              -,T,: are field delimiters
              Z is the (optional) time code terminator
              other fields are numbers as implied:
                Time Code A:
                   4 Year digits, 2 Month, 2 Day, 2 hour,
                   2 minute, 2 second, and up to 6 decimal
                   digits representing a fractional second
                Time Code B:
                   4 Year digits, 3 Day of year, 2 hour,
                   2 minute, 2 second, and up to 6 decimal
                   digits representing a fractional second

   TOOLKIT INTERNAL TIME (TAI):

     Toolkit internal time is the real number of continuous SI seconds since the
     epoch of UTC 12 AM 1-1-1993.  Toolkit internal time is also referred to in
     the toolkit as TAI (upon which it is based).

   TIME OFFSETS:

      This function accepts an ASCII UTC time, an array of time offsets and the
      number of offsets as input.  Each element in the offset array is an offset
      in seconds relative to the initial input ASCII UTC time.

      An error will be returned if the number of offsets specified is less than
      zero.  If the number of offsets specified is actually zero, the offsets
      array will be ignored.  In this case the input ASCII UTC time will be
      converted to Toolkit internal time (TAI) and this time will be used to
      process the data.  If the number of offsets specified is one (1) or
      greater, the input ASCII UTC time will converted to TAI and each element
      'i' of the input data will be processed at the time: (initial time) +
      (offset[i]).

      Examples:

         if numValues is 0 and asciiUTC is "1993-001T12:00:00" (TAI: 432000.0),
         then input[0] will be processed at time 432000.0 and return output[0]

	 if numValues is 1 and asciiUTC is "1993-001T12:00:00" (TAI: 432000.0),
         then input[0] will be processed at time 432000.0 + offsets[0] and
         return output[0]

         if numValues is N and asciiUTC is "1993-001T12:00:00" (TAI: 432000.0),
         then each input[i] will be processed at time 43000.0 + offsets[i] and
         the result will be output[i], where i is on the interval [0,N)

REQUIREMENTS:
   PGSTK - 0720

DETAILS:
   None

GLOBALS:
   None

FILES:
   None

FUNCTIONS_CALLED:
   PGS_IO_Gen_Open()
   PGS_IO_Gen_Close()
   PGS_TD_UTCtoTAI()
   PGS_TD_TAItoUTC()
   PGS_SMF_SetUnknownMsg()
   PGS_SMF_SetDynamicMsg()

END_PROLOG:
*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <PGS_SMF.h>
#include <PGS_IO.h>
#include <PGS_TD.h>
#include <PGS_EPH.h>

/* name of this function */

#define  FUNCTION_NAME "PGS_EPH_GetEphMet()"


PGSt_SMF_status
PGS_EPH_GetEphMet(                           /* get metadata associated with a
						toolkit ephemeris file */
    PGSt_tag         spacecraftTag,          /* spacecraft ID tag */
    PGSt_integer     numValues,              /* number of values requested */
    char             asciiUTC[28],           /* reference start time */
    PGSt_double      offsets[],              /* time offsets relative to
						asciiUTC */
    PGSt_integer*    numOrbits,              /* number of orbits spanned by
						data set */
    PGSt_integer     orbitNumber[],          /* array of orbit numbers */
    char             orbitAscendTime[][28],  /* array of ascend times */
    char             orbitDescendTime[][28], /* array of descend times */
    PGSt_double      orbitDownLongitude[])   /* array of s/c downward equator
						crossing longitudes */
{
    PGSt_ephemHeader   fileHeader;            /* s/c ephemeris file header */

    int                i;                     /* loop counter */
    int                j;                     /* loop counter */

    size_t             numCheck;              /* used to check calls to fread */

    PGSt_integer       num_files;             /* number of physical files
						 spanned by data set */
    PGSt_integer       totalOrbits=0;         /* total number of orbits spanned
						 by ephemeris files containing
						 at least some orbit data for
						 the requested input times */
    PGSt_integer       start_file_num;        /* file number of first physical
						 file */
    PGSt_integer       stop_file_num;         /* file number of last physical
						 file */
    PGSt_integer       previous_orbit_num=-2; /* stop orbit number of previous
						 file */
    PGSt_integer       *sortedIndex=NULL;     /* pointer to array holding sorted
						 indices of time offset array
						 (offsets) */

    PGSt_scTagInfo  scTagInfo;                /* s/c tag info. structure */

    PGSt_ephemMetadata *orbMetadata=NULL;     /* pointer to s/c metadata */

    PGSt_double        startTAI93;            /* initial time of data set */
    PGSt_double        stopTAI93;             /* final time of data set */

    FILE               *ephemFilePtr=NULL;    /* ptr to s/c ephemeris file */

    PGSt_SMF_status    returnStatus;          /* status of TK func. calls */

    PGSt_hdrSummary    *header_array=NULL;    /* pointer to array of header
					         summaries */

    /*******************
     * BEGIN EXECUTION *
     *******************/

    /* check spacecraft tag for validity */

    returnStatus = PGS_EPH_GetSpacecraftData(spacecraftTag, NULL,
					     PGSe_TAG_SEARCH, &scTagInfo);
    switch (returnStatus)
    {
      case PGS_S_SUCCESS:
	break;

      case PGSTD_E_SC_TAG_UNKNOWN:
      case PGS_E_TOOLKIT:
	return returnStatus;

      default:
	PGS_SMF_SetUnknownMsg(returnStatus, FUNCTION_NAME);
	return PGS_E_TOOLKIT;
    }

    /* convert input asciiUTC to equivalent TAI value */

    returnStatus = PGS_TD_UTCtoTAI(asciiUTC, &startTAI93);
    switch (returnStatus)
    {
      case PGS_S_SUCCESS:
      case PGSTD_E_NO_LEAP_SECS:
	break;

      case PGSTD_E_TIME_FMT_ERROR:
      case PGSTD_E_TIME_VALUE_ERROR:
      case PGS_E_TOOLKIT:
	return returnStatus;

      default:
	PGS_SMF_SetUnknownMsg(returnStatus, FUNCTION_NAME);
	return PGS_E_TOOLKIT;
    }

    if (numValues > 0)
    {
	returnStatus = PGS_MEM_Calloc((void **) &sortedIndex, 1,
				      sizeof(PGSt_integer)*numValues);
	if (returnStatus != PGS_S_SUCCESS)
	    return returnStatus;

	PGS_TD_sortArrayIndices(offsets,numValues,sortedIndex);

	stopTAI93 = startTAI93 + offsets[sortedIndex[numValues-1]];
	startTAI93 = startTAI93 + offsets[sortedIndex[0]];

	PGS_MEM_Free(sortedIndex);
    }
    else if (numValues == 0)
    {
	stopTAI93 = startTAI93;
    }
    else
    {
	PGS_SMF_SetStaticMsg(PGSEPH_E_BAD_ARRAY_SIZE, FUNCTION_NAME);
	return PGSEPH_E_BAD_ARRAY_SIZE;
    }

    /* get the header summary array */

    returnStatus = PGS_EPH_getEphemHeaders(&scTagInfo, &header_array,
					   &num_files);
    if (returnStatus != PGS_S_SUCCESS)
    {
	return returnStatus;
    }

    if (startTAI93 < header_array[0].startTAI93 ||
	stopTAI93 > header_array[num_files-1].stopTAI93)
    {
	returnStatus = PGSEPH_E_NO_SC_EPHEM_FILE;
	PGS_SMF_SetDynamicMsg(returnStatus,
			      "input time outside range of staged ephemeris "
			      "file(s)", FUNCTION_NAME);
	return returnStatus;
    }

    /* determine which of the staged physical s/c ephemeris files are spanned by
       the data set (as defined by the start and stop times) */

    start_file_num = 0;

    for (i=1; i<num_files; i++)
    {
	if (startTAI93 < header_array[i].startTAI93)
	{
	    start_file_num = i-1;
	    break;
	}
	else
	{
	    start_file_num = i;
	}
    }

    stop_file_num = 0;

    for (i=1; i<num_files; i++)
    {
	if (stopTAI93 < header_array[i].startTAI93)
	{
	    if ( stopTAI93 > header_array[i-1].stopTAI93  )
	    {
		stop_file_num = i;
	    }
	    else
	    {
		stop_file_num = i-1;
	    }
	    break;
	}
	else
	{
	    stop_file_num = i;
	}
    }

    /* extract the data set metadata */

    *numOrbits = 0;

    for (i=start_file_num;i<=stop_file_num;i++)
    {
	/* open s/c ephemeris file at requested time */

	returnStatus = PGS_IO_Gen_Open(PGSd_SC_EPHEM_DATA,
				       PGSd_IO_Gen_Read,
				       &ephemFilePtr,
				       header_array[i].file_version_num);

	if (returnStatus != PGS_S_SUCCESS)
	{
	    returnStatus = PGS_E_TOOLKIT;
	    PGS_SMF_SetDynamicMsg(returnStatus,
				  "Unexpected failure opening ephemeris file.",
				  FUNCTION_NAME);
	    if (orbMetadata != NULL)
	    {
		free(orbMetadata);
	    }
	    return returnStatus;
	}

	/* read the full ephemeris file header */

	numCheck = fread(&fileHeader, sizeof(PGSt_ephemHeader), 1,
			 ephemFilePtr);
	if (numCheck != 1)
	{
	    returnStatus = PGS_E_TOOLKIT;
	    PGS_SMF_SetDynamicMsg(returnStatus, "Unexpected failure reading "
				  "ephemeris file header.", FUNCTION_NAME);
	    if (orbMetadata != NULL)
	    {
		free(orbMetadata);
	    }
	    return returnStatus;
	}

	/* verify that this file is actually the next in line (in case there are
	   gaps in the staged files) */

	if (i != start_file_num)
	{
/* 	    if (fileHeader.orbitNumberStart > previous_orbit_num+1) */
	    if (fileHeader.orbitNumberStart != previous_orbit_num &&
		fileHeader.orbitNumberStart != previous_orbit_num+1)
	    {
		if (stopTAI93 > header_array[i-1].stopTAI93)
		{
		    returnStatus = PGSEPH_E_NO_SC_EPHEM_FILE;
		    PGS_SMF_SetDynamicMsg(returnStatus,
					  "input time outside range of staged "
					  "ephemeris file(s)", FUNCTION_NAME);
		    if (orbMetadata != NULL)
		    {
			free(orbMetadata);
		    }
		    return returnStatus;
		}
		break;
	    }
	}

	previous_orbit_num = fileHeader.orbitNumberEnd;
	totalOrbits += fileHeader.nOrbits;

	/* move to the end of the records (i.e. to the beginning of the s/c
	   ephemeris file metadata */

	fseek(ephemFilePtr,
	      (long)(sizeof(PGSt_ephemRecord)*fileHeader.nRecords) +
	      (long)(fileHeader.nURs*PGSd_UR_FIELD_SIZE),
	      SEEK_CUR);

	/* extract the appropriate metadata for the data set */

	orbMetadata = (PGSt_ephemMetadata*) realloc((void*)orbMetadata,
						    (totalOrbits+1)*sizeof(PGSt_ephemMetadata));
	numCheck = fread(orbMetadata+(totalOrbits-fileHeader.nOrbits),
			 sizeof(PGSt_ephemMetadata),
			 fileHeader.nOrbits, ephemFilePtr);

	if (numCheck != fileHeader.nOrbits)
	{
	    returnStatus = PGS_E_TOOLKIT;
	    PGS_SMF_SetDynamicMsg(returnStatus, "Unexpected failure reading"
				  " ephemeris file header.", FUNCTION_NAME);
	    if (orbMetadata != NULL)
	    {
		free(orbMetadata);
	    }
	    return returnStatus;
	}

	/* close this physical file */

	returnStatus = PGS_IO_Gen_Close(ephemFilePtr);

	if (returnStatus != PGS_S_SUCCESS)
	{
	    PGS_SMF_SetDynamicMsg(PGS_E_TOOLKIT,
				  "Unexpected failure closing ephemeris "
				  "file.", FUNCTION_NAME);
	    if (orbMetadata != NULL)
	    {
		free(orbMetadata);
	    }
	    return PGS_E_TOOLKIT;
	}
    }
    orbMetadata[totalOrbits].orbitAscendTime =
      orbMetadata[totalOrbits-1].orbitAscendTime + PGSd_GEO_ERROR_VALUE;

    for (j=0;j<totalOrbits;j++)
    {
	/* this metadata is only relevant if any part of this orbit is
	   spanned by the data set */

	if (startTAI93 >= orbMetadata[j+1].orbitAscendTime)
	{
	    continue;
	}

	if (stopTAI93 < orbMetadata[j].orbitAscendTime)
	{
	    break;
	}

	/* if this is not the first pass through (i.e. we have already recorded
	   information about at least one orbit) check the current orbit number
	   against the last orbit number recorded, don't record duplicate orbit
	   numbers (this can happen when the data for an orbit is found in two
	   separate physical files since each file will contain the metadata for
	   that particular orbit and this metadata will obviously be the same in
	   both cases) */

	if ( *numOrbits > 0 )
	{
	    if ( orbMetadata[j].orbitNumber ==
		 orbitNumber[(*numOrbits) - 1] )
	    {
		continue;
	    }
	}

	/* record the orbit number */

	orbitNumber[*numOrbits] = orbMetadata[j].orbitNumber;

	/* convert orbit start time to ASCII string and record it */

	returnStatus = PGS_TD_TAItoUTC(orbMetadata[j].orbitAscendTime,
				       orbitAscendTime[*numOrbits]);
	switch (returnStatus)
	{
	  case PGS_S_SUCCESS:
	  case PGSTD_E_NO_LEAP_SECS:
	    break;

	  case PGS_E_TOOLKIT:
	    if (orbMetadata != NULL)
	    {
		free(orbMetadata);
	    }
	    return returnStatus;

	  default:
	    PGS_SMF_SetUnknownMsg(returnStatus, FUNCTION_NAME);
	    if (orbMetadata != NULL)
	    {
		free(orbMetadata);
	    }
	    return PGS_E_TOOLKIT;
	}

	/* convert orbit stop time to ASCII string and record it */

	returnStatus = PGS_TD_TAItoUTC(orbMetadata[j].orbitDescendTime,
				       orbitDescendTime[*numOrbits]);
	switch (returnStatus)
	{
	  case PGS_S_SUCCESS:
	  case PGSTD_E_NO_LEAP_SECS:
	    break;

	  case PGS_E_TOOLKIT:
	    if (orbMetadata != NULL)
	    {
		free(orbMetadata);
	    }
	    return returnStatus;

	  default:
	    PGS_SMF_SetUnknownMsg(returnStatus, FUNCTION_NAME);
	    if (orbMetadata != NULL)
	    {
		free(orbMetadata);
	    }
	    return PGS_E_TOOLKIT;
	}

	/* record longitude of downward (i.e. south-bound) equator
	       crossing for this orbit */

	orbitDownLongitude[*numOrbits] =
	  orbMetadata[j].orbitDescendLongitude;

	/* increment the number of relevant orbits found */

	*numOrbits = (*numOrbits) + 1;
    }

    if (orbMetadata != NULL)
    {
	free(orbMetadata);
    }
    return PGS_S_SUCCESS;
}


/*******************************************************************************
BEGIN_PROLOG:

NAME:
  PGS_EPH_GetEphMetAux.c

DESCRIPTION:
  This function is a wraper around the function PGS_EPH_GetEphMet()
  This function (which has one more argument than  PGS_EPH_GetEphMet), is used
  to make the fortran binding work properly, where the size of 
  orbitAscendTime, and orbitDescendTime arrays are required in the 
  fortran binder, but not present in the argument list of PGS_EPH_GetEphMet.
  The size of orbitAscendTime, and orbitDescendTime arrays passed to this 
  function will be maxorbitarraysize.
  See PGS_EPH_GetEphMetF.f for more information

AUTHOR:
  Abe Taaheri / Space Applications Corporation

HISTORY:
  29-SEPT-1998  AT  Initial version

END_PROLOG:
*******************************************************************************/

PGSt_SMF_status
PGS_EPH_GetEphMetAux(                        /* get metadata associated with a
						toolkit ephemeris file */
    PGSt_tag         spacecraftTag,          /* spacecraft ID tag */
    PGSt_integer     numValues,              /* number of values requested */
    PGSt_integer     maxorbitarraysize,      /* number of elements in
						orbitAscendTime and
						orbitDescendTime for tempororay
						arrays defined in the fortran
						function pgs_eph_getephmet2 */
    char             asciiUTC[28],           /* reference start time */
    PGSt_double      offsets[],              /* time offsets relative to
						asciiUTC */
    PGSt_integer*    numOrbits,              /* number of orbits spanned by
						data set */
    PGSt_integer     orbitNumber[],          /* array of orbit numbers */
    char             orbitAscendTime[][28],  /* array of ascend times */
    char             orbitDescendTime[][28], /* array of descend times */
    PGSt_double      orbitDownLongitude[])   /* array of s/c downward equator
						crossing longitudes */
{

    PGSt_SMF_status    returnStatus;          /* status of TK func. calls */
    
    returnStatus =  PGS_EPH_GetEphMet(spacecraftTag, numValues, asciiUTC,
				      offsets, numOrbits, orbitNumber,
				      orbitAscendTime, orbitDescendTime,
				      orbitDownLongitude);
    return returnStatus;
    
}
