/*-------------------------------------------------------------------------*/
/*                                                                         */
/*  COPYRIGHT[copyright mark] 1996, Hughes Aircraft Company, its vendors,  */
/*  and suppliers.  ALL RIGHTS RESERVED.                                   */
/*                                                                         */
/*-------------------------------------------------------------------------*/
/*******************************************************************************
BEGIN_FILE_PROLOG:

FILENAME:
  PGS_EPH_getAttitRecords.c

DESCRIPTION:
  This file contains the function PGS_EPH_getAttitRecords()

AUTHOR:
  Guru Tej S. Khalsa / Applied Reseach Corporation

HISTORY:
  19-Apr-1996  GTSK  Initial version

END_FILE_PROLOG:
*******************************************************************************/

/*******************************************************************************
BEGIN_PROLOG:

TITLE:
   Get Array of Spacecraft Attitude Records

NAME:
   PGS_EPH_getAttitRecords()

SYNOPSIS:
C:
   #include <PGS_EPH.h>

   PGSt_SMF_status
   PGS_EPH_getAttitRecords(
       PGSt_scTagInfo   *scTagInfo,
       PGSt_double      secTAI93,
       PGSt_integer     maxArraySize,
       PGSt_attitRecord *attitRecordArray,
       PGSt_integer     *totalRecords)
      
DESCRIPTION:
   This tool gets an array of spacecraft attitude records (time and number of
   records specified in calling function).

INPUTS:
   Name              Description               Units        Min         Max
   ----              -----------               -----        ---         ---
   scTagInfo         spacecraft tag            N/A
                     information structure

   secTAI93          time of first requested   sec
                     record (in seconds since
		     12 AM UTC 1-1-1993)

   maxArraySize      size of s/c record        N/A
                     structure array passed

OUTPUTS:
   Name              Description               Units        Min         Max
   ----              -----------               -----        ---         ---
   attitRecordArray  array of s/c attitude    N/A
                     records

   totalRecords      number of records passed
                     back in attitRecordArray

RETURNS:
   PGS_S_SUCCESS               successful return 
   PGSEPH_M_SHORT_ARRAY        number of records returned is less than the max.
                               size allowed (i.e. the array size passed in)
   PGSEPH_E_NO_SC_EPHEM_FILE   no s/c attit files could be found for input times
   PGS_E_TOOLKIT               something really gonzo happened

EXAMPLES:
C:
        #define ARRAY_SIZE 500

	PGSt_double      secTAI93 = 205124400.0;
	
	PGSt_scTagInfo   scTagInfo;

	PGSt_attitRecord attitRecordArray[ARRAY_SIZE];

	PGSt_integer     totalRecords;

	PGSt_SMF_status  returnStatus;


	returnStatus = PGS_EPH_GetSpacecraftData(PGSd_EOS_AM, NULL,
                                                 PGSe_TAG_SEARCH, &scTagInfo);

	if (returnStatus != PGS_S_SUCCESS)
	{
	            :
	 ** do some error handling ***
		    :
	}
   
        returnStatus = PGS_EPH_getAttitRecords(&scTagInfo,secTAI93,ARRAY_SIZE,
	                                       attitRecordArray,&totalRecords);

	if (returnStatus != PGS_S_SUCCESS)
	{
	            :
	 ** do some error handling ***
		    :
	}
   
NOTES:
   None

REQUIREMENTS:
   PGSTK - 0720

DETAILS:
   None

GLOBALS:
   None

FILES:
   This file accesses s/c attitude files (specified in the PCF).

FUNCTIONS_CALLED:
   PGS_EPH_getAttitHeaders()
   PGS_IO_Gen_Open()
   PGS_IO_Gen_Close()
   PGS_SMF_SetStaticMsg()    sets the message buffer
   PGS_SMF_SetDynamicMsg()   sets the message buffer

END_PROLOG:
*******************************************************************************/

#include <string.h>
#include <sys/types.h>
#include <stdio.h>
#include <PGS_IO.h>
#include <PGS_EPH.h>

/* name of this function */

#define FUNCTION_NAME "PGS_EPH_getAttitRecords()"

PGSt_SMF_status
PGS_EPH_getAttitRecords(
    PGSt_scTagInfo*   scTagInfo,
    PGSt_double       secTAI93,
    PGSt_integer      maxArraySize,
    PGSt_attitRecord* attitRecordArray,
    PGSt_integer*     totalRecords)
{
    PGSt_attitHeader  fileHeader;

    int               i;
    int               seekCheck;

    size_t            temp;
    size_t            numCheck;

    PGSt_integer      numRecords;
    
    FILE              *attitFilePtr=NULL;
    
    PGSt_SMF_status   returnStatus;

    /* static variables */

    static PGSt_tag      oldSpacecraftTag=0;

    static long          nextRecordPtr;

    static PGSt_integer  lastArraySize=-1;
    static PGSt_integer  file_num=-1;
    static PGSt_integer  num_files=-1;

    static PGSt_hdrSummary *header_array=NULL;
    

    if (scTagInfo->spacecraftTag == oldSpacecraftTag &&
        lastArraySize > 0 &&
	secTAI93 >= attitRecordArray[0].secTAI93 &&
	secTAI93 <= attitRecordArray[lastArraySize-1].secTAI93)
    {
	return PGS_S_SUCCESS;	
    }
    
    if (scTagInfo->spacecraftTag != oldSpacecraftTag)
    {
	returnStatus = PGS_EPH_getAttitHeaders(scTagInfo, &header_array,
					       &num_files);
	if (returnStatus != PGS_S_SUCCESS)
	{
	    return returnStatus;
	}
	oldSpacecraftTag = scTagInfo->spacecraftTag;
	lastArraySize = -1;
	file_num = -1;
	nextRecordPtr = 0;
    }

    if (secTAI93 < header_array[0].startTAI93 ||
	secTAI93 > header_array[num_files-1].stopTAI93)
    {
	returnStatus = PGSEPH_E_NO_SC_EPHEM_FILE;
	PGS_SMF_SetDynamicMsg(returnStatus,
			      "input time outside range of staged attitude "
			      "file(s)", FUNCTION_NAME);
	return returnStatus;
    }
    
    /* If the currently open file has already been read from (true if this
       routine has been called with the same spacecraft ID and time as in a
       previous call) set the file pointer to next record to be read.  Otherwise
       save the new time and/or s/c ID information. */
    
    *totalRecords = 0;
    numRecords = maxArraySize;

    /* although the following is already as clear as mud, perhaps it could use a
       bit more explanation */

    for (i=0; i<num_files; i++)
    {
	if (secTAI93 < header_array[i].stopTAI93)
	{
	    if (secTAI93 < header_array[i].startTAI93)
	    {
		/* this handles the special case of a time that falls inbetween
		   the last time of one physical file and the first time of the
		   (chronologically) next physical file */

		i--;
	    }
	    
	    if (i == file_num && secTAI93 >= attitRecordArray[0].secTAI93)
	    {
		memcpy((char *) attitRecordArray,
		       (char*) (attitRecordArray+lastArraySize-1),
		       sizeof(PGSt_attitRecord));
		*totalRecords = 1;
		numRecords -= 1;
		break;
	    }
	    else
	    {
		file_num = i;
		nextRecordPtr = 0;
		*totalRecords = 0;
		break;
	    }
	}
    }

    /* open s/c attitude file at requested time */

    returnStatus = PGS_IO_Gen_Open(PGSd_SC_ATTIT_DATA, 
				   PGSd_IO_Gen_Read,
				   &attitFilePtr,
				   header_array[file_num].file_version_num);
    
    if (returnStatus != PGS_S_SUCCESS) 
    {
	returnStatus = PGS_E_TOOLKIT;
	PGS_SMF_SetDynamicMsg(returnStatus,
			      "Unexpected failure opening attitude file.",
			      FUNCTION_NAME);
	return returnStatus;
    }

    numCheck = fread(&fileHeader, sizeof(PGSt_attitHeader), 1, attitFilePtr);
    if (numCheck != 1)
    {
	returnStatus = PGS_E_TOOLKIT;
	PGS_SMF_SetDynamicMsg(returnStatus, "Unexpected failure reading "
			      "attitude file header.", FUNCTION_NAME);
	return returnStatus;
    }

    seekCheck = fseek(attitFilePtr, fileHeader.nURs*PGSd_UR_FIELD_SIZE,
		      SEEK_CUR);
    if (seekCheck != 0)
    {
	returnStatus = PGS_E_TOOLKIT;
	PGS_SMF_SetDynamicMsg(returnStatus, "Unexpected failure reading "
			      "attitude file header.", FUNCTION_NAME);
	return returnStatus;
    }

    if (numRecords > fileHeader.nRecords)
    {
	numRecords = fileHeader.nRecords;
    }

    if (nextRecordPtr != 0)
    {
	seekCheck = fseek(attitFilePtr,nextRecordPtr,SEEK_SET);
	if (seekCheck != 0)
	{
	    returnStatus = PGS_E_TOOLKIT;
	    PGS_SMF_SetDynamicMsg(returnStatus, "Unexpected failure reading "
				  "attitude file.", FUNCTION_NAME);
	    return returnStatus;
	}

	temp = nextRecordPtr - sizeof(PGSt_attitHeader);
	if (temp%sizeof(PGSt_attitRecord) != 0)
	{
	    PGS_SMF_SetDynamicMsg(PGS_E_TOOLKIT,
				  "Corrupt attitude file pointer.",
				  FUNCTION_NAME);
	    return PGS_E_TOOLKIT;
	}
	temp = temp/sizeof(PGSt_attitRecord);
	if (numRecords > (fileHeader.nRecords - temp))
	{
	    numRecords = fileHeader.nRecords - (PGSt_integer) temp;
	}
    }
    
    do
    {
	*totalRecords += fread(attitRecordArray+(*totalRecords),
			       sizeof(PGSt_attitRecord),numRecords,
			       attitFilePtr);

	lastArraySize = *totalRecords;
	nextRecordPtr = ftell(attitFilePtr);

	returnStatus = PGS_IO_Gen_Close(attitFilePtr);
	
	if (returnStatus != PGS_S_SUCCESS)
	{
	    PGS_SMF_SetDynamicMsg(PGS_E_TOOLKIT,
				  "Unexpected failure closing attitude "
				  "file.", FUNCTION_NAME);
	    return PGS_E_TOOLKIT;
	}
	
	if (*totalRecords < maxArraySize)
	{
	    numRecords = maxArraySize - *totalRecords;
	    file_num++;
	    if (file_num >= num_files)
	    {
                return PGSEPH_M_SHORT_ARRAY;		
	    }

	    returnStatus = PGS_IO_Gen_Open(PGSd_SC_ATTIT_DATA, 
					   PGSd_IO_Gen_Read,
					   &attitFilePtr,
					   header_array[file_num].
					   file_version_num);
	    
	    if (returnStatus != PGS_S_SUCCESS) 
	    {
		returnStatus = PGS_E_TOOLKIT;
		PGS_SMF_SetDynamicMsg(returnStatus,
				      "Unexpected failure opening attitude "
				      "file.", FUNCTION_NAME);
		return returnStatus;
	    }
	    
	    numCheck = fread(&fileHeader, sizeof(PGSt_attitHeader), 1, 
			     attitFilePtr);
	    if (numCheck != 1)
	    {
		returnStatus = PGS_E_TOOLKIT;
		PGS_SMF_SetDynamicMsg(returnStatus,"Unexpected failure reading "
				      "attitude file header.", FUNCTION_NAME);
		return returnStatus;
	    }

	    seekCheck = fseek(attitFilePtr, fileHeader.nURs*PGSd_UR_FIELD_SIZE,
			      SEEK_CUR);
	    if (seekCheck != 0)
	    {
		returnStatus = PGS_E_TOOLKIT;
		PGS_SMF_SetDynamicMsg(returnStatus, "Unexpected failure reading"
				      " attitude file header.", FUNCTION_NAME);
		return returnStatus;
	    }

	    if (numRecords > fileHeader.nRecords)
	    {
		numRecords = fileHeader.nRecords;
	    }
	}
    }
    while (*totalRecords < maxArraySize);

    if (attitRecordArray[*totalRecords-1].secTAI93 < secTAI93)
    {
	returnStatus = PGS_EPH_getAttitRecords(scTagInfo,secTAI93,
					       maxArraySize,attitRecordArray,
					       totalRecords);
    }
    
    return returnStatus;
}
