/*-------------------------------------------------------------------------*/
/*                                                                         */
/*  COPYRIGHT[copyright mark] 1996, Hughes Aircraft Company, its vendors,  */
/*  and suppliers.  ALL RIGHTS RESERVED.                                   */
/*                                                                         */
/*-------------------------------------------------------------------------*/
#include <stdlib.h>
#include <stdio.h>
#include <PGS_PC.h>
#include <PGS_IO.h>
#include <PGS_EPH.h>

/* name of this function */

#define FUNCTION_NAME "PGS_EPH_getEphemHeaders()"

PGSt_SMF_status
PGS_EPH_getEphemHeaders(
    PGSt_scTagInfo     *scTagInfo,
    PGSt_hdrSummary    **header_array,
    PGSt_integer       *num_files)
{
    PGSt_SMF_status returnStatus = PGS_S_SUCCESS;   /* return status */

    PGSt_integer     file_num;
    PGSt_integer     base;
    PGSt_integer     next;
    PGSt_integer     i;
    PGSt_integer     bogus_versions = 0;

    size_t           swap_size;
    size_t           numCheck;

    FILE*            ephemFilePtr;
    
    
    PGSt_ephemHeader header;

    static PGSt_tag         oldSpacecraftTag=0;
    
    static PGSt_hdrSummary  *staticHdrArray=NULL;
    PGSt_hdrSummary         sort_buf;
    
    char             details[PGS_SMF_MAX_MSG_SIZE]; /* detailed error message */

    if (scTagInfo->spacecraftTag == oldSpacecraftTag)
    {
	if (staticHdrArray != NULL)
	{
	    *header_array = staticHdrArray;
	    return PGS_S_SUCCESS;
	}
    }
    
    returnStatus = PGS_PC_GetNumberOfFiles(PGSd_SC_EPHEM_DATA, num_files);

    if (returnStatus != PGS_S_SUCCESS)
    {
	returnStatus = PGSEPH_E_NO_SC_EPHEM_FILE;
	PGS_SMF_SetDynamicMsg(returnStatus,
			      "error in accessing spacecraft ephemeris file(s)",
			      FUNCTION_NAME);
	return returnStatus;
    }

    /*
     *    Allocate the header array and pass the address back to
     *    the calling program.
     */

    if (*header_array != NULL)
    {
	free(*header_array);
    }
    
    *header_array = (PGSt_hdrSummary*) malloc(*num_files*
					      sizeof(PGSt_hdrSummary));

    if (*header_array == NULL)
    {
	PGS_SMF_SetDynamicMsg(PGS_E_TOOLKIT,
			      "Failed to allocate sufficient dynamic memory "
			      "for array of s/c ephemeris file headers.",
			      FUNCTION_NAME);
	return PGS_E_TOOLKIT;
    }

    for (file_num=0;file_num<*num_files;file_num++)
    {
	/* 
	 *    Open the specified physical file
	 */
	
	returnStatus = PGS_IO_Gen_Open(PGSd_SC_EPHEM_DATA, 
				       PGSd_IO_Gen_Read,
				       &ephemFilePtr,
				       file_num+1);

	if (returnStatus != PGS_S_SUCCESS) 
	{
	    PGS_SMF_SetDynamicMsg(PGSEPH_E_NO_SC_EPHEM_FILE,
				  "unable to open spacecraft ephemeris "
				  "file", FUNCTION_NAME);
	    (*header_array)[file_num].startTAI93 = 1.5E50;
	    bogus_versions++;
	    continue;
	}

	/* Read the file header info, checking to make sure that the read is
	   successfully accomplished. */
	
	numCheck = fread(&header, sizeof(PGSt_ephemHeader), 1,
			 ephemFilePtr);

	returnStatus = PGS_IO_Gen_Close(ephemFilePtr);
	
	if (returnStatus != PGS_S_SUCCESS) 
	{
	    PGS_SMF_SetDynamicMsg(PGS_E_TOOLKIT,
				  "Unexpected failure closing ephemeris file.",
				  FUNCTION_NAME);
	    return PGS_E_TOOLKIT;
	}

	if (numCheck != 1)
	{
	    PGS_SMF_SetDynamicMsg(PGSEPH_E_BAD_EPHEM_FILE_HDR,
				  "error attempting to read s/c ephemeris file "
				  "header",
				  FUNCTION_NAME);
	    (*header_array)[file_num].startTAI93 = 1.5E50;
	    bogus_versions++;
	    continue;
	}

	if (strcmp(scTagInfo->spacecraftName, header.spacecraftID) != 0)
	{
	    sprintf(details,"Encountered unexpected spacecraftID (%s) in "
		    "ephemeris file header (expected %s).", header.spacecraftID,
		    scTagInfo->spacecraftName);
	    PGS_SMF_SetDynamicMsg(PGSEPH_E_BAD_EPHEM_FILE_HDR, details,
				  FUNCTION_NAME);
	    (*header_array)[file_num].startTAI93 = 1.5E50;
	    bogus_versions++;
	    continue;
	}
	
	(*header_array)[file_num].file_version_num = file_num + 1;
	(*header_array)[file_num].startTAI93 = header.startTime;
	(*header_array)[file_num].stopTAI93 = header.endTime;
    }
    /*
     *    Sort the version array by start time, from earliest to latest
     */

    swap_size = sizeof(PGSt_hdrSummary); /* used for sorting (if necessary) */

    for (base=0; base<*num_files; base++)
    {

        /* 
         *    Get next earliest element
         */
        for (i=base, next=base; i<*num_files; i++) 
        {
            if ((*header_array)[i].startTAI93 < (*header_array)[next].startTAI93)
            {
                next=i;
            }
        }

        /* 
         *    Swap into position, if needed
         */
        if (next != base)
        {
	    memcpy((char*) &sort_buf, (char*) ((*header_array)+base), swap_size);
	    memcpy((char*) ((*header_array)+base), (char*) ((*header_array)+next),
		   swap_size);
	    memcpy((char*) ((*header_array)+next), (char*) &sort_buf,
		   swap_size);
	}
    }
    if (bogus_versions != 0)
    {
	*num_files -= bogus_versions;
	*header_array = (PGSt_hdrSummary*) realloc(*header_array, *num_files*
						   sizeof(PGSt_hdrSummary));
    }
    
    if (*num_files == 0)
    {
	returnStatus = PGSEPH_E_NO_SC_EPHEM_FILE;
	PGS_SMF_SetDynamicMsg(returnStatus,
			      "No valid ephemeris files found.",
			      FUNCTION_NAME);
    }

    staticHdrArray = *header_array;
    
    return returnStatus;
}
