/*-------------------------------------------------------------------------*/
/*                                                                         */
/*  COPYRIGHT[copyright mark] 1996, Hughes Aircraft Company, its vendors,  */
/*  and suppliers.  ALL RIGHTS RESERVED.                                   */
/*                                                                         */
/*-------------------------------------------------------------------------*/
/*******************************************************************************
BEGIN_FILE_PROLOG:

FILENAME:
   orbsim.c

DESCRIPTION:
   Orbit and attitude simulator.

AUTHOR:
   Guru Tej S. Khalsa / Applied Research Corporation
   Peter D. Noerdlinger / SM&A Inc.

HISTORY:
   17-Oct-1994  GTSK  Initial version
   31-Jan-1995  GTSK  Changed the way that record times are calculated.
   21-Jul-1995  GTSK  Changed printf and scanf statements so that they are only
                      using "int" and "double" types (or pointers) instead of
		      "PGSt_integer" or "PGSt_double" types which may be of the 
		      wrong type (i.e. they are NOT necessarily "int" and
		      "double" respectively).
   12-Feb-1999  PDN   Fixed nomenclature consistent with AM1 and later spacecraft
                      note that the function is also OK for TRMM, but in that case
                      please observe that angularVelocity is really (yaw rate,
                      pitch rate, and roll rate in that order).
   13-May-1999  PDN   Reduced the starting orbit number for simulations before
                      launch date from 116880 to 30303. This allows tests with
                      data server.

END_FILE_PROLOG:
*******************************************************************************/

/*******************************************************************************
BEGIN_PROLOG:

TITLE:
   Spacecraft Orbit and Attitude Simulator

NAME:
   orbsim

SYNOPSIS:
C:



DESCRIPTION:


INPUTS:
   Name         Description               Units       Min   Max
   ----         -----------               -----       ---   ---
 
OUTPUTS:
   Name         Description               Units       Min   Max
   ----         -----------               -----       ---   ---
 
          
RETURNS:
   

EXAMPLES:
C:

FORTRAN:
  
NOTES:
   None

REQUIREMENTS:
   PGSTK - 0720

DETAILS:
   None

GLOBALS:
   None

FILES:
   None

FUNCTIONS_CALLED:
   PGS_EPH_attOrbSim()
   PGS_TD_UTCtoTAI()
   PGS_TD_timeCheck()
   PGS_TD_ASCIItime_BtoA()
   PGS_TD_julday()
   PGS_TD_calday()

END_PROLOG:
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <PGS_math.h>
#include <PGS_PC.h>
#include <PGS_TD.h>
#include <PGS_CSC.h>
#include <PGS_EPH.h>
#include <PGS_SIM.h>

#ifndef M_PI
#define M_PI    3.14159265358979323846
#endif

#define  P    data[9]
#define  e    data[1]
#define  ma   data[5]
#define  ap   data[4]

#define  E(x) (acos((e + cos(x))/(1 + e*cos(x))))

#define  PGSd_TO_TIP -1
#define  PGSd_TO_ORB  1

#define  PGSd_CUSTOM -7             /* for consideration as possible value of
				       s/c tag, used to initiate special
				       processing in this program */

#define  START_BEFORE_LAUNCH 30303  /* hoked up number to prevent negatives */
#define  MAX_TRIES       5U         /* maximum number of times a user may
				       attempt to enter in a requested input
				       quantity before the program gives up on
				       said user */

static void quit(void);
static int file_exists(char *);
static PGSt_boolean get_scTag(char *, PGSt_tag *);
static void tipEuler(PGSt_double [3], PGSt_double [3], PGSt_integer [3],
		     PGSt_double [3], PGSt_integer);
static PGSt_double PGS_EPH_OrbitAscendTime(PGSt_tag, PGSt_double);
static PGSt_double PGS_EPH_OrbitDecendTime(PGSt_tag, PGSt_double);

typedef struct
{
  double       btai;
  double       bea[3];
  double       etai;
  double       eea[3];
  PGSt_boolean cont;
  PGSt_boolean lock;
} PGSt_customAttit;

int main(int argc,char *argv[])
{
    PGSt_ephemHeader ephemFileHeader;
    PGSt_attitHeader attitFileHeader;
    
    PGSt_ephemRecord ephemRecord;
    PGSt_attitRecord attitRecord;

    PGSt_ephemMetadata orbMetadata;
    
    PGSt_tag     spacecraftTag;
    
    double       yprNoise;
    double       yprRateNoise;

    PGSt_integer startOrbitNum;       /* orbit number of first orbit in file */

    PGSt_double  data[10];            /* keplerian orbital elements and other
					 stuff */
    PGSt_double  startTAI93;          /* file start time in seconds since 12AM
					 UTC 1-1-1993 */
    PGSt_double  stopTAI93;           /* file start time in seconds since 12AM
					 UTC 1-1-1993 */
    PGSt_double  defaultInterval;     /* default time interval between data sets
					 (spacecraft dependent) */
    PGSt_double  dataBytes;           /* number of bytes in a data record */
    PGSt_double  transform[3][3];     /* s/c to ECI transformation matrix */
    PGSt_double  numIntervals;        /* number of records in a file */

    PGSt_double  posvelECI[6];
    PGSt_double  posvelECR[6];
    PGSt_double  latitude;
    PGSt_double  longitude;
    PGSt_double  altitude;
    PGSt_double  time;
    PGSt_double  dt1;
    PGSt_double  dt2;
    PGSt_double  nu0;
    PGSt_double  nu;
    PGSt_double  k;
    PGSt_double  quatEuler[4];
    PGSt_double  quatORBtoECI[4];
    PGSt_double  quatSCtoECI[4];
    PGSt_double  quatECItoSC_0[4];
    PGSt_double  quatORBtoSC_0[4];
    PGSt_double  eulerAngle[3];
    PGSt_double  angularVelocity[3];
    PGSt_double  old_eea[3];
    PGSt_double  old_tai;

    double       tempDouble;          /* temporary "double" for use with scanf()
				       */
    unsigned int numTries;            /* number of times the user has attempted
					 to input something */
    int          numCheck;            /* used to verify calls to sscanf() and
					 fwrite() */
    int          year;                /* year */
    int          month;               /* month */
    int          day;                 /* day */
    
    PGSt_integer julianDay;           /* Julian day of file */
    PGSt_integer startJD;             /* Julian day of start day */
    PGSt_integer stopJD;              /* Julian day of stop day */
    PGSt_integer version=1;           /* version number for PC file */
    PGSt_integer count;               /* looping variable */
    PGSt_integer eao[3];
    PGSt_integer num_periods;
    PGSt_integer periodNum;
    
    PGSt_boolean gotData;             /* used to determine if an input value as
					 been successfully ingested */
    PGSt_boolean firstTime=PGS_TRUE;
    PGSt_boolean custom=PGS_FALSE;

    PGSt_customAttit ca[100];
    PGSt_customAttit tmp_ca;
    
    char         *charPtr;
    char         attitudeNoise[12];   /* attitude noise string for simulator */
    char         inputBuffer[100];    /* input buffer */
    char         ephFilename[100];    /* name of s/c ephemeris file */
    char         attFilename[100];    /* name of s/c attitude file */
    char         ephemDirectory[100]; /* name of directory for s/c ephemeris
					 files */
    char         scName[10];          /* spacecraft name */
    char         asciiUTC[28];        /* UTC time */
    char         startUTC[28];        /* beginning UTC time */
    char         stopUTC[28];         /* beginning UTC time */
    char         startUTC_print[28];  /* beginning UTC time (to stdout) */
    char         stopUTC_print[28];   /* beginning UTC time (to stdout) */
    char         *newlinePtr;         /* pointer to newline character in the
					 input buffer */
    char         parentUR[PGSd_UR_FIELD_SIZE];

    char         mnemonic[PGS_SMF_MAX_MNEMONIC_SIZE]; /* status mnemonic
							 returned by
							 PGS_SMF_GetMsg() */
    char         msg[PGS_SMF_MAX_MSG_SIZE];           /* status messsage
							 returned by call to
							 PGS_SMF_GetMsg() */
    
    FILE         *ephemFilePtr;       /* pointer to s/c ephemeris file */
    FILE         *attitFilePtr;       /* pointer to s/c attitude file */
    
    DIR          *directoryPtr;       /* pointer to directory for s/c ephemeris
					 files */
    
    PGSt_SMF_status returnStatus;     /* return status of calls to PGS toolkit
					 functions */
    PGSt_SMF_status code;             /* status code returned by
					 PGS_SMF_GetMsg() */
    
    /*******************
     * BEGIN EXECUTION *
     *******************/

    scName[0] = '\0';

    system("clear\n");

    /* print out lovely picture because Tejmo was bored one day */

    printf("\n       ********************\n"
	   "       * ------O--------- *\n"
	   "       * ___/\\__/\\_______ *\n"
	   "       * __/  \\/  \\______ *\n"
	   "       *  /    \\___\\_____ *\n"
	   "       * /                *\n"
	   "       * ^^^^^^^^^^^^^^^^ *\n"
	   "       * ^^^^^^^^^^^^^^^^ *\n"
	   "       * ^^^^^^^^^^^^^^^^ *\n"
	   "       * =EOS=            *\n"
	   "       ********************\n\n");
    printf("  ECS SPACECRAFT ORBIT SIMULATOR\n\n");

    memset((void*)parentUR, (unsigned char)0, PGSd_UR_FIELD_SIZE);
    strcpy(parentUR, "SIMULATED DATA");
    
  BEGINNING:;
    memset((void*)&ephemFileHeader, (unsigned char)0, sizeof(PGSt_ephemHeader));
    memset((void*)&attitFileHeader, (unsigned char)0, sizeof(PGSt_attitHeader));
    memset((void*)&tmp_ca, (unsigned char)0, sizeof(PGSt_customAttit));
    num_periods = 0;

    printf("Enter <return> at a prompt to select the default\n"
	   "option (indicated by []).  Enter 'q' at any prompt\n"
	   "to quit.\n\n");
    
    gotData = PGS_FALSE;
    numTries = 0;
 	
    /* determine spacecraft, set appropriate values to the spacecraft name
       string (scName) and default time interval (defaultInterval) */
   
    do
    {
	if (strlen(scName) == 0)
	  printf("enter spacecraft ID (TRMM, EOS_AM, EOS_PM):\n"
		 "-->");
	else
	  printf("enter spacecraft ID (TRMM, EOS_AM, EOS_PM) [%s]:\n"
		 "-->",scName);
	fgets(inputBuffer,99,stdin);
	if (inputBuffer[0] == 'q')
	  quit();
	if (inputBuffer[0] == '\n' && strlen(scName) != 0)
	  strcpy(inputBuffer,scName);
	newlinePtr = strchr(inputBuffer,'\n');
	if (newlinePtr != NULL)
	  *newlinePtr = '\0';
	
	if (get_scTag(inputBuffer,&spacecraftTag) == PGS_FALSE)
	  continue;
	   
	gotData = PGS_TRUE;
	switch (spacecraftTag)
	{
	  case PGSd_TRMM:
	    strcpy(scName,"TRMM");
	    defaultInterval = 60.0;
	    eao[0] = 3;
	    eao[1] = 2;
	    eao[2] = 1;
	    break;
	  case PGSd_EOS_AM:
	    strcpy(scName,"EOSAM1");
	    defaultInterval = 1.024;
	    eao[0] = 3;
	    eao[1] = 1;
	    eao[2] = 2;
	    break;
	  case PGSd_EOS_PM:
	    strcpy(scName,"EOSPM1");
	    defaultInterval = 1.024;
	    eao[0] = 3;
	    eao[1] = 1;
	    eao[2] = 2;
	    break;
	  case PGSd_CUSTOM:
	    custom = PGS_TRUE;
	    printf("\ncustom mode enabled...\n\n");
	    goto BEGINNING;
	  default:
	    gotData = PGS_FALSE;
	}	    
    }
    while (gotData == PGS_FALSE && numTries++ < MAX_TRIES);
    if (gotData == PGS_FALSE)
      quit();
    
    /* get start and stop days and time interval */
    
    do /* GET TIMES */
    {
	if (firstTime == PGS_TRUE)
	{
	    printf("\nenter start and stop dates in CCSDS ASCII "
		   "(format A or B)\n");
	    printf(" A) YYYY-MM-DD\n B) YYYY-DDD\n\n");
	}
	else
	  printf("\n");
	    
	gotData = PGS_FALSE;
	numTries = 0;
	
	/* get start day */

	do
	{
	    printf("enter start date:\n-->");
	    fgets(inputBuffer,99,stdin);
	    if (inputBuffer[0] == 'q')
	      quit();
	    newlinePtr = strchr(inputBuffer,'\n');
	    if (newlinePtr != NULL)
	      *newlinePtr = '\0';
	    returnStatus = PGS_TD_timeCheck(inputBuffer);
	    switch (returnStatus)
	    {
	      case PGS_S_SUCCESS:
		strncpy(startUTC,inputBuffer,10);
		startUTC[10] = '\0';
		strcpy(startUTC_print,startUTC);
		gotData = PGS_TRUE;
		break;
	      case PGSTD_M_ASCII_TIME_FMT_B:
		returnStatus = PGS_TD_ASCIItime_BtoA(inputBuffer,startUTC);
		if (returnStatus != PGS_S_SUCCESS)
		  continue;
		startUTC[10] = '\0';
		strcpy(startUTC_print,inputBuffer);
		startUTC_print[8] = '\0';
		gotData = PGS_TRUE;
		break;
	      default:
		PGS_SMF_GetMsg(&code,mnemonic,msg);
		printf("\n%s\n\n",msg);
		continue;
	    }
	}
	while (gotData == PGS_FALSE && ++numTries < MAX_TRIES);
	if (gotData == PGS_FALSE)
	  quit();

	gotData = PGS_FALSE;
	numTries = 0;
	
	/* get stop day, make sure stop day is not less than start day (the
	   default case is that the stop day is the same as the start day) */

	do
	{
	    printf("enter stop date [%.10s]:\n-->",startUTC_print);
	    fgets(inputBuffer,99,stdin);
	    if (inputBuffer[0] == 'q')
	      quit();
	    if (inputBuffer[0] == '\n')
	      strcpy(inputBuffer,startUTC_print);
	    newlinePtr = strchr(inputBuffer,'\n');
	    if (newlinePtr != NULL)
	      *newlinePtr = '\0';
	    returnStatus = PGS_TD_timeCheck(inputBuffer);
	    switch (returnStatus)
	    {
	      case PGS_S_SUCCESS:
		strncpy(stopUTC,inputBuffer,10);
		stopUTC[10] = '\0';
		strcpy(stopUTC_print,stopUTC);
		gotData = PGS_TRUE;
		break;
	      case PGSTD_M_ASCII_TIME_FMT_B:
		returnStatus = PGS_TD_ASCIItime_BtoA(inputBuffer,stopUTC);
		if (returnStatus != PGS_S_SUCCESS)
		  continue;
		stopUTC[10] = '\0';
		strcpy(stopUTC_print,inputBuffer);
		stopUTC_print[8] = '\0';
		gotData = PGS_TRUE;
		break;
	      default:
		PGS_SMF_GetMsg(&code,mnemonic,msg);
		printf("\n%s\n\n",msg);
		continue;
	    }
	    if (strcmp(stopUTC,startUTC) < 0)
	    {
		printf("\nstop time must be greater than or equal "
		       "to start time\n\n");
		gotData = PGS_FALSE;
	    }
	}
	while (gotData == PGS_FALSE && ++numTries < MAX_TRIES);
	if (gotData == PGS_FALSE)
	  quit();

	gotData = PGS_FALSE;
	numTries = 0;
	
	/* get time interval */

	do
	{
	    printf("enter time interval in seconds [%f sec]:\n"
		   "-->",(double) defaultInterval);
	    fgets(inputBuffer,99,stdin);
	    if (inputBuffer[0] == 'q')
	      quit();
	    if (inputBuffer[0] != '\n')
	    {
		numCheck = sscanf(inputBuffer,"%lf",&tempDouble);
		if (numCheck == 0)
		  continue;
		defaultInterval = tempDouble;
	    }
	    
	    gotData = PGS_TRUE;
	}
	while (gotData == PGS_FALSE && ++numTries < MAX_TRIES);
	if (gotData == PGS_FALSE)
	  quit();
	
	/* Calculate total size of data files to be generated.  The total size
	   of the data files generated will be approximately equal to the
	   (number of days requested)*(seconds per day)/(data time interval)*
	   (size in bytes of each record). */

	returnStatus = PGS_TD_UTCtoTAI(startUTC,&startTAI93);
	if (returnStatus == PGS_E_TOOLKIT)
	{
	    PGS_TD_UTCtoTAI(startUTC,&startTAI93);
	}
	
	PGS_TD_UTCtoTAI(stopUTC,&stopTAI93);
	stopTAI93 += 86400.0;
	dataBytes = (stopTAI93 - startTAI93 + 86400.)/
	            defaultInterval*sizeof(PGSt_ephemRecord); 
	dataBytes += (stopTAI93 - startTAI93 + 86400.)/
	             defaultInterval*sizeof(PGSt_attitRecord); 

	/* Print out the start and stop days, time interval and the
	   corresponding total size (in MB) of the files this will generate.
	   This is here because these files can be quite big. */

	printf("\nstart day:    %s\nstop day:     %s\ntime interval: %.4f"
	       " seconds\n\n",startUTC_print, stopUTC_print,
	       (double) defaultInterval); 
	printf("This will create approximately %.2f MB of data.\n",
	       (double) (dataBytes/1.e6));

	gotData = PGS_FALSE;
	numTries = 0;

	/* Verify start and stop time and time interval input.  This is to give
	   the user a chance to alter the input values if the size of the files
	   to be generated is larger than anticipated. */

	do
	{
	    printf("accept ([y]/n)?\n-->");
	    fgets(inputBuffer,99,stdin);
	    switch (inputBuffer[0])
	    {
	      case '\n':
	      case 'y':
	      case 'Y':
		gotData = PGS_TRUE;
		numTries = 2*MAX_TRIES;
		break;
	      case 'n':
	      case 'N':
		numTries = 2*MAX_TRIES;
		break;
	      case 'q':
		quit();
	    }
	}
	while (numTries++ < MAX_TRIES);

    } /* end do: GET TIMES */
    while (gotData == PGS_FALSE);

    /* get attitude noise */
    if (firstTime == PGS_TRUE)
      printf("\nYou may introduce random noise in the attitude and attitude\n"
	     "rates.  This noise will be deviations from the nominal values\n"
	     "of zero for these quantities (i.e. spacecraft reference frame\n"
	     "identical to orbital reference frame).  The number entered will\n"
	     "be the maximum deviation (+/-) from the nominal values.  The same"
	     "\nvalue will be used for all the components of a particular\n"
	     "quantity.  Enter 'N' at the first prompt for no noise at all.\n"
	     "Otherwise entering zero for a particular quantity will preclude\n"
	     "noise in that state.  The default case is no noise.\n"
	     "Enter noise level in arcseconds (0.00-999.99).\n\n");
    else
      printf("\nEnter noise level in arcseconds (0.00-999.99) or N for no"
	     " noise.\n\n");
    /* get attitude noise */

    numTries = 0;
    gotData = PGS_FALSE;
    do
    {
	printf("enter attitude noise level [N]:\n-->");
	fgets(inputBuffer,99,stdin);
	switch (inputBuffer[0])
	{
	  case '\n':
	  case 'n':
	  case 'N':
	    strcpy(attitudeNoise,"00000000000");
            yprNoise = 0.0;
	    gotData = PGS_TRUE;
	    break;
	  case 'q':
	    quit();
	  default:
	    attitudeNoise[0] = 'N';
	    numCheck = sscanf(inputBuffer,"%lf",&tempDouble);
	    if (numCheck == 1)
		if ( (tempDouble >= 0.0) && (tempDouble <= 999.99) )
		{
		    gotData = PGS_TRUE;
		    yprNoise = (PGSt_double) tempDouble;
		}
	}
    }
    while (gotData == PGS_FALSE && ++numTries < MAX_TRIES);
    if (gotData == PGS_FALSE)
      quit();
    
    /* get attitude rate noise */

    numTries = 0;
    gotData = PGS_FALSE;
    do
    {
	if (attitudeNoise[0] != 'N')
	{
	    gotData = PGS_TRUE;
	    yprRateNoise = 0.0;
	    continue;
	}
	printf("enter attitude rate noise level [%.2f]:\n-->",
	       (double) yprNoise);
	fgets(inputBuffer,99,stdin);
	if (inputBuffer[0] == 'q')
	  quit();
	if (inputBuffer[0] == '\n')
	  sprintf(inputBuffer,"%f",(double) yprNoise);
	numCheck = sscanf(inputBuffer,"%lf",&tempDouble);
	if (numCheck == 1)
	    if ( (tempDouble >= 0.0) && (tempDouble <= 999.99) )
	    {
		gotData = PGS_TRUE;
		yprRateNoise = (PGSt_double) tempDouble;
		sprintf(attitudeNoise+1,"%05d%05d",
			(int)(yprNoise*100.),
			(int)(yprRateNoise*100.));
	    }
    }
    while (gotData == PGS_FALSE && numTries++ < MAX_TRIES);
    if (gotData == PGS_FALSE)
      quit();
    
    /* Get default directory from the process control file.  This file
       associates a physical file name with the logical identifier 
       PGSd_SC_EPHEM_DATA (in this case it is actually a directory name).  This
       is where the SDP (aka PGS) toolkit spacecraft ephemeris access tool 
       (PGS_EPH_EphemAttit()) expects to find s/c ephemeris files. */
    
    if (firstTime == PGS_TRUE)
    {
	version = 1;
	ephemDirectory[0] = '\0';
	returnStatus = PGS_PC_GetReference(10601,&version,
					   ephemDirectory);
	charPtr = strrchr(ephemDirectory, 'C');
	if (charPtr != NULL)
	{
	    *charPtr = '\0';
	    strcat(ephemDirectory, "EPH");
	}
	printf("\n");
	if (returnStatus != PGS_S_SUCCESS || charPtr == NULL)
	{
	    printf("Unable to determine default ephemeris file location.\n"
		   "Setting default output location to current directory.\n\n");
	    strcpy(ephemDirectory,".");
	}
	else    
	  printf("By default this program will install the files it generates "
		 "in: %s.\n\n",ephemDirectory);
    }
    else 
      printf("\n");
    
    numTries = 0;
    gotData = PGS_FALSE;
    
    /* get install directory, make sure that the specified directory can be
       seen by the user */

    do
    {
	printf("install files in [%s]:\n-->",ephemDirectory);
	fgets(inputBuffer,99,stdin);
	if (inputBuffer[0] == 'q' && strlen(inputBuffer) == 2)
	  quit();
	if (inputBuffer[0] != '\n')
	{
	    newlinePtr = strchr(inputBuffer,'\n');
	    if (newlinePtr != NULL)
	      *newlinePtr = '\0';
	}
	else
	  strcpy(inputBuffer,ephemDirectory);
	
	directoryPtr = opendir(inputBuffer);
	if (directoryPtr == NULL)
	  printf("\nunable to open directory: %s\n\n",inputBuffer);
	else
	{
	    closedir(directoryPtr);
	    strcpy(ephemDirectory,inputBuffer);
	    gotData = PGS_TRUE;
	}
    }
    while (gotData == PGS_FALSE && ++numTries < MAX_TRIES);

    /* this section is executed only if we are doing attitude customization
       this feature is NOT an advertised one and therefore is likely only to be
       executed in house, at least until PDN tells everyone they can use it */

    if (custom == PGS_TRUE)
    {
	memset((void*)ca, (unsigned char)0, sizeof(PGSt_customAttit));
	printf("\nYou are now in the CUSTOM zone.  You may customize the "
	       "attitude of\nyour spacecraft here.  You will be prompted "
	       "for a start time, a stop\ntime, a starting set of euler "
	       "angles and an ending set of euler\nangles.  You will then "
	       "be asked if you would like to specify another\ncustom "
	       "period, at which point if you answer YES then the above "
	       "prompts\nwill be repeated.\n\nFor each period enter the "
	       "start time and the stop time of the period.\nThese times "
	       "need not fall within the time span of your current file\n"
	       "generation request but if there is no overlap at all then "
	       "you will get\nthe default file (i.e. you will have wasted a "
	       "lot of FLOPS and your\ntime).  Then at the first prompt for "
	       "euler angles enter the values of\nthe euler angles in the "
	       "order indicated (by the prompt) as three\nfloating point "
	       "numbers separated by spaces (on one line) (or\n'continue', "
	       "see below).  The simulator will force the the s/c attitude\n"
	       "to these values at the start time for this period.  At the "
	       "second\nprompt you may enter either another set of euler "
	       "angles OR the word\n'lock'.  If you enter 'lock', then the "
	       "s/c will enter an inertial lock\nmode and maintain its "
	       "current attitude relative to ECI.  If you enter\nanother "
	       "set of euler angles the spacecraft will rotate about a "
	       "fixed\naxis from the position defined by the first set of "
	       "euler angles at the\nstart time to the position defined by "
	       "the second set of euler angles\nat the end time.\n\nFor "
	       "each NEW period the s/c attitude will be set according to "
	       "the\nvalue of the starting Euler angles.  If you want to "
	       "ensure continuity,\nenter 'continue' at the first prompt.\n"
	       );

	count = 0;
	periodNum = -1;
	old_eea[0] = 0.0;
	old_eea[1] = 0.0;
	old_eea[2] = 0.0;

	do
	{
	    /* get start day */
	    
	    gotData = PGS_FALSE;

	    do
	    {
		printf("\nenter period start time (in CCSDS ASCII format):\n"
		       "-->");
		fgets(inputBuffer,99,stdin);
		if (inputBuffer[0] == 'q')
		  quit();
		newlinePtr = strchr(inputBuffer,'\n');
		if (newlinePtr != NULL)
		  *newlinePtr = '\0';
		returnStatus = PGS_TD_UTCtoTAI(inputBuffer, &ca[count].btai);
		switch (returnStatus)
		{
		  case PGS_S_SUCCESS:
		  case PGSTD_E_NO_LEAP_SECS:
		    if (ca[count].btai > stopTAI93)
		    {
			printf("\nInvalid period start time (start time is "
			       "AFTER ending time of this data set).\n");
			break;
		    }
		    gotData = PGS_TRUE;
		    break;
		  default:
		    PGS_SMF_GetMsg(&code,mnemonic,msg);
		    printf("\n%s\n\n",msg);
		    continue;
		}
	    }
	    while (gotData == PGS_FALSE && ++numTries < MAX_TRIES);
	    if (gotData == PGS_FALSE)
	      quit();

	    gotData = PGS_FALSE;

	    do
	    {
		printf("\nenter period stop time (in CCSDS ASCII format):\n"
		       "-->");
		fgets(inputBuffer,99,stdin);
		if (inputBuffer[0] == 'q')
		  quit();
		newlinePtr = strchr(inputBuffer,'\n');
		if (newlinePtr != NULL)
		  *newlinePtr = '\0';
		returnStatus = PGS_TD_UTCtoTAI(inputBuffer, &ca[count].etai);
		switch (returnStatus)
		{
		  case PGS_S_SUCCESS:
		  case PGSTD_E_NO_LEAP_SECS:
		    if (ca[count].etai < startTAI93 ||
			ca[count].etai < ca[count].btai)
		    {
			printf("\nInvalid period stop time (stop time is "
			       "BEFORE start time of this\nperiod OR this "
			       "data set).\n");
			break;
		    }
		    gotData = PGS_TRUE;
		    break;
		  default:
		    PGS_SMF_GetMsg(&code,mnemonic,msg);
		    printf("\n%s\n\n",msg);
		    continue;
		}
	    }
	    while (gotData == PGS_FALSE && ++numTries < MAX_TRIES);
	    if (gotData == PGS_FALSE)
	      quit();
	    
	    gotData = PGS_FALSE;
	    
	    
	    printf("\nenter starting Euler angles (%1d %1d %1d) or "
		   "'continue':\n-->",
		   eao[0], eao[1], eao[2]);
	    fgets(inputBuffer,99,stdin);
	    if (inputBuffer[0] == 'q')
	      quit();
	    if (strncmp(inputBuffer, "continue", 8) == 0 ||
		strncmp(inputBuffer, "CONTINUE", 8) == 0 ||
		strncmp(inputBuffer, "Continue", 8) == 0)
	    {
		ca[count].cont = PGS_TRUE;
	    }
	    else
	    {
		ca[count].cont = PGS_FALSE;
		sscanf(inputBuffer, "%lf %lf %lf", &ca[count].bea[0],
		       &ca[count].bea[1],&ca[count].bea[2]); 
	    }
	    
	    printf("\nenter ending Euler angles (%1d %1d %1d) or 'lock':\n-->",
		   eao[0], eao[1], eao[2]);
	    fgets(inputBuffer,99,stdin);
	    if (inputBuffer[0] == 'q')
	      quit();
	    if (strncmp(inputBuffer, "lock", 4) == 0 ||
		strncmp(inputBuffer, "LOCK", 4) == 0 ||
		strncmp(inputBuffer, "Lock", 4) == 0 ||
		strncmp(inputBuffer, "inertial", 8) == 0 ||
		strncmp(inputBuffer, "INERTIAL", 8) == 0 ||
		strncmp(inputBuffer, "Inertial", 8) == 0)
	    {
		ca[count].lock = PGS_TRUE;
		ca[count].eea[0] = 0.0;
		ca[count].eea[1] = 0.0;
		ca[count].eea[2] = 0.0;
		
	    }
	    else
	    {
		ca[count].lock = PGS_FALSE;
		sscanf(inputBuffer, "%lf %lf %lf", &ca[count].eea[0],
		       &ca[count].eea[1],&ca[count].eea[2]);
	    }

	    count++;
	    
	    if (ca[0].etai >= startTAI93)
	    {
		memcpy((void*)&tmp_ca, (void*)&ca[0], sizeof(PGSt_customAttit));
		old_eea[0] = tmp_ca.eea[0];
		old_eea[1] = tmp_ca.eea[1];
		old_eea[2] = tmp_ca.eea[2];
		if (tmp_ca.cont == PGS_TRUE)
		{
		    tmp_ca.bea[0] = 0.0;
		    tmp_ca.bea[1] = 0.0;
		    tmp_ca.bea[2] = 0.0;
		}
		periodNum = 0;
	    }

	    printf("\nDefine another period? (y/[n]):\n-->");
	    fgets(inputBuffer,99,stdin);
	    switch (inputBuffer[0])
	    {
	      case 'Y':
	      case 'y':
		continue;

	      case 'q':
	      case 'Q':
		quit();
		
	      default:
		break;
	    }
	    break;
	} while (count < 100);
    }
    
    num_periods = count;
    
    printf("\n");

    /* convert the start and stop time input by the user to Julian days for
       looping purposes */
    
    numCheck = sscanf(startUTC,"%d-%d-%d",&year,&month,&day);
    startJD = PGS_TD_julday(year,month,day);
    numCheck = sscanf(stopUTC,"%d-%d-%d",&year,&month,&day);
    stopJD = PGS_TD_julday(year,month,day);
    
    /* intialize some constant values in the s/c ephem/attit structures */

    strcpy(ephemFileHeader.spacecraftID, scName);
    strcpy(ephemFileHeader.source, "simulation");
    strcpy(ephemFileHeader.version, "1");
    strcpy(ephemFileHeader.keplerRefFrame, "J2000");
    ephemFileHeader.nURs = 0;
    
    strcpy(attitFileHeader.spacecraftID, scName);
    strcpy(attitFileHeader.source, "simulation");
    strcpy(attitFileHeader.version, "1");
    attitFileHeader.nURs = 0;
    attitFileHeader.eulerAngleOrder[0] = (PGSt_uinteger) eao[0];
    attitFileHeader.eulerAngleOrder[1] = (PGSt_uinteger) eao[1];
    attitFileHeader.eulerAngleOrder[2] = (PGSt_uinteger) eao[2];
    
    ephemRecord.qualityFlag = 0;
    attitRecord.qualityFlag = 0;

    /* loop from start day to stop day (inclusive), writing out an entire days
       worth of data for each day */

    for(julianDay = startJD;julianDay<=stopJD;julianDay++)
    {
	/* create the CCSDS ASCII version of the day corresponding to the
	   current Julian day and the next Julian day */

	PGS_TD_calday(julianDay,&year,&month,&day);
	sprintf(startUTC,"%04d-%02d-%02d", year, month, day);
	PGS_TD_calday(julianDay+1,&year,&month,&day);
	sprintf(stopUTC,"%04d-%02d-%02d", year, month, day);

	/* Create a file name of the form: <spacecraft>_<date>.eph.  Test to see
	   it this file already exists.  If it does do not overwrite the old
	   data. */

	sprintf(ephFilename,"%s/%s_%s.eph", ephemDirectory, scName, startUTC);
	if (file_exists(ephFilename))
	{
	    printf("file: %s already exists ...skipping\n",ephFilename);
	    continue;
	}

	/* Create a file name of the form: <spacecraft>_<date>.att.  Test to see
	   it this file already exists.  If it does do not overwrite the old
	   data. */

	sprintf(attFilename,"%s/%s_%s.att", ephemDirectory, scName, startUTC);
	if (file_exists(attFilename))
	{
	    printf("file: %s already exists ...skipping\n",attFilename);
	    continue;
	}

	/* Attempt to create the file for writing.  If there is a problem
	   opening the file (probably no write permission in specified
	   directory) issue a message and wait for a response.  This gives the
	   user a chance to identify the problem and fix it.  The user may quit
	   at this point or continue.  If the user wishes to continue attempt to
	   open the file once again.  If it again fails skip it without issuing
	   another warning. */

	ephemFilePtr = fopen(ephFilename,"w");
	if (ephemFilePtr == NULL)
	{
	    printf("\nunable to create file: %s\n",ephFilename);
	    printf("make sure that you have write permission in directory: %s\n"
		   ,ephemDirectory);
	    printf("continue (y/n)?\n-->");
	    fgets(inputBuffer,99,stdin);
	    printf("\n");
	    switch (inputBuffer[0])
	    {
	      case 'n':
	      case 'N':
	      case 'q':
		quit();
	    }
	    ephemFilePtr = fopen(ephFilename,"w");
	    if (ephemFilePtr == NULL)
	      continue;
	}
	printf("creating file: %s\n",ephFilename);
	      
	attitFilePtr = fopen(attFilename,"w");
	if (attitFilePtr == NULL)
	{
	    printf("\nunable to create file: %s\n",attFilename);
	    printf("make sure that you have write permission in directory: %s\n"
		   ,ephemDirectory);
	    printf("continue (y/n)?\n-->");
	    fgets(inputBuffer,99,stdin);
	    printf("\n");
	    switch (inputBuffer[0])
	    {
	      case 'n':
	      case 'N':
	      case 'q':
		quit();
	    }
	    attitFilePtr = fopen(attFilename,"w");
	    if (attitFilePtr == NULL)
	    {
		fclose(ephemFilePtr);
		continue;
	    }
	}
	printf("creating file: %s\n",attFilename);
	      
	/* calculate first record time and last record time */

	PGS_TD_UTCtoTAI(startUTC,&startTAI93);
	PGS_TD_UTCtoTAI(stopUTC,&stopTAI93);
	ephemFileHeader.startTime = startTAI93;
	attitFileHeader.startTime = startTAI93;
	PGS_TD_TAItoUTC(startTAI93, startUTC);
	
	/* calculate number of records of ephemeris data */

	numIntervals = (stopTAI93 - startTAI93 - 1.e-6)/
                       defaultInterval;
	numIntervals -= fmod(numIntervals,1.0);
	
	ephemFileHeader.endTime = startTAI93 + defaultInterval*
	                                         numIntervals;
	PGS_TD_TAItoUTC(ephemFileHeader.endTime, stopUTC);
	sprintf(ephemFileHeader.asciiTimeRange,
		"%.22s - %.22s",startUTC, stopUTC);
	
	ephemFileHeader.nRecords = (PGSt_integer) numIntervals + 1;
	
	/* calculate number of records of attitude data */

	numIntervals = (stopTAI93 - startTAI93 - 1.e-6)/
                       defaultInterval;
	numIntervals = (int) numIntervals; /*-= fmod(numIntervals,1.0);*/
	
	attitFileHeader.endTime = startTAI93 + defaultInterval*
	                                         numIntervals;
	PGS_TD_TAItoUTC(attitFileHeader.endTime, stopUTC);
	sprintf(attitFileHeader.asciiTimeRange,
		"%.21sZ - %.21sZ",startUTC, stopUTC);

	attitFileHeader.nRecords = (PGSt_integer) numIntervals + 1;

	/* get the orbital elements (required for the ephemeris file header) */

	PGS_EPH_orbitalElements(startTAI93,spacecraftTag, data, &startOrbitNum);

	if (startOrbitNum < 0)
	{
	    /* the definition of orbit numbers in ephemeris file headers only
	       allows for positive orbit numbers, so make up some sillyness if
	       the orbit number comes back negative from the simulator call
	       above (which it will if the requested time is BEFORE launch) */

	    startOrbitNum = START_BEFORE_LAUNCH;
	}

	ephemFileHeader.orbitNumberStart = startOrbitNum;
	ephemFileHeader.orbitNumberEnd = startOrbitNum-1;

/****
	nu = ma+2.0*e*sin(ma)+1.25*e*e*sin(2.0*ma);
	if ( nu < 0.0 )
	{
	    nu = 2.0*M_PI + nu;
	}
	
	if ( ap > 0.0 )
	{
	    nu0 = 2.0*M_PI - ap;
	}
	else
	{
	    nu0 = -ap;
	}
	
	if ( nu0 > nu )
	{
	    k = 1.0;
	}
	else
	{
	    k = 0.0;
	}
	dt1 = P/(2.0*M_PI)*(2.0*k*M_PI + 
			   (E(nu) - e*sin(E(nu))) - 
			   (E(nu0) - e*sin(E(nu0))));
****/
	dt1 = startTAI93 - PGS_EPH_OrbitAscendTime(spacecraftTag, startTAI93);
	dt2 = startTAI93 - PGS_EPH_OrbitDecendTime(spacecraftTag, startTAI93);
	time = startTAI93 - dt1;
	while ( time < stopTAI93 )
	{
	    ephemFileHeader.orbitNumberEnd += 1;
	    time += 2.0*(M_PI/data[8]);
	}

	ephemFileHeader.keplerElements[0] = data[0];
	ephemFileHeader.keplerElements[1] = data[1];
	ephemFileHeader.keplerElements[2] = data[2];
	ephemFileHeader.keplerElements[3] = data[3];
	ephemFileHeader.keplerElements[4] = data[4];
	ephemFileHeader.keplerElements[5] = data[5];

	/* just some bogus junk (i.e. filling out the number of orbits field in
	   the ephemeris header with bogus values, just so there are at LEAST
	   some values in these fields so that they can accessed in a test
	   situation) */

	ephemFileHeader.nOrbits = 1 + ephemFileHeader.orbitNumberEnd - 
	  ephemFileHeader.orbitNumberStart;
	
	/* write ephemeris file header */

	ephemFileHeader.interval = (PGSt_real) defaultInterval;
	numCheck = (int) fwrite(&ephemFileHeader,sizeof(PGSt_ephemHeader),1,
				ephemFilePtr);
	if (numCheck != 1)
	{
	    printf("error writing to file: %s ...aborting\n",ephFilename);
	    fclose(ephemFilePtr);
	    fclose(attitFilePtr);
	    break;
	}
	
	numCheck = (int) fwrite(parentUR, PGSd_UR_FIELD_SIZE,
				ephemFileHeader.nURs, ephemFilePtr);
	if (numCheck != ephemFileHeader.nURs)
	{
	    printf("error writing to file: %s ...aborting\n",ephFilename);
	    fclose(ephemFilePtr);
	    fclose(attitFilePtr);
	    break;
	}

	/* write attitude file header */

	attitFileHeader.interval = (PGSt_real) defaultInterval;
	numCheck = (int) fwrite(&attitFileHeader,sizeof(PGSt_attitHeader),1,
				attitFilePtr);
	if (numCheck != 1)
	{
	    printf("error writing to file: %s ...aborting\n",attFilename);
	    fclose(ephemFilePtr);
	    fclose(attitFilePtr);
	    break;
	}
	
	numCheck = (int) fwrite(parentUR, PGSd_UR_FIELD_SIZE,
				attitFileHeader.nURs, attitFilePtr);
	if (numCheck != attitFileHeader.nURs)
	{
	    printf("error writing to file: %s ...aborting\n",attFilename);
	    fclose(ephemFilePtr);
	    fclose(attitFilePtr);
	    break;
	}
			  
	/* write ephemeris data */

	ephemRecord.secTAI93=ephemFileHeader.startTime;
	attitRecord.secTAI93=attitFileHeader.startTime;

	for (count=1;ephemRecord.secTAI93<stopTAI93;count++)
	{
	    /* call spacecraft attitude and orbit simulator */

	    returnStatus = PGS_EPH_attOrbSim(ephemRecord.secTAI93,
					     spacecraftTag,
					     attitudeNoise,
					     ephemRecord.position,
					     ephemRecord.velocity,
					     attitRecord.eulerAngle,
					     attitRecord.angularVelocity,
					     transform);

	    /************************
	     * BEGIN CUSTOM SECTION *
	     ************************/

	    if (custom == PGS_TRUE)
	    {
		if (attitRecord.secTAI93 > tmp_ca.etai)
		{
		    if (periodNum < num_periods-1)
		    {
			if (ca[periodNum+1].btai > attitRecord.secTAI93)
			{
			    if (tmp_ca.lock == PGS_TRUE)
			    {
				memcpy((void*)tmp_ca.bea, (void*)old_eea,
				       sizeof(old_eea));
				memcpy((void*)tmp_ca.eea, (void*)old_eea,
				       sizeof(old_eea));
			    }
			    else
			    {
				memcpy((void*)tmp_ca.bea, (void*)tmp_ca.eea,
				       sizeof(old_eea));
			    }
			    
			    tmp_ca.btai = tmp_ca.etai;
			    tmp_ca.etai = ca[periodNum+1].btai - 1.0E-6;
			    tmp_ca.lock = PGS_FALSE;
			    tmp_ca.cont = PGS_FALSE;
			}
			else
			{
			    periodNum++;
			    memcpy((void*)old_eea, (void*)tmp_ca.eea,
				   sizeof(old_eea));
			    memcpy((void*)&tmp_ca, (void*)&ca[periodNum], 
				   sizeof(PGSt_customAttit));
			    if (tmp_ca.cont == PGS_TRUE)
			    {
				memcpy((void*)tmp_ca.bea, (void*)old_eea,
				       sizeof(old_eea));
				tmp_ca.btai = old_tai;
			    }
			}
		    }
		    else
		    {
			tmp_ca.bea[0] = tmp_ca.eea[0];
			tmp_ca.bea[1] = tmp_ca.eea[1];
			tmp_ca.bea[2] = tmp_ca.eea[2];
			tmp_ca.btai = tmp_ca.etai;
			tmp_ca.etai = stopTAI93;
			tmp_ca.cont = PGS_FALSE;
			tmp_ca.lock = PGS_FALSE;
			custom = PGS_FALSE;
		    }
		    if (tmp_ca.lock == PGS_TRUE)
		    {
			
			PGS_CSC_EulerToQuat(tmp_ca.bea,
					    eao,quatEuler);
			PGS_CSC_getORBtoECIquat(ephemRecord.position,
						ephemRecord.velocity,
						quatORBtoECI);
			PGS_CSC_quatMultiply(quatORBtoECI, quatEuler,
					     quatSCtoECI);
			quatECItoSC_0[0] = quatSCtoECI[0];
			quatECItoSC_0[1] = -quatSCtoECI[1];
			quatECItoSC_0[2] = -quatSCtoECI[2];
			quatECItoSC_0[3] = -quatSCtoECI[3];
		    }
		    
		}
		if (tmp_ca.lock != PGS_TRUE)
		{
		    PGS_EPH_interpolateAttitude(tmp_ca.btai, tmp_ca.bea,
						attitRecord.angularVelocity,
						tmp_ca.etai, tmp_ca.eea,
						attitRecord.angularVelocity,
						eao, ephemRecord.secTAI93,
						eulerAngle, angularVelocity );
		    old_tai = ephemRecord.secTAI93;
		    old_eea[0] = eulerAngle[0];
		    old_eea[1] = eulerAngle[1];
		    old_eea[2] = eulerAngle[2];
		}
		else
		{
		    PGS_CSC_getORBtoECIquat(ephemRecord.position,
					    ephemRecord.velocity,
					    quatORBtoECI);
		    PGS_CSC_quatMultiply(quatECItoSC_0, quatORBtoECI,
					 quatORBtoSC_0);
		    quatEuler[0] = quatORBtoSC_0[0];
		    quatEuler[1] = -quatORBtoSC_0[1];
		    quatEuler[2] = -quatORBtoSC_0[2];
		    quatEuler[3] = -quatORBtoSC_0[3];
		    PGS_CSC_QuatToEuler(quatEuler, eao, eulerAngle);

		    old_tai = ephemRecord.secTAI93;
		    old_eea[0] = eulerAngle[0];
		    old_eea[1] = eulerAngle[1];
		    old_eea[2] = eulerAngle[2];
		    if (spacecraftTag == PGSd_TRMM)
		    {
			tipEuler(ephemRecord.position, ephemRecord.velocity,
				 eao, old_eea, PGSd_TO_TIP);
		    }
		}
		attitRecord.eulerAngle[0] += eulerAngle[0];
		attitRecord.eulerAngle[1] += eulerAngle[1];
		attitRecord.eulerAngle[2] += eulerAngle[2];
	    }

	    /**********************
	     * END CUSTOM SECTION *
	     **********************/

	    /* If the spacecraft is TRMM assume simulator ephemeris output is
	       TOD and simulator attitude output is geodetic referenced.  These
	       must be transformed to J2000 and geocentric referenced,
	       respectively. */

	    if (spacecraftTag == PGSd_TRMM)
	    {
	    
		if (custom == PGS_FALSE || tmp_ca.lock != PGS_TRUE)
		{
		    tipEuler(ephemRecord.position, ephemRecord.velocity, eao,
			     attitRecord.eulerAngle, PGSd_TO_ORB);
		}
		
	        /* Convert the True of Date position and velocity to the J2000
		   reference frame (this function--PGS_EPH_EphemAttit()--returns
		   ephemeris data in J2000 coordinates).  The "6" in the
		   function call below indicates that ephemRecord.position is a
		   6-vector.  This is a questionable trick, taking advantage of
		   the fact that ephemRecord.velocity is a three vector in the
		   structure ephemRecord immediately following the three vector
		   position (this is not an ideal programming technique, but it
		   should work). */

		PGS_CSC_TODtoJ2000(6,ephemRecord.secTAI93,
				   ephemRecord.position,
				   ephemRecord.position);
	    }
		
	    returnStatus = PGS_EPH_TransformBodyRates(
		                              PGSe_ORBtoECI,
					      ephemRecord.position,
					      ephemRecord.velocity,
					      attitRecord.eulerAngle,
					      (PGSt_integer*)attitFileHeader.eulerAngleOrder,
					      attitRecord.angularVelocity,
					      attitRecord.angularVelocity);

	    /* write ephemeris data to file, abort if any error occurs and print
	       warning message */

	    numCheck = (int) fwrite(&ephemRecord,sizeof(PGSt_ephemRecord),1,
				    ephemFilePtr);
	    if (numCheck != 1)
	    {
		printf("error writing to file: %s ...aborting\n",ephFilename);
		break;
	    }
	    
	    /* write attitude data to file, abort if any error occurs and print
	       warning message */

	    numCheck = (int) fwrite(&attitRecord,sizeof(PGSt_attitRecord),1,
				    attitFilePtr);
	    if (numCheck != 1)
	    {
		printf("error writing to file: %s ...aborting\n",attFilename);
		break;
	    }
	    
	    /* increment the time
	       NOTE: the time is calculated as (startTime)+(record #)*(interval)
	             because calculating it as (previous record time)+(interval)
		     causes errors, presumably due to machine precision
		     limitations */

	    ephemRecord.secTAI93 = startTAI93 + (defaultInterval*count);
	    attitRecord.secTAI93 = startTAI93 + (defaultInterval*count);
	}
	
	for (count=0;count<ephemFileHeader.nOrbits;count++)
	{
	    orbMetadata.orbitNumber = startOrbitNum + count;
	    
	    time = startTAI93 - dt1 + count*P;
	    PGS_EPH_orbSim( time, spacecraftTag, posvelECI, posvelECI+3);
	    while ( posvelECI[2] < 0.0 && count > 0 )
	    {
		time -= 300.0;
		PGS_EPH_orbSim( time, spacecraftTag, posvelECI, posvelECI+3);
	    }
	    time = PGS_EPH_OrbitAscendTime( spacecraftTag, time );
	    orbMetadata.orbitAscendTime = time;
	    
	    time = startTAI93 - dt2 + count*P;
	    time = PGS_EPH_OrbitDecendTime( spacecraftTag, time );
	    orbMetadata.orbitDescendTime = time;

	    PGS_TD_TAItoUTC(orbMetadata.orbitDescendTime, asciiUTC);

	    PGS_EPH_orbSim( orbMetadata.orbitDescendTime,
			    spacecraftTag,
			    posvelECI,
			    posvelECI+3);
	    PGS_CSC_ECItoECR(0, asciiUTC, NULL,
			     (PGSt_double(*)[6]) posvelECI, 
			     (PGSt_double(*)[6]) posvelECR);
	    PGS_CSC_ECRtoGEO(posvelECR, "WGS84", &longitude, &latitude,
			     &altitude);
	    orbMetadata.orbitDescendLongitude = longitude;
	    numCheck = (int) fwrite(&orbMetadata,sizeof(PGSt_ephemMetadata),1,
				    ephemFilePtr);
	    if (numCheck != 1)
	    {
		printf("error writing to file: %s ...aborting\n",ephFilename);
		break;
	    }
	}
	
	fclose(ephemFilePtr);
	fclose(attitFilePtr);
    } /* END: for (julianDay... */
    printf("\n");
    printf("Done.  Generate another set of ephemeris files (y/[n]):\n-->");
    fgets(inputBuffer,99,stdin);
    switch (inputBuffer[0])
    {
      case 'y':
      case 'Y':
	printf("\n\n\n");
	firstTime = PGS_FALSE;
	custom = PGS_FALSE;
	goto BEGINNING;
    }
    printf("\n");
}

static void quit(void)
{
    printf("\nquitting\nno data generated\n\n");
    exit(0);
}


static PGSt_boolean get_scTag(char *scNameStr, PGSt_tag *spacecraftTag)
{
    char  *spacecraftNameString;
    
    spacecraftNameString = scNameStr;
    *spacecraftTag = 0;
    
    /* allow for use of new PGSd_ prefix */

    if (strncmp(spacecraftNameString,"PGSd_",5) == 0)
	spacecraftNameString += 5;
    else if (strncmp(spacecraftNameString,"pgsd_",5) == 0)
	spacecraftNameString += 5;
    
    if (!strcmp(spacecraftNameString,"TRMM"))
    {
	*spacecraftTag = PGSd_TRMM;
	return PGS_TRUE;
    }
    if (!strcmp(spacecraftNameString,"EOS_AM"))
    {
	*spacecraftTag = PGSd_EOS_AM;
	return PGS_TRUE;
    }
    if (!strcmp(spacecraftNameString,"EOS_PM"))
    {
	*spacecraftTag = PGSd_EOS_PM;
	return PGS_TRUE;
    }
    
    /* if input is not actually supported test every case before returning since
       I don't feel like putting a test after each case */

    if (!strcmp(spacecraftNameString,"trmm"))
      *spacecraftTag = PGSd_TRMM;
    if (!strcmp(spacecraftNameString,"eos_am"))
      *spacecraftTag = PGSd_EOS_AM;
    if (!strcmp(spacecraftNameString,"EOS AM"))
      *spacecraftTag = PGSd_EOS_AM;
    if (!strcmp(spacecraftNameString,"EOS-AM"))
      *spacecraftTag = PGSd_EOS_AM;
    if (!strcmp(spacecraftNameString,"eos am"))
      *spacecraftTag = PGSd_EOS_AM;
    if (!strcmp(spacecraftNameString,"eos-am"))
      *spacecraftTag = PGSd_EOS_AM;
    if (!strcmp(spacecraftNameString,"EOSAM"))
      *spacecraftTag = PGSd_EOS_AM;
    if (!strcmp(spacecraftNameString,"eosam"))
      *spacecraftTag = PGSd_EOS_AM;
    if (!strcmp(spacecraftNameString,"eos_pm"))
      *spacecraftTag = PGSd_EOS_PM;
    if (!strcmp(spacecraftNameString,"EOS PM"))
      *spacecraftTag = PGSd_EOS_PM;
    if (!strcmp(spacecraftNameString,"EOS-PM"))
      *spacecraftTag = PGSd_EOS_PM;
    if (!strcmp(spacecraftNameString,"eos pm"))
      *spacecraftTag = PGSd_EOS_PM;
    if (!strcmp(spacecraftNameString,"eos-pm"))
      *spacecraftTag = PGSd_EOS_PM;
    if (!strcmp(spacecraftNameString,"EOSPM"))
      *spacecraftTag = PGSd_EOS_PM;
    if (!strcmp(spacecraftNameString,"eospm"))
      *spacecraftTag = PGSd_EOS_PM;

    if (!strcmp(spacecraftNameString,"custom"))
      *spacecraftTag = PGSd_CUSTOM;
    if (!strcmp(spacecraftNameString,"CUSTOM"))
      *spacecraftTag = PGSd_CUSTOM;

    if (*spacecraftTag)
      return PGS_TRUE;
    else
      return PGS_FALSE;
}


/*
 * function: file_split_path()
 * author:   mike sucher
 * purpose:  split path prefix from file name
 * arguments:
 *     char *name;  input:  full name of file including path
 *     char *path;  output: the path, split from the filename
 *     char *fname; output: the filename, split from the path
 * returns:
 *     (1) Success
 * notes:
 *     the path separator is the slash ('/') character, thus this
 *     routine is only intended for Unix filesystems
 */
static int file_split_path(char *name, char *path, char *fname )
{
    char c, *p1, *p2, *q;

    p1 = p2 = name;
        
    while(1)
    {
	for(q=p2; (*q != '/') && *q; q++);
	if(*q) p2 = q+1;
	else break;
    }
    
    strcpy(fname, p2);
 
    if(p2 > p1)
    {
	c = *(p2-1); 
	*(p2-1) = 0;
	strcpy(path, p1);
	*(p2-1) = c; 
    }
    else path[0] = 0;
        
    return 1;
}

/*
 * function: file_exists()
 * author:   mike sucher
 * purpose:  check to see if file exists
 * arguments:
 *     char *name; name, including path, of the file to check
 * returns:
 *     FOUND     (1) file does exist
 *     NOT_FOUND (0) file does not exist
 */

#define FOUND 1
#define NOT_FOUND 0

static int file_exists(char *name)
{
    DIR *dirp;
    struct dirent *dp;
    char fname[128], path[128];

    file_split_path(name,path,fname);
    if(strlen(path) != 0) dirp = opendir(path);
    else dirp = opendir( "." );
    if (dirp == NULL)
      return NOT_FOUND;
    while ( (dp = readdir( dirp )) != NULL )
    if( strcmp( dp->d_name, fname ) == 0 )
    {
	closedir(dirp);
	return FOUND;
    }
    closedir(dirp);
    return NOT_FOUND;
}

static void
tipEuler(
    PGSt_double   position[3],
    PGSt_double   velocity[3],
    PGSt_integer  eulerAngleOrder[3],
    PGSt_double   eulerAngles[3],
    PGSt_integer  direction)
{
    PGSt_double   quatEuler[4];
    PGSt_double   quatTIPtoORB[4];
    PGSt_double   *quatORBtoTIP;
    PGSt_double   *quatSCtoORB;
    PGSt_double   *quatSCtoTIP;
    
    quatORBtoTIP = quatTIPtoORB;
    quatSCtoORB = quatEuler;
    quatSCtoTIP = quatEuler;
    
    /* Get the quaternion that gives the equivalent rotation as the s/c Euler
       angles. */
    
    PGS_CSC_EulerToQuat(eulerAngles,
			eulerAngleOrder, quatEuler);
    
    /* The numbers 6378137.0 and 6356752.314245 below are the Earth's equatorial
       and polar radii, respectively.  Note that PGS_CSC_TiltYaw() MUST be
       called with True of Date position and velocity.  TRMM ephemeris elements
       are True of Date. */
    
    PGS_CSC_TiltYaw(position, velocity, 6378137.0, 6356752.314245,
		    quatTIPtoORB);
    
    if (direction == PGSd_TO_TIP)
    {
	/* The quaternion defining the rotation from the TIP to SC reference
	   frame is determined by the successive rotations: SC to ORB => ORB to
	   TIP. */
	
	/*  The ORB to TIP quaternion is the conjugate of the TIP to ORB
	    quaternion. */

	quatORBtoTIP[1] = -quatTIPtoORB[1];
	quatORBtoTIP[2] = -quatTIPtoORB[2];
	quatORBtoTIP[3] = -quatTIPtoORB[3];

	/* quatEuler is now quatSCtoTIP */

	PGS_CSC_quatMultiply(quatORBtoTIP, quatSCtoORB, quatEuler);
    }
    else
    {
	/* The quaternion defining the rotation from the SC to ORB reference
	   frame is determined by the successive rotations: SC to TIP => TIP to
	   ORB. */
	
	/* quatEuler is now quatSCtoORB */

	PGS_CSC_quatMultiply(quatTIPtoORB, quatSCtoTIP, quatEuler);
    }
    
    /* finally convert the SC to ORB (or SC to TIP) quaternion back to a set of
       Euler angles */
    
    PGS_CSC_QuatToEuler(quatEuler, eulerAngleOrder, eulerAngles);

}

static PGSt_double
PGS_EPH_OrbitAscendTime(
    PGSt_tag              spacecraftTag,
    PGSt_double           startTAI93)
{
    PGSt_double           posECI[3];
    PGSt_double           velECI[3];
    PGSt_double*          zPos;
    PGSt_double           deltaT;
    PGSt_double           t1;
    PGSt_double           t2;
    
    zPos = posECI+2;
    t1 = startTAI93;
    deltaT = 300.0;
    
    PGS_EPH_orbSim( t1, spacecraftTag, posECI, velECI );
    while ( *zPos < 0.0 )
    {
	t1 -= deltaT;
	PGS_EPH_orbSim( t1, spacecraftTag, posECI, velECI );
    }
    
    t2 = t1 - deltaT;

    while ( deltaT > 0.001 )
    {
	PGS_EPH_orbSim( t2, spacecraftTag, posECI, velECI );
	if ( *zPos < 0.0 )
	{
	    deltaT = deltaT/2.0;
	    t2 = t1 - deltaT;
	}
	else
	{
	    t1 = t2;
	    t2 = t2 - deltaT;
	}
    }

    return (t1 + deltaT);
}

static PGSt_double
PGS_EPH_OrbitDecendTime(
    PGSt_tag              spacecraftTag,
    PGSt_double           startTAI93)
{
    PGSt_double           posECI[3];
    PGSt_double           velECI[3];
    PGSt_double*          zPos;
    PGSt_double           deltaT;
    PGSt_double           t1;
    PGSt_double           t2;
    
    zPos = posECI+2;
    t1 = startTAI93;
    deltaT = 300.0;
    
    PGS_EPH_orbSim( t1, spacecraftTag, posECI, velECI );
    while ( *zPos < 0.0 )
    {
	t1 -= deltaT;
	PGS_EPH_orbSim( t1, spacecraftTag, posECI, velECI );
    }
    
    t2 = t1 + deltaT;

    while ( deltaT > 0.001 )
    {
	PGS_EPH_orbSim( t2, spacecraftTag, posECI, velECI );
	if ( *zPos < 0.0 )
	{
	    deltaT = deltaT/2.0;
	    t2 = t1 + deltaT;
	}
	else
	{
	    t1 = t2;
	    t2 = t2 + deltaT;
	}
    }

    return (t1 + deltaT);
}
