/*-------------------------------------------------------------------------*/
/*                                                                         */
/*  COPYRIGHT[copyright mark] 1996, Hughes Aircraft Company, its vendors,  */
/*  and suppliers.  ALL RIGHTS RESERVED.                                   */
/*                                                                         */
/*-------------------------------------------------------------------------*/
/*******************************************************************************
BEGIN_FILE_PROLOG:

FILENAME:   
   PGS_MEM_ShmWriteF.c

DESCRIPTION:
   This file contains the function PGS_MEM_ShmWriteF().
   This function copies the contents of a user allocated (may be dynamically or
   statically allocated) memory area into shared memory.
 
AUTHOR:
   Guru Tej S. Khalsa /	Applied Research Corp.

HISTORY:
   26-Mar-1994  GTSK  Initial version

END_FILE_PROLOG:
*******************************************************************************/
/*******************************************************************************
BEGIN_PROLOG:

TITLE:  
    Write To Shared Memory

NAME:   
    PGS_MEM_ShmWriteF

SYNOPSIS:       
C:  
    #include <PGS_MEM1.h>                                                  

    PGSt_SMF_status
    PGS_MEM_ShmWriteF(
        void*        mem_ptr,
	PGSt_integer size);

FORTRAN:
      include 'PGS_SMF.f'
      include 'PGS_MEM_9.f'

      integer function pgs_mem_shmwrite(mem_ptr, size)
      integer   size
      character mem_ptr(size)

DESCRIPTION:
      This function copies the contents of a user allocated (may be dynamically
      or statically allocated) memory area into shared memory.  This function is
      is meant to be used by FORTRAN (77/90) users who cannot take advantage of
      the C shared memory tool PGS_MEM_ShmAttach() and PGS_MEM_ShmDetach().

INPUTS:        
    NAME         DESCRIPTION
    ----         -----------
    mem_ptr      pointer to memory area the contents of which are to be written
                 shared memory

    size         size (in bytes) of the memory area to copy into shared memory

OUTPUTS:        
    NONE

RETURNS:
    PGS_S_SUCCESS
    PGS_E_UNIX
    PGSMEM_E_SHM_ENV
    PGSMEM_E_SHM_NOTCREATE
    PGSMEM_E_SHM_MULTIATTACH    
    PGSMEM_E_SHM_NOTATTACH    

EXAMPLES:
 FORTRAN:
      integer    pgs_mem_shmwrite
      integer    size
      character  shm_buffer(1000)
      integer    returnstatus

!  fill shm_buffer with interesting data...

      returnstatus = pgs_mem_shmwrite(shm_buffer, size)
      if (returnstatus .ne. pgs_s_success) goto 999

!  the contents of shm_buffer have been written to the shared memory
!  area which can be accessed by a subsequent process

 999  continue  ! process errors conditions

NOTES:
    This tool is meant to be used by FORTRAN (77/90) users ONLY.  C users should
    use the functions PGS_MEM_ShmAttach() and PGS_MEM_ShmDetach().

    This tool is not part of POSIX and is subjected to change when the POSIX.4
    implementation becomes available. This function will only detach the shared
    memory segment from the process. The shared memory segment will not be
    removed from the system by calling this tool; therefore one can re-attach it
    again.

REQUIREMENTS:
    PGSTK-1241

DETAILS:
    The user passes in a pointer to a user defined memory area (an area of
    memory which has been either statically or dynamically allocated by the
    user) and the size of that area.  This function will retrieve the pointer
    to the shared memory area and write the contents of the users memory area
    to the shared memory area OVERWRITING whatever was previously in the shared
    memory area.  This function will then detach the shared memory from the
    current process.  Before exiting from the PGE, the system will make sure
    that the attached shared memory segment will be removed from the system.

GLOBALS:
    NONE

FILES:
    NONE

FUNCTIONS_CALLED:   
    PGS_MEM_ShmAttach()
    PGS_MEM_ShmDetach()

END_PROLOG:
*******************************************************************************/

#include <PGS_MEM1.h>

PGSt_SMF_status
PGS_MEM_ShmWriteF(                    /* write to shared memory */
    void             *mem_ptr,        /* memory area to write to */
    PGSt_integer     size)            /* size of memory area to write to */
{
    void             *shared_mem_ptr; /* pointer ot shared memory */
    PGSt_SMF_status  returnStatus;    /* return value of this function */

    /* retrieve pointer to shared memory */

    returnStatus = PGS_MEM_ShmAttach(&shared_mem_ptr);
    if (returnStatus == PGS_S_SUCCESS)
    {
	/* copy contents of input buffer to shared memory and then detach shared
	   memory */

	memcpy(shared_mem_ptr, mem_ptr, size);
	returnStatus = PGS_MEM_ShmDetach();
    }

    return returnStatus;
}
