/*-------------------------------------------------------------------------*/
/*                                                                         */
/*  COPYRIGHT[copyright mark] 1996, Hughes Aircraft Company, its vendors,  */
/*  and suppliers.  ALL RIGHTS RESERVED.                                   */
/*                                                                         */
/*-------------------------------------------------------------------------*/
/*******************************************************************************
BEGIN_FILE_PROLOG:

FILENAME:   
        PGS_MET_Write
 
DESCRIPTION:
	Enables the user to write different master groups of metadata to separate
	HDF attributes. Whole of the MCF can also be written to a separate ascii file. 
	Attributes can only be read back using PGS_MET_GetPCAttr tool.
	
AUTHOR: 
        Alward N. Siyyid/ EOSL
        Carol S. W. Tsai / Applied Reseach Corporation

HISTORY:
        18-MAY-95      ANS     Initial version
	01-JUN-95      ANS     Code inspection comments update
	13-July-95     ANS     Improved Fortran example
	20-July-95	ANS	Fixed bug ECSed01021
	21-July-95	ANS     Cleaned up fix for ECSed01021
	04-MAR-96	ANS	Updated for tk5+
	19-Aug-96       ANS     added fclose for the separate ascii file
	                        HDF attributes more than MAX_ORDER can now be
				written.
        06-Feb-97      CSWT    added PGS_PC_GetConfigData and
                               PGS_PC_GetReference to make use of the ASCII
                               metadata files the TOOLKIT creates for non
                               HDF-EOS product output files.

        14-APR-97      CSWT    added code to provide this function can automatically
                               populate ProcuctionDateTime with the value of the time
                               at which science software processing of a given data
                               granule is being completes.
        24-Jun-97      CSWT    Fixed the code for setting up the attribute
                               ProductionDateTime by calling the UNIX system 
                               time is also available when writing the Meta
                               data to an ASCCI file using this function in 
                               which the mdHandle argument, the first input
                               parameter of the function PGS_MET_Write() is 
                               not the group name of INVENTORYMETADATA.
        22-Sep-97      CSWT    Fixed code of writing .met files to the product/data
                               directory instead of writing to the runtime (current
                               working) directory
        04-Dec-97      CSWT    Fixed code of replacing the C library function strtok used  
                               to break the string into a sequence of tokens with the C 
                               standard I/O function sscanf used to read characters,
                               interpret them according to a format, and store the results
                               in its argument/arguments  
                               Removed all the status checking codes followed after calling   
                               the function WriteLabel. The function WriteLabel is a avoid
                               function, set up the status checking code will cause the 
                               memory problem of attempting to read from uninitialized 
                               memory.
                               Changed the character string "PGS_MET_LoadAggregate", the  
                               second parameter passed into the function PGS_MET_ErrorMsg, 
                               to funcName declared for character string "PGS_MET_Write" 
        14-Jul-98      CSWT    Added code to provide the capability of writing the metadata 
                               file  that are associated with a particular Product ID and
                               Version number listed in the PCF file for the modes of
                               INTERMEDIATE INPUT and INTERMEDIATE OUTPUT (This change is
                               for NCR ECSed15639 about PGS_MET_Write does not support 
                               intermediate files)
	27-Jul-98 Abe Taaheri  Fixed the problem with the number of arguments in
	                       PGS_IO_Gen_Temp_Open

END_FILE_PROLOG
*******************************************************************************/

/* include files */

#define _BSD_COMPAT /* this definition is necessary for the local_nc.h to compile with -ansiposix flag */
#include <stdio.h>
#include <time.h>
#include "PGS_MET.h"
#include <PGS_MEM.h>

#include <errno.h>

/* odl include files */

#include <CUC/odldef.h>
#include <CUC/odlinter.h>

/* hdf files */

#ifdef DEC_ALPHA        /* DEC Alpha platform */
#include <rpc/types.h>  /* this avoids typedef conflict with int32, uint32 */
#endif /* DEC_ALPHA */
#include <hdf.h>
#include <mfhdf.h>

#include "local_nc.h"
#include "hfile.h"

/***************************************************************************
BEGIN_PROLOG:

TITLE: 
        Writes metadata and their values to HDF attributes
  
NAME:  
        PGS_MET_Write()

SYNOPSIS:
C:
        #include "PGS_MET.h"

	PGSt_SMF_status
		PGS_MET_Write(
			PGSt_MET_handle mdHandle,
			char *          hdfAttrName,
			PGSt_integer    hdfFileId)
FORTRAN:
         include 'PGS_MET_13.f'
	 include 'PGS_MET.f' 
	 include 'PGS_SMF.h'
         integer function pgs_met_write(mdHandle, hdfAttrName, hdfFileId)

	 character*   mdHandle
         character*   hdfAttrName
	 integer      hdfFileId
   
DESCRIPTION:
	This is the final tool that pge uses when all the metadata parameters
	are set in memory. The tool checks that all the mandatory parameters 
	are set. 

INPUTS:
        Name            Description            Units    	Min	Max
        ----            -----------             -----   	---     ---
	mdHandle	metadata group		none		N/A	N/A
			in MCF
        hdfAttrName	HDF file attribute 	none            N/A     N/A
			name
	hdfFileId	HDF file ID		none            N/A     N/A

OUTPUTS:
	None

RETURNS:   
   	PGS_S_SUCCESS			
	PGSMET_E_NO_INITIALIZATION	Metadata file is not initialized
	PGSMET_E_ODL_MEM_ALLOC		ODL routine failed to malloc memmory space
	PGSMET_E_GROUP_NOT_FOUND	No group called <name> found in the MCF
	PGSMET_E_OPEN_ERR		Unable to open <temporary> file with file id <fileId>
	PGSMET_E_SD_SETATTR		Unable to set the HDF file attribute
					NOTE: HDF4.0r1p1 and previous versions of HDF have imposed a limit
					of 32k on the size attribute
	PGSMET_E_MALLOC_ERR		Unable to allocate memory for the hdf attribute
	PGSMET_E_MAND_NOT_SET		Some of the mandatory parameters were not set
					Note: HDF attribute is still written out
	PGSMET_E_FGDC_ERR		Unable to conver UTC input date time string to FGDC values
	PGS_MET_E_ILLEGAL_HANDLE        Handle is illegal. Check that initialization has taken place
	PGS_MET_E_HDFFILENAME_ERR	Unable to obtain hdf file name
	PGSMET_E_MET_ASCII_ERR		Unable to open MET ascii file
	

EXAMPLES:
C:

This is an extract from the main example in PGS_MET_Init.c

			ret = PGS_MET_Write(handles[ODL_IN_MEMMORY], NULL, (PGSt_integer)NULL);
			if(ret != PGS_S_SUCCESS)
			{	
				printf("ASCII Write failed\n");
			}
			ret = PGS_MET_Write(handles[INVENTORYMETADATA], "metadata", sdid);
			if(ret != PGS_S_SUCCESS)
                        {
                                printf("HDF Write failed\n");
                        }

FORTRAN:
	This is just an extract of the call from a full example given in PGS_MET_Init() prolog.

C	ascii file is written to file with id 10255
			result = pgs_met_write(groups(ODL_IN_MEMORY), dummyStr, dummyInt)
			if(result.NE.PGS_S_SUCCESS  .AND. result.NE.PGSMET_MAND_NOT_SET) then
				print *,"ASCII Write failed"
			endif
C			write the first group as attribute 
			result = pgs_met_write(groups(INVENTORYMETADATA), "coremetadata.0", sdid)
			if(result.NE.PGS_S_SUCCESS .AND. result.NE.PGSMET_MAND_NOT_SET) then
				print *,"ASCII Write failed"
			endif
NOTES:
 	NONE

REQUIREMENTS:
        PGSTK-0290, PGSTK-0380, PGSTK-0400, PGSTK-0450 PGSTK-0510

DETAILS:
	This routine can be used multiple times to write/attach separate mastergroups
	as local or global HDF attributes.

	Addendum for tk5+

	A number of changes have been introduced for tk5+:

	1.	A separate ascii dump of the MCF in memory can now be written out. The name of the file 
		should be defined in the PCF. A file id has been reserved  for this purpose and is 10255
		(PGSd_MET_ASCII_DUMP). The user needs to call PGS_MET_Write with mdHandle[0] and the rest
		of the arguments set to null.

		**RELEASE A ADDENDUM**

		This is still true. However, its now a default behavior. If user wishes to use his/her 
		own file, he/she must supply the id in the "hdffileid" argument. Note that in this case
		the file id is the toolkit file id and not the hdf file id.

		**END RELEASE A ADDENDUM**

	2.	If MANDATORY parameters are not set, an error PGSMET_E_MAND_NOT_SET is returned. The value 
		of the metadata is set to as follows:
				DATA_LOCATION		VALUE
				PGE			"NOT SET"
				PCF			"NOT FOUND"
				MCF			"NOT SUPPLIED"
		The writing of the hdf header is not affected

		NOTE:	A warning PGSMET_W_METADATA_NOT_SET is issued if MANDATORY has the value FALSE in the MCF
			
		**RELEASE A ADDENDUM**

			Also anything which is not mandatory and not set by the user is removed

		**END RELEASE A ADDENDUM**

	3.	Only system errors such as memory failurer, file openings etc should be able to abort the write procedure.

	4.	NUM_VAL and CLASS fields are written in the hdf header

	5.	This routine creates two new metadata objects for metadata with type DATETIME. The UTC time
		string is used to create an FGDC date object and FGDC time object as strings. The object names 
		are same as parent metadata with DATE and TIME appended at the end.

		eg.	OBJECT                 = DATE_TIME
    			NUM_VAL              = 1
    			VALUE                = "1989-04-11T12:30:45.7Z"
  			END_OBJECT             = DATE_TIME
 
  			OBJECT                 = DATE_TIMEDATE
    			NUM_VAL              = 1
    			VALUE                = "19890411"
  			END_OBJECT             = DATE_TIMEDATE
 
  			OBJECT                 = DATE_TIMETIME
    			NUM_VAL              = 1
    			VALUE                = "123045700000Z"
  			END_OBJECT             = DATE_TIMETIME

		The original MCF simply contained the definition for DATE_TIME
	**RELEASE A ADDENDUM**

	1.	When inventory metadata group is written out, it is also dumped in a file as ascii with the name
		hdffilename.met in the current working directory. "hdffilename" is the hdf file that metadata is
		written to.

	2.	For the following metadata of type DATETIME additional metadata is produced as follows:

			CALENDERDATATIEM	produces	CALENDERDATE and CALENDERTIME
			RANGEBEGINNINGDATETIME  produces 	RANGEBEGINNINGDATE and RANGEBEGINNINGTIME
			RANGEENDINGDATETIME  	produces        RANGEENDINGDATE and RANGEENDINGTIME

		e.g.	"1989-04-11T12:30:45.7Z" produces	"1989-04-11" and "12:30:45.70000"

		This is only true for these three metadata and therefore point 5. in the previous
		addendum is now obsolete

	3.	User no longer has to worry about the size of MCF exceeding the HDF limit on attribute sizes
		This is now handled internally. 


	**END RELEASE A ADDENDUM**

			
GLOBALS:
	PGSg_MET_MasterNode

FILES:
	product files

FUNCTIONS_CALLED:
	PGS_SMF_SetStaticMsg
	PGS_MET_CheckAttr
	PGS_MEM_Malloc
	PGS_MEM_Free
	PGS_IO_Gen_Open
	PGS_IO_Gen_Temp_Open
	PGS_IO_Gen_Close
	PGS_IO_Gen_Close
	PGS_IO_Gen_Temp_Delete
	PGS_MET_ErrorMsg
	FindGroup
	NextGroup
	RemoveAggregate
	CopyAggregate
	WriteLabel
	SDsetattr
	
END_PROLOG:
***************************************************************************/


PGSt_SMF_status
PGS_MET_Write(
             PGSt_MET_handle mdHandle,  /* Handle for the MCF meta data group in memory */
             char *          hdfAttrName, /* name of hdf attribute to be attached */
             PGSt_integer    hdfFileId) /* HDF product file id or local
                                          * data id
                                          */
{
        struct tm *local;
        time_t lt;

	extern AGGREGATE	PGSg_MET_MasterNode[];
	extern char		PGSg_MET_AttrHandle[][PGSd_MET_SIGNATURE_L];
	PGSt_IO_Gen_FileHandle  *fileHandle = NULL;
	AGGREGATE		groupNode = NULL;
	AGGREGATE               tempNode = NULL;
	AGGREGATE               newNode = NULL;
	AGGREGATE 		oldParentNode  = NULL;
	PGSt_SMF_status 	retVal = PGS_S_SUCCESS;	/* SDPS toolkit ret value */
	PGSt_SMF_status         checkRetVal = PGS_S_SUCCESS;
	int32			hdfRetVal = 0;
	intn			hdfRet;
        char                    details[PGS_SMF_MAX_MSG_SIZE];
	char *			errInserts[PGSd_MET_MAX_ERR_INSERTS] = {NULL}; 
							/* Dynamic strings inserted in 
							 * error messages
							 */
	char *			funcName = "PGS_MET_Write";
        char                    *createtime=NULL;
	char 			*mcfBuffer = NULL;
	char 			fileIdStr[10];
      	char *			newAttrName= NULL;
        PGSt_integer            logicalId; /* user defined logical ID that
                                              represents the Product listed
                                              in the PCF  */

	PGSt_integer		i = 0; /* loop count */
	PGSt_integer		numBytesRead = 0;
        PGSt_integer            version;   /* The number of versions associated
                                              with the requested product Logical
                                              ID */
	PGSt_integer		numAttr = 0;
	PGSt_integer		filType;
	FILE *			asciiFile = NULL;
        char                    configParam[PGSd_PC_VALUE_LENGTH_MAX]; 
                                           /* a string represents the value of
                                              the configration parameter */
        char                    configParamm[PGSd_PC_VALUE_LENGTH_MAX]; 
                                           /* a string represents the value of
                                              the configration parameter */
        char                    configDataString1[PGSd_PC_VALUE_LENGTH_MAX]; 
                                           /* a substring extracts from the value
                                              of the configration parameter */
        char                    configDataString2[PGSd_PC_VALUE_LENGTH_MAX]; 
                                           /* a substring extracts from the value
                                              of the configration parameter */
        char                    handleString[PGSd_PC_VALUE_LENGTH_MAX]=" ";
	char * 			hFileName = NULL;
	char                    asciiFileRef[PGSd_PC_FILE_PATH_MAX];
	char                    asciiFileName[PGSd_PC_FILE_PATH_MAX];
	intn                    access;
	intn			attach;
	NC			*hdfhandle = NULL;
	PGSt_integer            mcfNumber = 0;
        char *                  mcfNumPtr = NULL;
        char *                  handlePtr = NULL;
        char *                  errPtr = NULL;
        char                    groupHandle[PGSd_MET_GROUP_NAME_L] = "";
        char                    locMdHandle[PGSd_MET_GROUP_NAME_L] = "";

	char                    signature[PGSd_MET_SIGNATURE_L] = "Razia";
	PGSt_integer            fileInBuffer = 1;
        PGSt_integer            bufferIndex = 0;
        PGSt_IO_Gen_Duration    file_duration = 1;
        PGSt_boolean            tempReferenceflag = PGS_FALSE;
        PGSt_boolean            existFlag = PGS_FALSE;

	/* clear the errno for the ODL routines */

        if (errno == ERANGE)
        {
                errno = 0;
        }

        /* find out the MCF number contained in the mdHandle argument*/

	if(mdHandle != NULL)
	{
        	mcfNumPtr = strrchr(mdHandle, '#');
	}
        if(mcfNumPtr == NULL)
        {
                (void) PGS_SMF_SetStaticMsg(PGSMET_E_ILLEGAL_HANDLE, funcName);
                return(PGSMET_E_ILLEGAL_HANDLE);
        }
        /* Generate the Greenwich mean time of the system to be the value of
           metadata parameter, ProductionDataTime,  in order this attribute
           value will be automatically written to an ASCII file and to an HDF-EOS
           file as appropriate */
 
        strcpy(groupHandle, mdHandle);
        handlePtr = strrchr(groupHandle, '#');
 
        if (handlePtr != NULL)
        {
              sscanf(groupHandle, "%[^#]", handleString);
            if ((strcmp(handleString,PGSd_MET_INVENT_DATA) == 0) || (strcmp(handleString,"MCF") == 0))
            {
	        createtime=(char *) malloc(30);
                lt=time(NULL);
                local=gmtime(&lt);
                strftime(createtime,80,"%Y-%m-%dT%X.000Z",local);
                retVal=PGS_MET_SetAttr(mdHandle,"ProductionDateTime",&createtime);
                if (retVal != PGS_S_SUCCESS)
                {
		    free(createtime);
                    (void) PGS_SMF_SetStaticMsg(PGSMET_E_SET_ATTRIBUTE_ERROR, funcName);
                    return(PGSMET_E_SET_ATTRIBUTE_ERROR);
                }
                free(createtime);
            }
        }

        mcfNumPtr++;
        mcfNumber = (PGSt_integer)strtol(mcfNumPtr, &errPtr, 10);
        if(mcfNumber < 0 || mcfNumber > PGSd_MET_NUM_OF_MCF || errPtr == mcfNumPtr)
        {
                (void) PGS_SMF_SetStaticMsg(PGSMET_E_ILLEGAL_HANDLE, funcName);
                return(PGSMET_E_ILLEGAL_HANDLE);
        }

	

	/* if master node is null then initialization has not taken palce */

        if(PGSg_MET_MasterNode[mcfNumber] == NULL)
        {
                (void) PGS_SMF_SetStaticMsg(PGSMET_E_NO_INITIALIZATION, funcName);
                return(PGSMET_E_NO_INITIALIZATION);
        }

	/* now make a local copy of the mdHandle */

        strcpy(locMdHandle, mdHandle);
        mcfNumPtr = strrchr(locMdHandle, '#');
        *mcfNumPtr = '\0';

	/* make a copy of the master node and use that */

	tempNode = CopyAggregate(PGSg_MET_MasterNode[mcfNumber]);
	if(tempNode == NULL)
	{
		(void) PGS_MET_ErrorMsg(PGSMET_E_ODL_MEM_ALLOC,
                                     funcName, 0, errInserts);
		return(PGSMET_E_ODL_MEM_ALLOC);
	}

	/* find the group in the MCF to be written */

	groupNode = FindGroup(tempNode, locMdHandle);
	if(groupNode == NULL)
	{
            errInserts[0] = locMdHandle;

            /* error message is:
            "No group called <name> found in the MCF" */

            (void) PGS_MET_ErrorMsg(PGSMET_E_GROUP_NOT_FOUND,
            funcName, 1, errInserts);
	    tempNode = RemoveAggregate(tempNode);
            return(PGSMET_E_GROUP_NOT_FOUND);
        }
	
	/* check that all the mandatory parameters, if any are set */
	
	checkRetVal = PGS_MET_CheckAttr(groupNode);
        if(checkRetVal != PGS_S_SUCCESS)
        {
                if(checkRetVal == PGSMET_E_ODL_MEM_ALLOC)
                {
                        tempNode = RemoveAggregate(tempNode);
                        return(PGSMET_E_ODL_MEM_ALLOC);
                }
		if(checkRetVal == PGSMET_E_FGDC_ERR)
                {
                        tempNode = RemoveAggregate(tempNode);
                        return(PGSMET_E_FGDC_ERR);
                }
        }
	if(strcmp(locMdHandle, PGSd_MET_MCF_NAME) == 0)
        {
	        /* write metadata for non-HDF file */

                /* if input hdfFileId is NULL use default value */

		if(hdfFileId == (PGSt_integer)NULL)
		{
                        hdfFileId = PGSd_MET_ASCII_DUMP;
		}
                
                /* retrieve the user defined configuration paratemeters 
                   associated with the input logical identifer from the
                   PCF */
		
                retVal = PGS_PC_GetConfigData(hdfFileId,configParam);
		
                /* if the return status is unsuccessful, then set up the 
                   return value */
		
                if (retVal != PGS_S_SUCCESS)
                {
#ifndef PGS_DAAC
		    /* if this is not the DAAC version then try the old method
		       of writing ASCII metadata */

		    goto OLD_CODE;
#else
                    retVal = PGSMET_E_MET_ASCII_ERR;
                    sprintf(details,
                            "Unable to write metadata for non-HDF data file. "
                            "Invalid hdfFileId value (%d) specified.",
                            hdfFileId);
                    PGS_SMF_SetDynamicMsg(retVal, details, funcName); 
                    return retVal;
#endif
                }
  		
                /* if the return status is successful, then call strtok()
                   to get tokens, the logical ID and version number
                   associated with a data file, from the configuration   
                   parameters  */
		
                strcpy(configParamm,configParam);
                sscanf(configParam,"%[^:]",configDataString1);

                if (configDataString1 == NULL)
                {
                    retVal = PGSMET_E_MET_ASCII_ERR;
                    sprintf(details,
                            "Unable to write metadata for non-HDF data file. "
                            "Invalid data file association for specified hdfFileId (%d).",
                            hdfFileId);
                    PGS_SMF_SetDynamicMsg(retVal, details, funcName);
                    return retVal;
                }
                logicalId=atoi(configDataString1);
 		
                sscanf(configParamm,"%*[^:]:%s",configDataString2);

                if (configDataString2 == NULL)
                {
                    retVal = PGSMET_E_MET_ASCII_ERR;
                    sprintf(details,
                            "Unable to write metadata for non-HDF data file. "
                            "Invalid data file association for specified hdfFileId (%d).",
                            hdfFileId);
                    PGS_SMF_SetDynamicMsg(retVal, details, funcName);
                    return retVal;
                }
		
                version=atoi(configDataString2);
                if (version <= 0)
                {
                    retVal = PGSMET_E_MET_ASCII_ERR;
                    sprintf(details,
                            "Unable to write metadata for non-HDF data file. "
                            "Invalid data file association for specified hdfFileId (%d).",
                             hdfFileId);
                    PGS_SMF_SetDynamicMsg(retVal, details, funcName);
                    return retVal;
                }
 		
                /* if the return status is successful, then call PGS_PC_GetReferenceType()
                   with the logical ID */
 
                retVal = PGS_PC_GetReferenceType(logicalId,
                                                 &filType);
 
                /* if the return status is successful, then call PGS_PC_GetRef-
                   erence() to retrieve the physical file reference associated 
                   with the logical ID */
  	
                if (retVal == PGS_S_SUCCESS &&
                    (filType == PGSd_PC_INPUT_FILE_NAME ||
                     filType == PGSd_PC_OUTPUT_FILE_NAME ||
                     filType == PGSd_PC_SUPPORT_IN_NAME || 
                     filType == PGSd_PC_SUPPORT_OUT_NAME ))
                {
                     retVal = PGS_PC_GetReference(logicalId,
		         			  &version,
					          asciiFileRef);
                }
                else if (retVal == PGS_S_SUCCESS &&
                         (filType == PGSd_PC_TEMPORARY_FILE ||
                          filType == PGSd_PC_INTERMEDIATE_INPUT ||
                          filType == PGSd_PC_INTERMEDIATE_OUTPUT))
                {
                     retVal = PGS_IO_Gen_Temp_Reference(file_duration,
                                                        logicalId,
                                                        PGSd_IO_Gen_Write,
                                                        asciiFileRef,
                                                        &existFlag);
                     tempReferenceflag = PGS_TRUE;
                }

		
                if (retVal != PGS_S_SUCCESS)
                {
                    retVal = PGSMET_E_MET_ASCII_ERR;
                    sprintf(details,
                            "Unable to write metadata for non-HDF data file. "
                            "Invalid data file association (%d:%d) "
                            "for specified hdfFileId (%d).",
                            logicalId,version,hdfFileId);
                    PGS_SMF_SetDynamicMsg(retVal, details, funcName);
                    return retVal;
                }

                /* non HDF_EOS product output .met files will contains the same full path/name 
                   as the non HDF_EOS product input files */
                  
	        strcpy(asciiFileName,asciiFileRef);	
                strcat(asciiFileName,".met");
                fileHandle = fopen(asciiFileName, "w"); 
                

           	if(fileHandle == NULL)
           	{
                	sprintf(fileIdStr, "%d", hdfFileId);
                	errInserts[0] = "output" ;
                	errInserts[1] = fileIdStr;
 
                	/* error message is:
                	"Unable to open <input> file with file id <aggregate name>" */
 
                	(void) PGS_MET_ErrorMsg(PGSMET_E_OPEN_ERR, funcName,
                                    2, errInserts);
			tempNode = RemoveAggregate(tempNode);
                	return(PGSMET_E_OPEN_ERR);
           	}
                WriteLabel(fileHandle, tempNode);
                tempNode = RemoveAggregate(tempNode);
		fclose(fileHandle);
        	return(PGS_S_SUCCESS);
#ifndef PGS_DAAC
	OLD_CODE:
                if (tempReferenceflag == PGS_FALSE)
                {
         		if(hdfFileId != (PGSt_integer)NULL)
         		{
			    retVal = PGS_IO_Gen_Open(hdfFileId, PGSd_IO_Gen_Write, &fileHandle, 1);
	        	}
	        	else
	        	{
	         	    retVal = PGS_IO_Gen_Open(PGSd_MET_ASCII_DUMP, PGSd_IO_Gen_Write, &fileHandle, 1);
		        }
		}
                else
                {
                        if(hdfFileId != (PGSt_integer)NULL)
                        {
			    retVal = PGS_IO_Gen_Temp_Open(file_duration, hdfFileId, PGSd_IO_Gen_Write, &fileHandle);
                        }
                        else
                        {
                            retVal = PGS_IO_Gen_Open(PGSd_MET_ASCII_DUMP, PGSd_IO_Gen_Write, &fileHandle, 1);
                        }
                }
           	if(retVal != PGS_S_SUCCESS)
           	{
                	sprintf(fileIdStr, "%d", PGSd_MET_ASCII_DUMP);
                	errInserts[0] = "output" ;
                	errInserts[1] = fileIdStr;
 
                	/* error message is:
                	"Unable to open <input> file with file id <aggregate name>" */
 
                	(void) PGS_MET_ErrorMsg(PGSMET_E_OPEN_ERR, funcName,
                                    2, errInserts);
			tempNode = RemoveAggregate(tempNode);
                	return(PGSMET_E_OPEN_ERR);
           	}

		/* write a warning message to the log to let the user know that
		   an obsolete method of writing ASCII metadata is being used */

                /* this next open brace is only there to create the new 
                   message variable and call PGS_SMF_GenerateStatusReport(). */
                {
		char strPtr[] =
		  "\n"
		  "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"
		  "!!                   W A R N I N G                     !!\n"
		  "!! Using  obsolete method of writing ASCII metadata.   !!\n"
		  "!! A direct reference to an ASCII metadata file should !!\n"
		  "!! no longer be used in the PCF.  Instead, replace the !!\n"
		  "!! logical ID of the ASCII metadata file in the PCF    !!\n"
		  "!! with a runtime parameter using the same logical ID  !!\n"
		  "!! and having a value of <logical_id>:<version_number> !!\n"
		  "!! where <logical_id> and <version_number> are the     !!\n"
		  "!! logical ID and version number (respectively) of the !!\n"
		  "!! non-HDF output product for which this ASCII         !!\n"
		  "!! metadata is being written out.                      !!\n"
		  "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n";

		PGS_SMF_GenerateStatusReport(strPtr);
                }  /* close brace - only opened for new variable */

                WriteLabel(fileHandle, tempNode);
		tempNode = RemoveAggregate(tempNode);
		(void) PGS_IO_Gen_Close(fileHandle);
        	return(PGS_S_SUCCESS);
#endif
        }
	/* find out the hdf file name for later use */

        if(hdfFileId != (PGSt_integer)NULL)
        {
                hdfhandle = NC_check_id((hdfFileId >> 24) & 0xff);
		hdfRet = FAIL;
		if(hdfhandle != NULL)
		{
                	hdfRet = Hfidinquire(hdfhandle->hdf_file, &hFileName, &access, &attach);
		}
		if(hdfRet != SUCCEED)
                {
                        /* error message is:
                        "Unable to obtain hdf filename" */

                        (void) PGS_MET_ErrorMsg(PGSMET_E_HDFFILENAME_ERR, funcName,
                                   0, errInserts);
                        tempNode = RemoveAggregate(tempNode);
                        return(PGSMET_E_HDFFILENAME_ERR);
                }
        }

        /* prepare the signature */

        if(hFileName != NULL && hdfAttrName != NULL)
        {
                sprintf(signature,"%s/%s", hFileName, hdfAttrName);
        }
	/* open temporary file to and write the group node */
	retVal = PGS_IO_Gen_Open(PGSd_MCF_TEMP_FILE,
                                        PGSd_IO_Gen_Write, &fileHandle, 1);
        if(retVal != PGS_S_SUCCESS)
        {
                sprintf(fileIdStr, "%d", PGSd_MCF_TEMP_FILE);
                errInserts[0] = "temporary";
                errInserts[1] = fileIdStr;
                /* error message is:
                "Unable to open <temporary> file with file id <fileId>" */

                (void) PGS_MET_ErrorMsg(PGSMET_E_OPEN_ERR, funcName,
                                    2, errInserts);
		tempNode = RemoveAggregate(tempNode);
                return(PGSMET_E_OPEN_ERR);
        }

	/* write the label to a temporary file. This is necessary to convert
	 * odl memory representation into an odl ASCII representation
	 */

	(void) rewind((FILE *)fileHandle);
	oldParentNode = NextGroup(tempNode);
	if(oldParentNode!= NULL)
	{
		newNode = oldParentNode->right_sibling;
	}
	else
	{
		newNode = NULL;
	}
	if(oldParentNode != NULL)
	{
		/* remove unnecessary groups */
		do
		{
			if(strcmp(groupNode->name, oldParentNode->name) != 0)
			{
				oldParentNode = RemoveAggregate(oldParentNode);
			}
			oldParentNode = newNode;
			if(newNode != NULL)
			{
				newNode = newNode->right_sibling;
			}
		}
		while(oldParentNode != NULL);
	}

	/* write out the inventory matadata as a separate ascii file */
	if(strcmp(locMdHandle, PGSd_MET_INVENT_DATA) == 0)
        {
        /*        hFileId = SDidtofid(hdfFileId); */

		hdfhandle = NC_check_id((hdfFileId >> 24) & 0xff);
                hdfRet = Hfidinquire(hdfhandle->hdf_file, &hFileName, &access, &attach);
                if(hdfRet != SUCCEED)
                {
                        /* error message is:
                        "Unable to obtain hdf filename" */

                        (void) PGS_MET_ErrorMsg(PGSMET_E_HDFFILENAME_ERR, funcName,
                                   0, errInserts);
			(void) PGS_IO_Gen_Close(fileHandle);
                        tempNode = RemoveAggregate(tempNode);
                        return(PGSMET_E_HDFFILENAME_ERR);
                }

                /* HDF_EOS product output .met files will contains the same full path/name
                   as the HDF_EOS product input files */

                sprintf(asciiFileName, "%s.met", hFileName);
                asciiFile = fopen(asciiFileName, "w");
                if(asciiFile == NULL)
                {
                        /* error message is:
                        "Unable to open MET ascii file" */

                        (void) PGS_MET_ErrorMsg(PGSMET_E_MET_ASCII_ERR, funcName,
                                   0, errInserts);
                        tempNode = RemoveAggregate(tempNode);
                        (void) fclose(asciiFile);
			(void) PGS_IO_Gen_Close(fileHandle);
                        return(PGSMET_E_MET_ASCII_ERR);
                }
                WriteLabel(asciiFile, tempNode);
                (void) fclose(asciiFile);
        }
	WriteLabel(fileHandle, tempNode);
	(void) PGS_IO_Gen_Close(fileHandle);
	tempNode = RemoveAggregate(tempNode);

	/* open the file again for reading */

	retVal = PGS_IO_Gen_Open(PGSd_MCF_TEMP_FILE,
                              		PGSd_IO_Gen_Read, &fileHandle, 1);
        if(retVal != PGS_S_SUCCESS)
        {
                sprintf(fileIdStr, "%d", PGSd_MCF_TEMP_FILE);
                errInserts[0] = "temporary";
                errInserts[1] = fileIdStr;
                /* error message is:
                "Unable to open <temporary> file with file id <aggregate name>" */

                (void) PGS_MET_ErrorMsg(PGSMET_E_OPEN_ERR, funcName,
                                    2, errInserts);
                return(PGSMET_E_OPEN_ERR);
        }

	/* read single characters from the file into a buffer */

       retVal = PGS_MEM_Malloc((void **)&newAttrName, (PGSt_integer)(strlen(hdfAttrName) + 11));
        if(retVal != PGS_S_SUCCESS)
        {
                /* error message is:
                "Unable to allocate memory for the hdf attribute" */
 
                (void) PGS_MET_ErrorMsg(PGSMET_E_MALLOC_ERR, funcName,
                                    0, errInserts);
		(void) PGS_IO_Gen_Close(fileHandle);
                return(PGSMET_E_MALLOC_ERR);
        }
       retVal = PGS_MEM_Malloc((void **)&mcfBuffer, (PGSt_integer)MAX_ORDER);
        if(retVal != PGS_S_SUCCESS)
        {
                /* error message is:
                "Unable to allocate memory for the hdf attribute" */
 
                (void) PGS_MET_ErrorMsg(PGSMET_E_MALLOC_ERR, funcName,
                                    0, errInserts);
		(void) PGS_IO_Gen_Close(fileHandle);
                return(PGSMET_E_MALLOC_ERR);
        }
       numAttr = 0;
        while(feof(fileHandle) == 0)
        {

	  numBytesRead = fread(mcfBuffer, sizeof(char), (size_t)MAX_ORDER, fileHandle);
	  if(numBytesRead != 0) /* if numBytesRead is equal to MAX_ORDER */
	    {
	      numAttr++;
	    }
        }
	(void) rewind(fileHandle);
       for (i = 0; i < numAttr; i++)
	 {
	   numBytesRead = fread(mcfBuffer, sizeof(char), (size_t) MAX_ORDER, fileHandle);

/*
	   mcfBuffer[numBytesRead + 1]= '\0';
*/

	   /* attach the buffer to the HDF attribute */
	   if(i == 0)
	     {
	       sprintf(newAttrName, "%s", hdfAttrName);
	     }
	   else
	     {
	       sprintf(newAttrName, "%s.%d", hdfAttrName, i);
	     }
	   hdfRetVal = SDsetattr((int32)hdfFileId, newAttrName,
			      DFNT_CHAR, (int32) (numBytesRead), mcfBuffer);

	   if(hdfRetVal != SUCCEED)
	     {
/*
       	         (void) HEprint(stderr, (int32) 0);
*/
                         /* error message is:
                        "Unable to set the HDF file attribute" */

                        (void) PGS_MET_ErrorMsg(PGSMET_E_SD_SETATTR, funcName,
                                    0, errInserts);
       	        (void) PGS_MEM_Free(mcfBuffer);
        	         mcfBuffer = (char *) NULL;
	         (void) PGS_MEM_Free(newAttrName);
	         newAttrName = (char *) NULL;
			(void) PGS_IO_Gen_Close(fileHandle);
                          return(PGSMET_E_SD_SETATTR);
	     }
	 }
       (void) PGS_IO_Gen_Close(fileHandle);
       /* (void) PGS_IO_Gen_Temp_Delete(PGSd_MCF_TEMP_FILE); */
       (void) PGS_MEM_Free(mcfBuffer);
        mcfBuffer = (char *) NULL;
       (void) PGS_MEM_Free(newAttrName);
        newAttrName = (char *) NULL;
	/* check that this particular attribute is not in the GetPCAttr buffers */
	bufferIndex = 0;
	if(hFileName != NULL && hdfAttrName != NULL)
        {
		while(PGSg_MET_AttrHandle[bufferIndex][0] != '\0' && bufferIndex < PGSd_MET_NUM_OF_GROUPS)
        	{
	
                        fileInBuffer = strcmp(signature, PGSg_MET_AttrHandle[bufferIndex]);
                        if(fileInBuffer == 0)
                        {
				strcpy(PGSg_MET_AttrHandle[bufferIndex], "alward");
                                break;
                        }
                        bufferIndex++;
		}
        }
	
       if(checkRetVal != PGS_S_SUCCESS && checkRetVal != PGSMET_E_ODL_MEM_ALLOC && checkRetVal != PGSMET_W_META_NOT_SET)
        {
                (void) PGS_MET_ErrorMsg(checkRetVal,
                                     funcName, 0, errInserts);
                return(checkRetVal);
        }

	return(PGS_S_SUCCESS);
}
