/*-------------------------------------------------------------------------*/
/*                                                                         */
/*  COPYRIGHT[copyright mark] 1996, Hughes Aircraft Company, its vendors,  */
/*  and suppliers.  ALL RIGHTS RESERVED.                                   */
/*                                                                         */
/*-------------------------------------------------------------------------*/
/***************************************************************************
BEGIN_FILE_PROLOG:

FILENAME:
	PGS_PC_GenUniqueID.c

DESCRIPTION:
	This file contains the function PGS_PC_GenUniqueID().

AUTHOR:
	Ray Milburn / Applied Research Corp.

HISTORY:
	05-Aug-94 RM Initial Version
	14-Oct-94 RM Added call to PGS_SMF_CallerID
	11-Jul-95 RM Changed LONG to INT in cfortran.h macros to correct
			"Unaligned access error" as per DR ECSed00931.
 
END_FILE_PROLOG:
***************************************************************************/

#include <stdio.h>
#include <PGS_PC.h>

/***************************************************************************
BEGIN_PROLOG:

TITLE:
	Generate a unique ID.
 
NAME:
	PGS_PC_GenUniqueID()

SYNOPSIS:

C:	
		#include <PGS_PC.h>
	
		PGSt_SMF_status
		PGS_PC_GenUniqueID(
			PGSt_PC_Logical		prodID,
			char			*uniqueID)

FORTRAN:

		include 'PGS_SMF.f'
		include 'PGS_PC.f'
		include 'PGS_PC_9.f'

		integer function pgs_pc_genuniqueid(prodid,uniqueid)
			integer			prodid
			character*200 		uniqueid

DESCRIPTION:
	This tool may be used to generate a unique product identifier.  This
	identifier may be attached to file metadata to facilitate tracking of
	production output.  The identifier may include Production Run ID, the
	Science Software Program ID, and the actual Product ID.

INPUTS:
	Name		Description			Units	Min	Max

	prodID		The logical identifier as 
			defined by the user.  The user's
			definitions will be mapped into 
			actual identifiers during the 
			Integration & Test procedure.

OUTPUTS:
	Name		Description			Units	Min	Max

	uniqueID	The unique ID generated by this 
			function.  This ID will be
			returned as a string.  The ID is 
			guaranteed to be no greater than 
			PGSd_PC_LABEL_SIZE_MAX in length 
			(see PGS_PC.h).

RETURNS:
	PGS_S_SUCCESS               successful execution
	PGSPC_E_DATA_ACCESS_ERROR   error accessing PCS data

EXAMPLES:

C:	
	#define CERES3A 300

	PGSt_SMF_status	returnStatus;
	char		uniqueID[PGSd_PC_LABEL_SIZE_MAX];

	returnStatus = PGS_PC_GenUniqueID(CERES3A,uniqueID);

	if (returnStatus != PGS_S_SUCCESS)
		goto EXCEPTION;
	else
	{

	 /# attach uniqueID into file metadata field #/

	}
			.
			.
			.
	EXCEPTION:
		return returnStatus;

FORTRAN:	
		IMPLICIT NONE 

		integer		returnstatus
		character*200	uniqueid
		integer		pgs_pc_genuniqueid
		integer		ceres3a
		parameter	(ceres3a = 300)

		returnstatus = pgs_pc_genuniqueid(ceres3a,uniqueid)

		if (returnstatus .ne. pgs_s_success) then
			goto 9999
		else

C	attach uniqueid into file metadata field

		endif
			.
			.
			.
		return

NOTES:
	If more than one product is being generated from the same PGE, then the 
	appropriate product identifier must be used as input to this function
	when called from within the science software.  Upon entry into this function 
	all input values will be checked to determine that legal values were passed
	in.  If any value is illegal, the function will return the proper error 
	value to the calling function.  All unique identifier strings are guaranteed
	to be no greater than PGSd_PC_LABEL_SIZE_MAX characters in length (see
	PGS_PC.h).

REQUIREMENTS:  
	PGSTK-1280

DETAILS:
	NONE

GLOBALS:
	NONE

FILES:
	NONE

FUNCTIONS_CALLED:
	PGS_SMF_CallerID		 Get value of user variable
	PGS_PC_GetPCSData                Get process control information.
	PGS_SMF_SetStaticMsg		 Set a static message

END_PROLOG:
***************************************************************************/


PGSt_SMF_status
PGS_PC_GenUniqueID(                    /* generate unique id string */
    PGSt_PC_Logical   prodID,	       /* logical value */
    char             *uniqueID)        /* unique ID */

{
    char              str[PGSd_PC_LABEL_SIZE_MAX]; /* temp string */
    PGSt_integer      numFiles;        /* num files - dummy hold */
    PGSt_SMF_status   whoCalled;       /* user who called this function */
    PGSt_SMF_status   returnStatus;    /* function return */

/***************************************************************************
*    Initialize status.
***************************************************************************/
    returnStatus = PGS_S_SUCCESS;

/***************************************************************************
*    Initialize caller variable.
***************************************************************************/
    whoCalled = PGS_SMF_CallerID();

/***************************************************************************
*    Get production run id.
***************************************************************************/
    returnStatus = PGS_PC_GetPCSData(PGSd_PC_PRODUCTION_RUN_ID,0,str,&numFiles);

/***************************************************************************
*    If that worked, then let's start building our string and get the 
*    software id.
***************************************************************************/
    if (returnStatus == PGS_S_SUCCESS)
    {
        sprintf(uniqueID,"PRID - %s",str);
        returnStatus = PGS_PC_GetPCSData(PGSd_PC_SOFTWARE_ID,0,str,&numFiles);

/***************************************************************************
*    If retrieving the software id worked let's finish off our string.
***************************************************************************/
        if (returnStatus == PGS_S_SUCCESS)
        {
            sprintf(uniqueID,"%s SID - %s PRODID - %d",uniqueID,str,prodID);
        } 
    }
 
/***************************************************************************
*    Check to see if we had a problem.  Remember, we do not have to tell 
*    the user what exactly the problem was, but just let them know that
*    there was a problem.
***************************************************************************/
    if (returnStatus != PGS_S_SUCCESS)
    {
        returnStatus = PGSPC_E_DATA_ACCESS_ERROR;
    }

/***************************************************************************
*    Update the message log and return.
***************************************************************************/
    if (whoCalled != PGSd_CALLERID_SMF)
    {
        PGS_SMF_SetStaticMsg(returnStatus,"PGS_PC_GenUniqueID()");
    }
    return returnStatus;
}
