/*-------------------------------------------------------------------------*/
/*                                                                         */
/*  COPYRIGHT[copyright mark] 1996, Hughes Aircraft Company, its vendors,  */
/*  and suppliers.  ALL RIGHTS RESERVED.                                   */
/*                                                                         */
/*-------------------------------------------------------------------------*/
/***************************************************************************
BEGIN_FILE_PROLOG:

FILENAME:
	PGS_PC_TempDeleteCom.c

DESCRIPTION:
	This file contains all functions necessary to build 
	PGS_PC_TempDeleteCom.

AUTHOR:
	Ray Milburn / Applied Research Corp.

HISTORY:
	09-Jan-95 RM Initial Version
        26-Jan-95 RM Updated from comments of code inspection held
                        on 25-Jan-95.
 
END_FILE_PROLOG:
***************************************************************************/

#include <stdio.h>
#include <PGS_PC.h>

/***************************************************************************
BEGIN_PROLOG:

TITLE:  
	Delete temporary file command.
 
NAME:  
	PGS_PC_TempDeleteCom

SYNOPSIS:
	PGS_PC_TempDeleteCom <logical ID>

DESCRIPTION:
	This program will flag a temporary file as deleted in the 
	PCF or shared memory at the command line.
 
INPUTS:
        Name            Description                     Units   Min     Max

	argc		number of command line arguments

	argv[0]		executable name (not processed 
			but listed here anyway)

	argv[1]		logical ID of the temporary file
			to be deleted.

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

RETURNS:
	PGS_S_SUCCESS
        PGS_SH_SYS_PARAM
	PGS_SH_PC_NODATA
	PGS_SH_PC_TOOLERROR

EXAMPLES:
	#  This is within a shell script - probably within the 
	#  PGE script.

	LogicalID=12297

	#  Delete the temporary file with the logical ID 12297
	PGS_PC_TempDeleteCom $LogicalID
	RETVAL=$?

	#  Check the return value
	if [ $RETVAL -eq 0 ]
	then
	#  continue normal processing
	else
	#  report an error found
	fi
	.
	.	
	.

NOTES:
	This program is designed to be run from within the PGE.

REQUIREMENTS:
	PGSTK-0360
	PGSTK-521

DETAILS:
	NONE	

GLOBALS:
	NONE

FILES:
	NONE	

FUNCTIONS_CALLED:
	PGS_PC_PutPCSData		- update the PCF or shared memory
					containing the PCS data.
	PGS_PC_BuildNumericInput	- build a number from Com Tool input.
        PGS_SMF_TestStatusLevel         - Determine the level of the
                                         error/status code

END_PROLOG:
***************************************************************************/

/***************************************************************************
*    Function main().
***************************************************************************/
int
main(                                    /* main function */
    int                 argc,            /* number of command line arguments */
    char               *argv[])          /* array of command line arguments */
{
    PGSt_integer        retNum;          /* number returned */
    PGSt_PC_Logical     logicalID;       /* logical ID passed in */
    PGSt_SMF_status     returnStatus;    /* return from functions */

/***************************************************************************
*    Initialize variables.
***************************************************************************/
    returnStatus = PGS_S_SUCCESS;

/***************************************************************************
*    Determine if the correct number of parameters were passed in.
***************************************************************************/
    if (argc != 2)
    {
        returnStatus = PGS_SH_SYS_PARAM;
    }

/***************************************************************************
*    We have the correct number of input parameters, let's ensure that
*    the parameter sent in is all digit.  At the same time let's put it
*    in our own variable.
***************************************************************************/
    else
    {
        returnStatus = PGS_PC_BuildNumericInput(argv[1],&retNum);
   
/***************************************************************************
*    If everything is still OK then we need to load our logical ID 
*    variable and call PGS_PC_PutPCSData().
***************************************************************************/
        if (returnStatus == PGS_S_SUCCESS)
        {
            logicalID = (PGSt_PC_Logical) retNum;
            returnStatus = PGS_PC_PutPCSData(PGSd_PC_DELETE_TEMP,
                                             (void *) logicalID);

/***************************************************************************
*    If we did not get success from out PGS_PC_PutPCSData() call 
*    let's set up the proper return value.
***************************************************************************/
            if (returnStatus != PGS_S_SUCCESS)
            {
                switch (PGS_SMF_TestStatusLevel(returnStatus))
                {
                    case PGS_SMF_MASK_LEV_E:
                        returnStatus = PGS_SH_PC_TOOLERROR;
                        break;
                    case PGS_SMF_MASK_LEV_W:
                        returnStatus = PGS_SH_PC_NODATA;
                        break;
                }
            }
        }  /* end if */
    }  /* end else */

/***************************************************************************
*    Exit program.
***************************************************************************/
    exit(returnStatus);
}
